package com.vmware.psc.alerting.persister.integration.test;

import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.AlertType.ALERT_TYPE_THREAT;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.CbAlertConfidence.ALERT_CONFIDENCE_LOW;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.CbAlertConfidence.ALERT_CONFIDENCE_MEDIUM;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlertClassification.CbAlertClassificationType.FALSE_POSITIVE;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlertOperator.CbAlertOperatorType.SUPPRESSION;
import static com.carbonblack.protocols.psc.response.PSCRProto.ThreatScore.CbThreatSource.THREAT_SOURCE_DEFENSE_ANALYTICS;
import static org.awaitility.Awaitility.await;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assumptions.assumeTrue;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRAugmentedEventBatch;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.vmware.psc.alerting.data.uae.UaeAlertCreator;
import com.vmware.psc.alerting.data.uds.UdsEventCreator;
import com.vmware.psc.alerting.persister.AlertPersisterApplication;
import com.vmware.psc.alerting.persister.es.AlertEsClient;
import com.vmware.psc.alerting.persister.es.IPartitionStrategyV2;
import com.vmware.psc.alerting.persister.es.notes.NotesWriter;
import com.vmware.psc.alerting.persister.integration.config.TestConfig;
import com.vmware.psc.alerting.persister.integration.data.TestData;
import com.vmware.psc.alerting.persister.kinesis.producer.IKinesisProducer;
import com.vmware.psc.alerting.persister.kinesis.producer.KinesisKPLProducer;
import com.vmware.psc.alerting.persister.model.notes.Notes;
import com.vmware.psc.alerting.persister.properties.KinesisConfig;
import com.vmware.psc.alerting.sampledata.UdsInputSampleBuilder;
import com.vmware.psc.alerting.sampledata.UdsInputSamples;
import com.vmware.time.helper.TimeHelper;
import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.function.Predicate;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.junit.jupiter.api.*;
import org.junit.jupiter.api.parallel.Execution;
import org.junit.jupiter.api.parallel.ExecutionMode;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.opensearch.OpenSearchStatusException;
import org.opensearch.action.get.GetRequest;
import org.opensearch.action.get.GetResponse;
import org.opensearch.client.RequestOptions;
import org.opensearch.client.RestHighLevelClient;
import org.opensearch.search.fetch.subphase.FetchSourceContext;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.util.CollectionUtils;

@SpringBootTest(classes = {AlertPersisterApplication.class})
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@Slf4j
@TestMethodOrder(MethodOrderer.OrderAnnotation.class)
@RequiredArgsConstructor(onConstructor_ = {@Autowired})
@Execution(ExecutionMode.CONCURRENT)
class PersisterIntegrationTest {
  private final KinesisConfig config;
  private final TestConfig testConfig;
  private final IKinesisProducer testProducer;
  private final ObjectMapper objectMapper;
  private final Map<String, RestHighLevelClient> esClusterMap;
  private final IPartitionStrategyV2 monthlyFixedRetentionPartitionStrategy;
  private final AlertEsClient alertEsClient;

  private static final int MAX_WAIT_SECONDS = 65;

  private RestHighLevelClient notesClient() {
    return esClusterMap.get("notes");
  }

  private List<Arguments> allTestDevices() {
    assumeTrue(!CollectionUtils.isEmpty(testConfig.getOrgKeys()), "No org keys to test in!");
    return testConfig.getOrgKeys().stream()
        .map(
            orgKey -> {
              var deviceName = TestData.generateTestDeviceName();
              var defaultPolicy = TestData.generateTestPolicy();
              var device =
                  TestData.generateTestDevice(
                      orgKey,
                      deviceName,
                      defaultPolicy,
                      PSCRAugmentedEventBatch.CbDeviceOsType.WINDOWS);
              return Arguments.of(Named.of(String.format("[%s/%s]", orgKey, deviceName), device));
            })
        .toList();
  }

  @ParameterizedTest(name = "{0}: Test alert written successfully")
  @MethodSource("allTestDevices")
  void testEndToEnd(PSCRAugmentedEventBatch.CbDeviceMetadata.Builder device) throws Exception {
    String alertId = UUID.randomUUID().toString();

    // This alert is in today's index.
    // To avoid extra load on ES, other tests should use a date in the past
    var alertTimestamp = Instant.now();
    UdsInputSampleBuilder builder =
        UdsInputSamples.threatEventWithBlockedAnydesk()
            .backdatedTo(alertTimestamp)
            .withAlertCreationTimestamp(TimeHelper.toTimestamp(alertTimestamp))
            .withDevice(device)
            .withTriggeredAlertId(alertId);

    publishBatch(builder.build());

    PSCAlert.CbUaeAlert writtenAlert = waitForAlert(alertId, MAX_WAIT_SECONDS);
    assertEquals(alertId, writtenAlert.getAlert().getAlertIdGuid());
  }

  @ParameterizedTest(name = "{0}: Test alert written with aggregate ID instead of alert ID guid")
  @MethodSource("allTestDevices")
  void testEndToEndAggregated(PSCRAugmentedEventBatch.CbDeviceMetadata.Builder device)
      throws Exception {
    String alertId = UUID.randomUUID().toString();
    String aggregateGuid = UUID.randomUUID().toString();

    var alertTimestamp = Instant.now().minus(Duration.ofDays(31));
    UdsInputSampleBuilder builder =
        UdsInputSamples.threatEventWithBlockedAnydesk()
            .backdatedTo(alertTimestamp)
            .withAlertCreationTimestamp(TimeHelper.toTimestamp(alertTimestamp))
            .withDevice(device)
            .withTriggeredAlertId(alertId)
            .withAlertAggregation(
                PSCRProto.CbAlertAggregation.newBuilder()
                    .setAggregateGuid(aggregateGuid)
                    .setIsAggregated(true));

    publishBatch(builder.build());

    var writtenAlert = waitForAlert(aggregateGuid, MAX_WAIT_SECONDS);
    assertEquals(aggregateGuid, writtenAlert.getAlert().getAlertIdGuid());
    assertEquals(aggregateGuid, writtenAlert.getAlert().getAlertAggregation().getAggregateGuid());
  }

  @ParameterizedTest(name = "{0}: Test suppression notes are created")
  @MethodSource("allTestDevices")
  void testSuppressionNotesEndToEnd(PSCRAugmentedEventBatch.CbDeviceMetadata.Builder device)
      throws Exception {
    String alertId = UUID.randomUUID().toString();
    String suppressionRuleId = UUID.randomUUID().toString();

    var alertTimestamp = Instant.now().minus(Duration.ofDays(31));
    UdsInputSampleBuilder builder =
        UdsInputSamples.deviceControlEvent()
            .backdatedTo(alertTimestamp)
            .withAlertCreationTimestamp(TimeHelper.toTimestamp(alertTimestamp))
            .withDevice(device)
            .withTriggeredAlertId(alertId);

    // override the suppression rule ID so we don't have conflicts between tests
    // (the sample hardcoded one ID, which means subsequent tests will generate the same notes)
    var alert = builder.buildToMutableCbUaeAlert();
    var mdrState = alert.getAlertBuilder().getMdrStateDetailsBuilder().getWorkflowStateBuilder();
    mdrState.getAssignedToBuilder().setSuppressionRuleId(suppressionRuleId);
    mdrState.getChangedByBuilder().setSuppressionRuleId(suppressionRuleId);

    publishBatch(UdsEventCreator.convertToUds(alert.build()));

    waitForAlert(alertId, MAX_WAIT_SECONDS);
    var note = waitForNote(alert.getAlert().getAlertIdGuid(), suppressionRuleId, true);
    assertEquals("Internally, I donno what to say", note.getNote());

    note = waitForNote(alert.getAlert().getAlertIdGuid(), suppressionRuleId, false);
    assertEquals("Closed by mdr because of xxxx", note.getNote());
  }

  @ParameterizedTest(
      name = "{0}: Test suppression state from kinesis in MDR state details is preserved")
  @MethodSource("allTestDevices")
  void testEndToEndWithMdrState(PSCRAugmentedEventBatch.CbDeviceMetadata.Builder device)
      throws Exception {
    String alertId = UUID.randomUUID().toString();
    String suppressionRuleId = UUID.randomUUID().toString();

    var alertTimestamp = Instant.now().minus(Duration.ofDays(31));
    UdsInputSampleBuilder builder =
        UdsInputSamples.deviceControlEvent()
            .backdatedTo(alertTimestamp)
            .withAlertCreationTimestamp(TimeHelper.toTimestamp(alertTimestamp))
            .withDevice(device)
            .withTriggeredAlertId(alertId);

    var suppressionTime = TimeHelper.timestampNow();
    var suppressionOperator =
        PSCRProto.CbAlertOperator.newBuilder()
            .setName("SUPPRESSION_RULE")
            .setType(SUPPRESSION)
            .setSuppressionRuleId(suppressionRuleId)
            .build();

    var alert = UaeAlertCreator.convertAlertsAsMutable(builder.build()).get(0);
    var mdrState = alert.getAlertBuilder().getMdrStateDetailsBuilder();
    mdrState
        .getWorkflowStateBuilder()
        .setStatus("TRIAGE_COMPLETE")
        .setLastUpdateTimestamp(suppressionTime)
        .setChangedBy(suppressionOperator);

    mdrState
        .getClassificationBuilder()
        .setAlertClassification(FALSE_POSITIVE)
        .setLastUpdateTimestamp(suppressionTime)
        .setChangedBy(suppressionOperator);

    publishBatch(UdsEventCreator.convertToUds(alert.build()));

    var alertFromEs = waitForAlert(alertId, MAX_WAIT_SECONDS);
    PSCRProto.CbAlertStateDetails mdrStateFromEs = alertFromEs.getAlert().getMdrStateDetails();
    // Verify expected workflow state, suppression ID unchanged.
    assertEquals("TRIAGE_COMPLETE", mdrStateFromEs.getWorkflowState().getStatus());
    assertEquals(
        suppressionRuleId, mdrStateFromEs.getWorkflowState().getChangedBy().getSuppressionRuleId());

    // Verify create time matches workflow update time
    assertEquals(
        alertFromEs.getCreateTimestamp(),
        mdrStateFromEs.getWorkflowState().getLastUpdateTimestamp());

    // Verify expected classification state, suppression ID unchanged.
    assertEquals(FALSE_POSITIVE, mdrStateFromEs.getClassification().getAlertClassification());
    assertEquals(
        suppressionRuleId,
        mdrStateFromEs.getClassification().getChangedBy().getSuppressionRuleId());

    // Verify create time matches classification update time
    assertEquals(
        alertFromEs.getCreateTimestamp(),
        mdrStateFromEs.getClassification().getLastUpdateTimestamp());
  }

  private Notes waitForNote(String alertId, String suppressionRuleId, boolean isInternal) {
    var note =
        await()
            .atMost(Duration.ofSeconds(MAX_WAIT_SECONDS))
            .pollInterval(Duration.ofSeconds(2))
            .until(
                () ->
                    getNotesById(
                        NotesWriter.buildNotesId(alertId, suppressionRuleId, 1, isInternal)),
                Optional::isPresent);
    assertTrue(note.isPresent());
    return note.get();
  }

  @ParameterizedTest(name = "{0}: Test alerts updated with higher scores fill update details")
  @MethodSource("allTestDevices")
  void testEndToEndAlertScoreChange(PSCRAugmentedEventBatch.CbDeviceMetadata.Builder device)
      throws Exception {
    String aggregateGuid = UUID.randomUUID().toString();

    Instant firstEvent = Instant.now().minus(Duration.ofDays(31));

    // write one with aggregateGuid with score 5

    UdsInputSampleBuilder builder1 =
        UdsInputSamples.mappedThreatProcessEvent()
            .withDevice(device)
            .withTriggeredAlertId(UUID.randomUUID().toString())
            .withAlertCreationTimestamp(TimeHelper.toTimestamp(firstEvent))
            .withAlertUpdateTimestamp(TimeHelper.toTimestamp(firstEvent))
            .backdatedTo(firstEvent)
            .withAlertAggregation(
                PSCRProto.CbAlertAggregation.newBuilder()
                    .setAggregateGuid(aggregateGuid)
                    .setIsAggregated(true))
            .withAlertThreatScore(5);

    publishBatch(builder1.build());

    PSCAlert.CbUaeAlert writtenAlert = waitForAlert(aggregateGuid, MAX_WAIT_SECONDS);
    assertEquals(5, writtenAlert.getAlert().getAlertThreatScore().getThreatScore());

    // write a second one with the same aggregateGuid with score 7
    var secondEventTimestamp = TimeHelper.toTimestamp(firstEvent.plus(Duration.ofMinutes(2)));
    var newAlertToWrite = writtenAlert.toBuilder();
    newAlertToWrite
        .getAlertBuilder()
        .setAlertUpdateTimestamp(secondEventTimestamp)
        .getAlertThreatScoreBuilder()
        .setThreatScore(7);

    newAlertToWrite.getEventBuilder().getHeaderBuilder().setDeviceTimestamp(secondEventTimestamp);
    publishBatch(UdsEventCreator.convertToUds(newAlertToWrite.build()));

    writtenAlert =
        waitForAlert(
            aggregateGuid,
            MAX_WAIT_SECONDS,
            alert -> alert.isPresent() && alert.get().getUpdateCount() == 1);
    assertEquals(5, writtenAlert.getAlert().getAlertUpdateDetails().getOldThreatScore());
    assertEquals(7, writtenAlert.getAlert().getAlertThreatScore().getThreatScore());

    // write a third one with the same aggregateGuid with score 9
    var thirdEventTimestamp = TimeHelper.toTimestamp(firstEvent.plus(Duration.ofMinutes(4)));
    newAlertToWrite = writtenAlert.toBuilder();
    newAlertToWrite
        .getAlertBuilder()
        .setAlertUpdateTimestamp(thirdEventTimestamp)
        .getAlertThreatScoreBuilder()
        .setThreatScore(9);

    newAlertToWrite.getEventBuilder().getHeaderBuilder().setDeviceTimestamp(thirdEventTimestamp);
    publishBatch(UdsEventCreator.convertToUds(newAlertToWrite.build()));

    writtenAlert =
        waitForAlert(
            aggregateGuid,
            MAX_WAIT_SECONDS,
            alert -> alert.isPresent() && alert.get().getUpdateCount() == 2);

    assertEquals(aggregateGuid, writtenAlert.getAlert().getAlertIdGuid());
    assertEquals(aggregateGuid, writtenAlert.getAlert().getAlertAggregation().getAggregateGuid());
    assertEquals(9, writtenAlert.getAlert().getAlertThreatScore().getThreatScore());
    assertEquals(7, writtenAlert.getAlert().getAlertUpdateDetails().getOldThreatScore());
  }

  @ParameterizedTest(name = "{0}: Test alerts updated with confidence fill update details")
  @MethodSource("allTestDevices")
  void testEndToEndAlertConfidenceChange(PSCRAugmentedEventBatch.CbDeviceMetadata.Builder device)
      throws Exception {
    String alertId = UUID.randomUUID().toString();

    var alertTimestamp = Instant.now().minus(Duration.ofDays(31));

    UdsInputSampleBuilder builder =
        UdsInputSamples.mappedThreatProcessEvent()
            .withDevice(device)
            .withTriggeredAlertId(alertId)
            .withAlertCreationTimestamp(TimeHelper.toTimestamp(alertTimestamp))
            .withAlertUpdateTimestamp(TimeHelper.toTimestamp(alertTimestamp))
            .backdatedTo(alertTimestamp)
            .withAlertThreatScore(5);

    var alertNew = builder.buildToMutableCbUaeAlert();
    alertNew
        .getAlertBuilder()
        .setAlertConfidence(ALERT_CONFIDENCE_LOW)
        .setAlertType(ALERT_TYPE_THREAT)
        .setAlertProducer("F745EEE6-706F-3DDD-8852-85D943D8F230")
        .getAlertThreatScoreBuilder()
        .setThreatSource(THREAT_SOURCE_DEFENSE_ANALYTICS);

    publishBatch(UdsEventCreator.convertToUds(alertNew.build()));
    PSCAlert.CbUaeAlert writtenAlert = waitForAlert(alertId, MAX_WAIT_SECONDS);
    assertEquals(ALERT_CONFIDENCE_LOW, writtenAlert.getAlert().getAlertConfidence());

    // write a second one with the same aggregateGuid with score 7
    var secondEventTimestamp = TimeHelper.toTimestamp(alertTimestamp.plus(Duration.ofMinutes(2)));
    var newAlertToWrite = writtenAlert.toBuilder();
    newAlertToWrite
        .getAlertBuilder()
        .setAlertUpdateTimestamp(secondEventTimestamp)
        .setAlertConfidence(ALERT_CONFIDENCE_MEDIUM)
        .setAlertProducer("F745EEE6-706F-3DDD-8852-85D943D8F230_ELEVATED");
    newAlertToWrite.getEventBuilder().getHeaderBuilder().setDeviceTimestamp(secondEventTimestamp);
    newAlertToWrite.setLastEventTimestamp(secondEventTimestamp);
    publishBatch(UdsEventCreator.convertToUds(newAlertToWrite.build()));
    writtenAlert =
        waitForAlert(
            alertId,
            MAX_WAIT_SECONDS,
            alert -> alert.isPresent() && alert.get().getUpdateCount() == 1);
    assertEquals(ALERT_CONFIDENCE_MEDIUM, writtenAlert.getAlert().getAlertConfidence());
    assertEquals(
        "F745EEE6-706F-3DDD-8852-85D943D8F230_ELEVATED",
        writtenAlert.getAlert().getAlertProducer());
  }

  @ParameterizedTest(name = "{0}: Test update count is tracked in upsert script")
  @MethodSource("allTestDevices")
  void testEndToEndAlertUpdateCount(PSCRAugmentedEventBatch.CbDeviceMetadata.Builder device)
      throws Exception {
    String aggregateGuid = UUID.randomUUID().toString();

    var alertTime = Instant.now().minus(Duration.ofDays(31));

    // write one alert
    // note: alert needs to have defense_analytics_primary_event: true for updates to work
    UdsInputSampleBuilder builder1 =
        UdsInputSamples.threatEventWithBlockedAnydesk()
            .withDevice(device)
            .withTriggeredAlertId(UUID.randomUUID().toString())
            .backdatedTo(alertTime)
            .withAlertUpdateTimestamp(TimeHelper.toTimestamp(alertTime))
            .withAlertAggregation(
                PSCRProto.CbAlertAggregation.newBuilder()
                    .setAggregateGuid(aggregateGuid)
                    .setIsAggregated(true));

    publishBatch(builder1.build());

    PSCAlert.CbUaeAlert writtenAlert = waitForAlert(aggregateGuid, MAX_WAIT_SECONDS);
    assertEquals(0, writtenAlert.getUpdateCount());

    // write a second one with the same aggregateGuid - update 1
    var secondEvent = alertTime.plus(Duration.ofMinutes(2));
    builder1
        .withAlertUpdateTimestamp(TimeHelper.toTimestamp(secondEvent))
        .backdatedTo(secondEvent)
        .withIngressTimestamp(TimeHelper.toTimestamp(secondEvent));
    // note: need to set the last event timestamp too, as that's what the aggregation logic is
    // checking
    publishBatch(
        UdsEventCreator.convertToUds(
            builder1
                .buildToMutableCbUaeAlert()
                .setLastEventTimestamp(TimeHelper.toTimestamp(secondEvent))
                .build()));

    waitForAlert(
        aggregateGuid,
        MAX_WAIT_SECONDS,
        alert -> alert.isPresent() && alert.get().getUpdateCount() == 1);

    // write a third one with the same aggregateGuid - update 2
    var thirdEvent = alertTime.plus(Duration.ofMinutes(5));
    builder1
        .withAlertUpdateTimestamp(TimeHelper.toTimestamp(thirdEvent))
        .backdatedTo(secondEvent)
        .withIngressTimestamp(TimeHelper.toTimestamp(thirdEvent));
    publishBatch(
        UdsEventCreator.convertToUds(
            builder1
                .buildToMutableCbUaeAlert()
                .setLastEventTimestamp(TimeHelper.toTimestamp(thirdEvent))
                .build()));

    writtenAlert =
        waitForAlert(
            aggregateGuid,
            MAX_WAIT_SECONDS,
            alert -> alert.isPresent() && alert.get().getUpdateCount() == 2);

    assertEquals(aggregateGuid, writtenAlert.getAlert().getAlertIdGuid());
    assertEquals(2, writtenAlert.getUpdateCount());
  }

  void publishBatch(PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg payload) throws Exception {
    testProducer.addRawRecord(
        config.getClient().getStreamName(), payload.toByteArray(), UUID.randomUUID().toString());

    if (testProducer instanceof KinesisKPLProducer kplProducer) {
      kplProducer.flush();
    }
    log.debug("publishBatch: Flushed event batch to kinesis");
  }

  Optional<Notes> getNotesById(String id) throws IOException {
    var index =
        monthlyFixedRetentionPartitionStrategy.getIndexForDocument("notes-alert", Instant.now());
    GetRequest getRequest =
        new GetRequest(index, id).fetchSourceContext(FetchSourceContext.FETCH_SOURCE);
    GetResponse response;
    try {
      response = notesClient().get(getRequest, RequestOptions.DEFAULT);
    } catch (OpenSearchStatusException e) {
      if (e.status().getStatus() == 404) {
        log.debug("getNotesById: index does not exist yet", e);
      } else {
        log.warn("getNotesById: Get exception from es", e);
      }
      return Optional.empty();
    } catch (Exception e) {
      log.warn("getNotesById: Get exception from es", e);
      return Optional.empty();
    }

    if (!response.isExists()) {
      log.debug("getNotesById: note {} does not exist in {}", id, index);
      return Optional.empty();
    }

    log.debug("getNotesById: found note {}", id);
    return Optional.of(objectMapper.readValue(response.getSourceAsString(), Notes.class));
  }

  private PSCAlert.CbUaeAlert waitForAlert(String alertId, int seconds) {
    return waitForAlert(alertId, seconds, Optional::isPresent);
  }

  private PSCAlert.CbUaeAlert waitForAlert(
      String alertId, int seconds, Predicate<Optional<PSCAlert.CbUaeAlert>> condition) {
    Optional<PSCAlert.CbUaeAlert> alert =
        await()
            .atMost(Duration.ofSeconds(seconds))
            .pollInterval(Duration.ofSeconds(2))
            .until(() -> getAlertById(alertId), condition);
    assertTrue(alert.isPresent());
    return alert.get();
  }

  Optional<PSCAlert.CbUaeAlert> getAlertById(String id) throws IOException {
    return alertEsClient.findAlertById(id);
  }
}
