package com.vmware.psc.alerting.persister;

import static com.google.common.base.Strings.isNullOrEmpty;

import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.auth.STSAssumeRoleSessionCredentialsProvider;
import com.amazonaws.regions.AwsRegionProvider;
import com.amazonaws.services.kinesis.producer.KinesisProducerConfiguration;
import com.amazonaws.services.securitytoken.AWSSecurityTokenServiceClientBuilder;
import com.carbonblack.psc.backoff.ExponentialBackoff;
import com.carbonblack.psc.backoff.retrylocker.IAbortRetryLocker;
import com.carbonblack.psc.backoff.sleep.ISleeper;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Strings;
import com.vmware.psc.alerting.persister.kinesis.client.KinesisApplication;
import com.vmware.psc.alerting.persister.kinesis.client.health.KinesisApplicationHealthIndicator;
import com.vmware.psc.alerting.persister.kinesis.producer.IKinesisProducer;
import com.vmware.psc.alerting.persister.kinesis.producer.KinesisBasicProducer;
import com.vmware.psc.alerting.persister.kinesis.producer.KinesisKPLProducer;
import com.vmware.psc.alerting.persister.properties.KinesisConfig;
import com.vmware.psc.alerting.persister.util.RecordProcessingSupport;
import io.micrometer.core.instrument.MeterRegistry;
import java.util.Optional;
import java.util.UUID;
import lombok.extern.slf4j.Slf4j;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.context.annotation.Profile;
import software.amazon.awssdk.services.cloudwatch.CloudWatchAsyncClient;
import software.amazon.awssdk.services.dynamodb.DynamoDbAsyncClient;
import software.amazon.awssdk.services.kinesis.KinesisAsyncClient;

@Slf4j
@Configuration
@Profile({"aws", "localstack"})
public class PersisterContext {
  private final MeterRegistry registry;
  private final KinesisAsyncClient kinesis;
  private final DynamoDbAsyncClient dynamo;
  private final CloudWatchAsyncClient cloudwatch;
  private final KinesisConfig config;
  private final AwsRegionProvider awsRegionProvider;

  public PersisterContext(
      MeterRegistry registry,
      KinesisAsyncClient kinesis,
      DynamoDbAsyncClient dynamo,
      CloudWatchAsyncClient cloudwatch,
      KinesisConfig config,
      AwsRegionProvider awsRegionProvider) {
    this.registry = registry;
    this.kinesis = kinesis;
    this.dynamo = dynamo;
    this.cloudwatch = cloudwatch;
    this.config = config;
    this.awsRegionProvider = awsRegionProvider;
  }

  @Bean
  ExponentialBackoff processBackoff(ISleeper sleeper, IAbortRetryLocker locker) {
    return new ExponentialBackoff(config.getClient().getBackoff().getProcess(), sleeper, locker);
  }

  @Bean
  ExponentialBackoff checkpointBackoff(ISleeper sleeper, IAbortRetryLocker locker) {
    return new ExponentialBackoff(config.getClient().getBackoff().getCheckpoint(), sleeper, locker);
  }

  @Bean
  @ConditionalOnProperty(value = "alert-persister.kinesis.producer.basic.enabled")
  IKinesisProducer basicProducer(
      ISleeper sleeper, IAbortRetryLocker locker, ObjectMapper objectMapper) {
    // TODO configure to use optional port, assume role
    return new KinesisBasicProducer(
        kinesis, config.getProducer(), sleeper, locker, registry, objectMapper);
  }

  @Bean
  @ConditionalOnProperty(value = "alert-persister.kinesis.producer.kpl.enabled")
  @Primary
  IKinesisProducer kplProducer(
      KinesisProducerConfiguration producerConfig,
      ObjectMapper objectMapper,
      AWSCredentialsProvider awsCredentialsProvider) {

    // assume role if the role is provided
    getStreamAssumeRoleProvider(awsCredentialsProvider)
        .ifPresent(producerConfig::setCredentialsProvider);

    getKinesisEndpoint().ifPresent(producerConfig::setKinesisEndpoint);

    // optionally override port (in local/testing env)
    config.getPort().ifPresent(producerConfig::setKinesisPort);

    return new KinesisKPLProducer(producerConfig, config.getProducer(), registry, objectMapper);
  }

  @Bean
  @ConditionalOnProperty(value = "alert-persister.kinesis.client.enabled", matchIfMissing = true)
  KinesisApplication persisterApplication(
      PersisterRecordProcessorFactory processorFactory,
      KinesisApplicationHealthIndicator healthIndicator) {
    return new KinesisApplication(
        config.getClient(), kinesis, dynamo, cloudwatch, processorFactory, healthIndicator);
  }

  private Optional<AWSCredentialsProvider> getStreamAssumeRoleProvider(
      AWSCredentialsProvider awsCredentialsProvider) {
    if (config.getOutputStream() == null
        || Strings.isNullOrEmpty(config.getOutputStream().getAwsRole())) {
      return Optional.empty();
    }

    String streamAccessRole = config.getOutputStream().getAwsRole();

    log.debug(
        "kcl stream name: {}, stream role: {}",
        config.getOutputStream().getStreamName(),
        config.getOutputStream().getAwsRole());

    if (!isNullOrEmpty(streamAccessRole)) {
      log.debug("Assuming kcl stream role: {}", streamAccessRole);

      return Optional.of(
          new STSAssumeRoleSessionCredentialsProvider.Builder(
                  streamAccessRole, UUID.randomUUID().toString())
              .withStsClient(
                  AWSSecurityTokenServiceClientBuilder.standard()
                      .withCredentials(awsCredentialsProvider)
                      .withRegion(awsRegionProvider.getRegion())
                      .build())
              .build());
    }

    return Optional.empty();
  }

  private Optional<String> getKinesisEndpoint() {
    return Optional.ofNullable(config.getEndpoint());
  }

  @Bean
  public RecordProcessingSupport recordProcessingSupport(
      final KinesisConfig config,
      final ExponentialBackoff processBackoff,
      final ExponentialBackoff checkpointBackoff) {
    return new RecordProcessingSupport(
        registry, config.getClient().getRecordProcessor(), processBackoff, checkpointBackoff);
  }
}
