package com.vmware.psc.alerting.persister;

import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlertClassification.CbAlertDeterminationCodeType.TP;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlertUpdateDetails.AlertUpdateType.NOTE_ADDED;
import static com.carbonblack.protocols.psc.response.PSCRProto.ThreatScore.CbThreatSource.THREAT_SOURCE_DEFENSE_ANALYTICS;
import static com.vmware.psc.alerting.persister.es.notes.NotesWriter.buildNotesId;
import static com.vmware.psc.alerting.persister.service.AggregationService.updateTimestamp;
import static com.vmware.psc.alerting.persister.util.CbUaeAlertUtil.getAlertId;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRAugmentedEventBatch;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.carbonblack.psc.orgkey.exception.IllegalOrgKeyException;
import com.google.common.base.Strings;
import com.google.protobuf.InvalidProtocolBufferException;
import com.google.protobuf.util.JsonFormat;
import com.vmware.psc.alerting.data.uae.UaeAlertCreator;
import com.vmware.psc.alerting.exception.NonRetryableDecompressionException;
import com.vmware.psc.alerting.persister.cache.properties.CacheAlertAggProperties;
import com.vmware.psc.alerting.persister.cache.service.CacheService;
import com.vmware.psc.alerting.persister.enums.AggregationUpdateType;
import com.vmware.psc.alerting.persister.es.AlertWriter;
import com.vmware.psc.alerting.persister.es.notes.NotesWriter;
import com.vmware.psc.alerting.persister.exception.RetryableOrgThreatException;
import com.vmware.psc.alerting.persister.kinesis.client.RecordProcessorImpl;
import com.vmware.psc.alerting.persister.kinesis.producer.ProducerSupport;
import com.vmware.psc.alerting.persister.metrics.PerAlertTypeCounter;
import com.vmware.psc.alerting.persister.model.notes.Notes;
import com.vmware.psc.alerting.persister.orgthreat.OrgThreat;
import com.vmware.psc.alerting.persister.orgthreat.OrgThreatClient;
import com.vmware.psc.alerting.persister.orgthreat.ThreatWorkflow;
import com.vmware.psc.alerting.persister.orgthreat.WorkflowState;
import com.vmware.psc.alerting.persister.service.AggregationService;
import com.vmware.psc.alerting.persister.util.RecordProcessingSupport;
import java.io.IOException;
import java.util.*;
import java.util.function.Predicate;
import lombok.extern.slf4j.Slf4j;

@Slf4j
public class PersisterRecordProcessor
    extends RecordProcessorImpl<PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg> {

  public static final String ALERTS_TO_FLUSH_SIZE = "alerts_to_flush_size";
  public static final String CLOSED = "CLOSED";
  private static final Predicate<PSCAlert.CbUaeAlert.Builder> PREDICATE_CLOSED_WORK_FLOW =
      alertBuilder ->
          Optional.ofNullable(alertBuilder)
              .map(PSCAlert.CbUaeAlert.Builder::getAlert)
              .filter(PSCRProto.CbAlert::hasUserStateDetails)
              .map(PSCRProto.CbAlert::getUserStateDetails)
              .map(PSCRProto.CbAlertStateDetails::getWorkflowState)
              .map(PSCRProto.CbAlertWorkflowState::getStatus)
              .map(status -> status.equalsIgnoreCase(CLOSED))
              .orElse(Boolean.FALSE);

  private final Map<String, List<PSCAlert.CbUaeAlert.Builder>> alertsMap;
  private final Map<String, Notes> notesToFlush;
  private final AlertWriter alertWriter;
  private final NotesWriter notesWriter;
  private final OrgThreatClient orgThreatClient;
  private final CacheService<String, PSCAlert.CbUaeAlert> alertCacheService;

  private final CacheAlertAggProperties cacheAlertAggProperties;

  private final ProducerSupport producerSupport;

  private final PerAlertTypeCounter alertsReceived;
  private final PerAlertTypeCounter alertsProcessed;
  private final PerAlertTypeCounter alertsSkipped;

  public PersisterRecordProcessor(
      final RecordProcessingSupport recordProcessingSupport,
      AlertWriter alertWriter,
      NotesWriter notesWriter,
      OrgThreatClient orgThreatClient,
      final CacheService<String, PSCAlert.CbUaeAlert> alertCacheService,
      final CacheAlertAggProperties cacheAlertAggProperties,
      final ProducerSupport producerSupport) {
    super(
        recordProcessingSupport.registry(),
        recordProcessingSupport.processorProperties(),
        recordProcessingSupport.processBackoff(),
        recordProcessingSupport.checkpointBackoff());
    this.alertWriter = alertWriter;
    this.notesWriter = notesWriter;
    this.orgThreatClient = orgThreatClient;
    this.alertCacheService = alertCacheService;
    this.cacheAlertAggProperties = cacheAlertAggProperties;
    notesToFlush = new HashMap<>();
    alertsMap = new HashMap<>();
    recordProcessingSupport
        .registry()
        .gauge(
            ALERTS_TO_FLUSH_SIZE,
            Collections.emptyList(),
            alertsMap,
            map -> map.values().stream().mapToDouble(List::size).sum());
    this.producerSupport = producerSupport;
    this.alertsReceived =
        new PerAlertTypeCounter(
            recordProcessingSupport.registry(), "kinesis.input.alerts_received");
    this.alertsSkipped =
        new PerAlertTypeCounter(recordProcessingSupport.registry(), "kinesis.input.alerts_skipped");
    this.alertsProcessed =
        new PerAlertTypeCounter(
            recordProcessingSupport.registry(), "kinesis.input.alerts_processed");
  }

  @Override
  protected PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg deserialize(byte[] serializedData)
      throws IOException {
    return PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg.parseFrom(serializedData);
  }

  /**
   * Process each event batch message to generate alerts
   *
   * @throws RetryableOrgThreatException - Error looking up orgthreat data from dynamo. Can retry.
   */
  @Override
  protected boolean process(PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch)
      throws RetryableOrgThreatException {
    List<PSCAlert.CbUaeAlert.Builder> alerts;
    try {
      alerts = UaeAlertCreator.convertAlertsAsMutable(batch);
    } catch (NonRetryableDecompressionException e) {
      // Error parsing the record. There's no way we're going to be able to fix this so just abort
      // processing.
      log.debug(
          "process: Cannot parse record from device={}, dropping it", batch.getDeviceMetadata());
      return true;
    }

    processAlerts(alerts);
    return true;
  }

  private PSCRProto.CbAlert.AlertType getAlertUpdateType(PSCAlert.CbUaeAlert.Builder alert) {
    return Optional.ofNullable(alert)
        .map(PSCAlert.CbUaeAlert.Builder::getAlert)
        .map(PSCRProto.CbAlert::getAlertType)
        .orElse(PSCRProto.CbAlert.AlertType.UNRECOGNIZED);
  }

  /**
   * Process the alerts.
   *
   * @param alerts List of CbUaeAlert
   * @throws RetryableOrgThreatException retry with exponential backoff.
   */
  private void processAlerts(List<PSCAlert.CbUaeAlert.Builder> alerts)
      throws RetryableOrgThreatException {
    for (var alert : alerts) {
      PSCRProto.CbAlert.AlertType alertUpdateType = getAlertUpdateType(alert);
      alertsReceived.increment(alertUpdateType);
      try {
        if (shouldSkipAlertRecord(alert)) {
          alertsSkipped.increment(alertUpdateType);
          continue;
        }
        buildNotesForFlush(alert);
        if (cacheAlertAggProperties.isEnabled()) {
          aggregateAlertWithCache(alert);
        } else {
          updateAlertFromOrgThreat(alert);
          setAlertIdToAggregateIdIfExists(alert);
          addDataToFlush(AggregationUpdateType.NEW, alert);
        }
        alertsProcessed.increment(alertUpdateType);
      } catch (IllegalOrgKeyException ex) {
        try {
          // PSCR smoke tests use org keys that start with an underscore
          // We can safely ignore them
          if (!ex.getOrgKey().startsWith("_")) {
            log.error("Invalid org key alert: {}", JsonFormat.printer().print(alert));
            throw ex;
          }
        } catch (InvalidProtocolBufferException e) {
          log.error("Error serializing alert for troubleshooting", e);
        }
      }
    }
  }

  /**
   * Skipping the alert record if more than one threat details present.
   *
   * @param alert CbUaeAlert.
   * @return boolean.
   */
  private static boolean shouldSkipAlertRecord(final PSCAlert.CbUaeAlert.Builder alert) {

    boolean hasMultipleThreatIndicators =
        Optional.ofNullable(alert)
            .map(PSCAlert.CbUaeAlert.Builder::getAlert)
            .map(PSCRProto.CbAlert::getThreatIndicatorsList)
            .filter(list -> list.size() > 1)
            .isPresent();
    boolean isJavaAnalyticsAlert =
        Optional.ofNullable(alert)
            .map(PSCAlert.CbUaeAlert.Builder::getAlert)
            .map(PSCRProto.CbAlert::getAlertThreatScore)
            .map(PSCRProto.ThreatScore::getThreatSource)
            .filter(THREAT_SOURCE_DEFENSE_ANALYTICS::equals)
            .isPresent();

    if (hasMultipleThreatIndicators && !isJavaAnalyticsAlert) {
      log.warn(
          "Skipping the alert, more than one Threat details present in alert id {}",
          getAlertId(alert));
      return true;
    }
    return false;
  }

  /**
   * aggregateAlertWithCache.
   *
   * @param incomingAlert CbUaeAlert.
   * @throws RetryableOrgThreatException exception.
   */
  private void aggregateAlertWithCache(final PSCAlert.CbUaeAlert.Builder incomingAlert)
      throws RetryableOrgThreatException {
    final String alertId = getAlertId(incomingAlert);
    final PSCAlert.CbUaeAlert cachedAlert = alertCacheService.get(alertId);
    final AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlert, incomingAlert.build(), cacheAlertAggProperties.getFlushInterval());
    updateAlertFromOrgThreat(incomingAlert);
    log.debug("Alert Update Type {} and alert id {}", updateType, alertId);

    // For the no change aggregation type, incoming alert published to the UDS stream
    // and update timestamps in cached alert.
    if (updateType == AggregationUpdateType.NO_CHANGE) {
      alertCacheService.put(alertId, updateTimestamp(cachedAlert, incomingAlert).build());
      producerSupport.publishUaeAlert(incomingAlert.build(), getAlertId(incomingAlert));
    } else {
      // When the update type is not NO_CHANGE
      // we will be writing the alert to OS and handle publishing back to UDS after receiving the
      // response from OS.
      final PSCAlert.CbUaeAlert.Builder aggregatedAlert =
          AggregationService.updateIncomingAlertFromCache(incomingAlert, cachedAlert, updateType);
      setAlertIdToAggregateIdIfExists(aggregatedAlert);
      alertCacheService.put(alertId, aggregatedAlert.build());
      addDataToFlush(updateType, aggregatedAlert);
    }
  }

  /**
   * Add Data to Flush.
   *
   * @param updateType AggregationUpdateType.
   * @param alert PSCAlert.CbUaeAlert.Builder.
   */
  private void addDataToFlush(
      final AggregationUpdateType updateType, final PSCAlert.CbUaeAlert.Builder alert) {
    alertsMap.computeIfAbsent(updateType.name(), key -> new ArrayList<>());
    alertsMap.computeIfPresent(
        updateType.name(),
        (key, alertsToPersist) -> {
          alertsToPersist.add(alert);
          return alertsToPersist;
        });
  }

  private void buildNotesForFlush(PSCAlert.CbUaeAlert.Builder alert) {
    PSCRProto.CbAlertWorkflowState workflowState =
        alert.getAlert().getMdrStateDetails().getWorkflowState();
    String author = workflowState.getChangedBy().getName();
    String ruleId = workflowState.getChangedBy().getSuppressionRuleId();
    String orgKey = alert.getDeviceMetadata().getOrgKey();
    // For public suppression note
    // if alert determination is TP
    // 2 way comm will be started (locked: false)
    if (workflowState.hasExternalNote()) {
      var externalNote = workflowState.getExternalNote();
      PSCRProto.CbAlertClassification.CbAlertDeterminationCodeType alertDeterminationCode =
          alert.getAlert().getMdrStateDetails().getClassification().getAlertDeterminationCode();
      Notes notes =
          createSuppressionNotes(
              orgKey,
              alert.getAlert().getAlertIdGuid(),
              author,
              ruleId,
              externalNote,
              false,
              alertDeterminationCode != TP);
      log.debug("buildNotesForFlush: Created external notes={}", notes);
      notesToFlush.put(notes.getId(), notes);
      // The actual update will happen in painless script in es-config
      alert.getAlertBuilder().getAlertUpdateDetailsBuilder().addUpdateTypes(NOTE_ADDED);
      // When adding public note, make alert visible to customers
      if (alert.getAlert().getAlertConfidence()
          == PSCRProto.CbAlert.CbAlertConfidence.ALERT_CONFIDENCE_SILENT) {
        alert
            .getAlertBuilder()
            .setAlertConfidence(PSCRProto.CbAlert.CbAlertConfidence.ALERT_CONFIDENCE_MEDIUM);
      }
    }

    if (workflowState.hasInternalNote()) {
      // For internal suppression note
      // locked will remain to true
      var internalNote = workflowState.getInternalNote();
      Notes notes =
          createSuppressionNotes(
              orgKey, alert.getAlert().getAlertIdGuid(), author, ruleId, internalNote, true, true);
      log.debug("buildNotesForFlush: Created internal notes={}", notes);
      notesToFlush.put(notes.getId(), notes);
    }
  }

  private Notes createSuppressionNotes(
      String orgKey,
      String alertId,
      String author,
      String ruleId,
      PSCRProto.CbAlertSuppressionNote suppressionNote,
      boolean isInternal,
      boolean locked) {
    String content = suppressionNote.getContent();
    if (content.length() > 20000) {
      log.warn(
          "createSuppressionNotes: Suppression note content has more than 20000 chars. orgKey={}, ruleId={}, note={}",
          orgKey,
          ruleId,
          suppressionNote);
    }
    long notesVersion = suppressionNote.getVersionId();
    String id = buildNotesId(alertId, ruleId, notesVersion, isInternal);
    return Notes.builder()
        .alertId(alertId)
        .note(content)
        .orgKey(orgKey)
        .author(author)
        .internal(isInternal)
        .locked(locked)
        .source("SUPPRESSION")
        .suppressionRuleId(ruleId)
        .suppressionNoteVersion(notesVersion)
        .id(id)
        .build();
  }

  private void updateAlertFromOrgThreat(PSCAlert.CbUaeAlert.Builder alert)
      throws RetryableOrgThreatException {
    String orgKey = alert.getDeviceMetadata().getOrgKey();
    String threatId = alert.getAlert().getAlertHash();
    OrgThreat threat = orgThreatClient.getById(orgKey, threatId);

    // Create default object of mdrstatedetails, if alert is mdr enabled and state details not
    // present
    if (alert.getAlert().getMdrAlert() && !alert.getAlertBuilder().hasMdrStateDetails()) {
      alert.getAlertBuilder().setMdrStateDetails(buildDefaultMdrState());
    }

    if (threat == null) {
      log.trace("OrgThreat(orgkey={}, threat_id={}) not found", orgKey, threatId);
      alert.getAlertBuilder().setUserStateDetails(buildDefaultState());
      return;
    }

    if (threat.getTags() != null) {
      // Tags *theoretically* can never be null but there are some test orgs where they've slipped
      // through validation. This condition helps weed out any bad tags.
      alert.addAllInvestigationTags(
          threat.getTags().stream().filter(Predicate.not(Strings::isNullOrEmpty)).toList());
    }
    alert.setThreatNotesPresent(threat.isNotePresent());

    if (hasExistingWorkflow(alert)) {
      return;
    }

    ThreatWorkflow workflow = threat.getWorkFlow();
    if (workflow == null) {
      // No existing workflow
      log.debug(
          "Ignoring OrgThreat(orgkey={}, threat_id={}) without workflow set", orgKey, threatId);
      alert.getAlertBuilder().setUserStateDetails(buildDefaultState());
    } else if (workflow.getState() != WorkflowState.DISMISSED) {
      // Existing workflow, but not marked as dismissed (seems unlikely but still possible)
      log.debug(
          "Ignoring OrgThreat(orgkey={}, threat_id={}, workflow={}) because state is not DISMISSED",
          orgKey,
          threatId,
          threat.getWorkFlow());
      alert.getAlertBuilder().setUserStateDetails(buildDefaultState());
    } else {
      // Existing workflow: marked this as dismissed
      alert
          .getAlertBuilder()
          .setUserStateDetails(
              PSCRProto.CbAlertStateDetails.newBuilder()
                  .setWorkflowState(buildStateForWorkflow(workflow))
                  .setClassification(buildClassificationForWorkflow(workflow))
                  .build());
    }
  }

  /**
   * Check the work flow is closed.
   *
   * @param alert CbUaeAlert.
   * @return boolean.
   */
  private static boolean hasExistingWorkflow(PSCAlert.CbUaeAlert.Builder alert) {
    return PREDICATE_CLOSED_WORK_FLOW.test(alert);
  }

  private static void setAlertIdToAggregateIdIfExists(PSCAlert.CbUaeAlert.Builder alert) {
    // set the alert-id guid to the alert-aggregate guid if the alert-aggregate guid exists
    if (alert.getAlert().hasAlertAggregation()) {
      alert
          .getAlertBuilder()
          .setAlertIdGuid(alert.getAlert().getAlertAggregation().getAggregateGuid());
    }
  }

  /**
   * Build a default Alert State Details
   *
   * <p>Default values:
   *
   * <ul>
   *   <li><strong>status</strong>: OPEN
   *   <li><strong>changed by</strong>: from the workflow {@link ThreatWorkflow#getChangedBy()}
   * </ul>
   */
  private PSCRProto.CbAlertStateDetails buildDefaultState() {
    return PSCRProto.CbAlertStateDetails.newBuilder()
        .setWorkflowState(
            PSCRProto.CbAlertWorkflowState.newBuilder()
                .setStatus("OPEN")
                .setChangedBy(
                    PSCRProto.CbAlertOperator.newBuilder()
                        .setType(PSCRProto.CbAlertOperator.CbAlertOperatorType.SYSTEM)
                        .setName("ALERT_CREATION")
                        .build())
                .build())
        .setClassification(
            PSCRProto.CbAlertClassification.newBuilder()
                .setAlertClassification(
                    PSCRProto.CbAlertClassification.CbAlertClassificationType
                        .ALERT_CLASSIFICATION_UNKNOWN)
                .setChangedBy(
                    PSCRProto.CbAlertOperator.newBuilder()
                        .setType(PSCRProto.CbAlertOperator.CbAlertOperatorType.SYSTEM)
                        .setName("ALERT_CREATION")
                        .build())
                .build())
        .build();
  }

  /**
   * Build a default Mdr Alert State Details
   *
   * <p>Default values:
   *
   * <ul>
   *   <li><strong>status</strong>: UNCLAIMED
   *   <li><strong>changed by</strong>: SYSTEM
   * </ul>
   */
  private PSCRProto.CbAlertStateDetails buildDefaultMdrState() {
    return PSCRProto.CbAlertStateDetails.newBuilder()
        .setWorkflowState(
            PSCRProto.CbAlertWorkflowState.newBuilder()
                .setStatus("UNCLAIMED")
                .setChangedBy(
                    PSCRProto.CbAlertOperator.newBuilder()
                        .setType(PSCRProto.CbAlertOperator.CbAlertOperatorType.SYSTEM)
                        .setName("ALERT_CREATION")
                        .build())
                .build())
        .setClassification(
            PSCRProto.CbAlertClassification.newBuilder()
                .setAlertClassification(
                    PSCRProto.CbAlertClassification.CbAlertClassificationType
                        .ALERT_CLASSIFICATION_UNKNOWN)
                .setChangedBy(
                    PSCRProto.CbAlertOperator.newBuilder()
                        .setType(PSCRProto.CbAlertOperator.CbAlertOperatorType.SYSTEM)
                        .setName("ALERT_CREATION")
                        .build())
                .build())
        .build();
  }

  /**
   * Build an Alert State based on an existing dismissed workflow
   *
   * <p>Default values:
   *
   * <ul>
   *   <li><strong>status</strong>: closed
   *   <li><strong>comment</strong>: from the workflow {@link ThreatWorkflow#getComment()}
   *   <li><strong>changed by</strong>: system
   * </ul>
   */
  private PSCRProto.CbAlertWorkflowState buildStateForWorkflow(ThreatWorkflow workflow) {
    return PSCRProto.CbAlertWorkflowState.newBuilder()
        .setComment(ensureNonNull(workflow.getComment()))
        .setStatus(CLOSED)
        .setChangedBy(
            PSCRProto.CbAlertOperator.newBuilder()
                .setType(PSCRProto.CbAlertOperator.CbAlertOperatorType.SYSTEM)
                .setName("AUTO_DISMISSAL")
                .build())
        .setCloseCode(ensureNonNull(workflow.getRemediation()))
        .build();
  }

  /**
   * Build an Alert Classification based on an existing dismissed workflow
   *
   * <p>Default values:
   *
   * <ul>
   *   <li><strong>classification</strong>: false positive
   *   <li><strong>justification</strong>:from the workflow {@link ThreatWorkflow#getRemediation()}
   *   <li><strong>changed by</strong>: from the workflow {@link ThreatWorkflow#getChangedBy()}
   * </ul>
   */
  private PSCRProto.CbAlertClassification buildClassificationForWorkflow(ThreatWorkflow workflow) {
    PSCRProto.CbAlertClassification.Builder classification =
        PSCRProto.CbAlertClassification.newBuilder()
            .setAlertClassification(
                PSCRProto.CbAlertClassification.CbAlertClassificationType.FALSE_POSITIVE);

    if (workflow.getChangedBy() != null) {
      classification.setChangedBy(
          PSCRProto.CbAlertOperator.newBuilder()
              .setType(PSCRProto.CbAlertOperator.CbAlertOperatorType.USER)
              .setName(workflow.getChangedBy())
              .build());
    } else {
      classification.setChangedBy(
          PSCRProto.CbAlertOperator.newBuilder()
              .setType(PSCRProto.CbAlertOperator.CbAlertOperatorType.SYSTEM)
              .setName("AUTO_DISMISSAL")
              .build());
    }

    return classification.build();
  }

  private static String ensureNonNull(String str) {
    return str == null ? "" : str;
  }

  @Override
  protected void beforeCheckpoint() throws Exception {
    if (alertsMap.isEmpty()) {
      log.debug("beforeCheckpoint: no alerts to write");
    }

    alertsMap.forEach(
        (key, value) -> {
          log.debug("beforeCheckpoint: found {} alert(s) of type {}", value.size(), key);
          try {
            alertWriter.writeAndClearAlerts(value, AggregationUpdateType.valueOf(key));
          } catch (IOException e) {
            log.warn("error writing alerts to ES", e);
            throw new IllegalStateException(e);
          }
        });

    try {
      if (!notesToFlush.isEmpty()) {
        notesWriter.writeNotesBatch(new ArrayList<>(notesToFlush.values()));
      }
    } catch (IOException e) {
      log.warn("error writing alerts to ES", e);
      throw e;
    }
  }
}
