package com.vmware.psc.alerting.persister.aws.localstack;

import static software.amazon.awssdk.core.SdkSystemSetting.CBOR_ENABLED;

import com.amazonaws.SdkClientException;
import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.auth.AWSStaticCredentialsProvider;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.regions.AwsRegionProvider;
import com.amazonaws.regions.AwsRegionProviderChain;
import com.amazonaws.services.kinesis.producer.KinesisProducerConfiguration;
import com.vmware.psc.alerting.persister.properties.KinesisConfig;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.UUID;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Profile;
import software.amazon.awssdk.auth.credentials.AwsBasicCredentials;
import software.amazon.awssdk.auth.credentials.AwsCredentials;
import software.amazon.awssdk.auth.credentials.AwsCredentialsProvider;
import software.amazon.awssdk.auth.credentials.StaticCredentialsProvider;
import software.amazon.awssdk.core.internal.http.loader.DefaultSdkAsyncHttpClientBuilder;
import software.amazon.awssdk.http.SdkHttpConfigurationOption;
import software.amazon.awssdk.http.async.SdkAsyncHttpClient;
import software.amazon.awssdk.regions.Region;
import software.amazon.awssdk.services.cloudwatch.CloudWatchAsyncClient;
import software.amazon.awssdk.services.dynamodb.DynamoDbAsyncClient;
import software.amazon.awssdk.services.kinesis.KinesisAsyncClient;
import software.amazon.awssdk.utils.AttributeMap;

@Configuration
@Profile("localstack")
public class LocalstackAwsConfig {

  private static final String LOCALSTACK_HOST = "mockaws";
  private static final long LOCALSTACK_PORT = 4566;
  private static final String LOCALSTACK_ACCESS_KEY = "dummy";
  private static final String LOCALSTACK_SECRET_KEY = UUID.randomUUID().toString();

  @Bean
  public AWSCredentialsProvider localstackCredentials() {
    System.setProperty("com.amazonaws.sdk.disableCertChecking", "true");
    System.setProperty("com.amazonaws.sdk.disableCbor", "true");
    BasicAWSCredentials awsCredentials =
        new BasicAWSCredentials(LOCALSTACK_ACCESS_KEY, LOCALSTACK_SECRET_KEY);
    return new AWSStaticCredentialsProvider(awsCredentials);
  }

  @Bean
  public AwsCredentialsProvider localstackCredentialsSdkV2() {
    System.setProperty(CBOR_ENABLED.property(), "false");
    AwsCredentials awsCredentials =
        AwsBasicCredentials.create(LOCALSTACK_ACCESS_KEY, LOCALSTACK_SECRET_KEY);
    return StaticCredentialsProvider.create(awsCredentials);
  }

  @Bean
  AwsRegionProvider localstackRegionProvider() {
    return new AwsRegionProviderChain(
        new AwsRegionProvider() {

          @Override
          public String getRegion() throws SdkClientException {
            return "us-east-1";
          }
        });
  }

  @Bean
  SdkAsyncHttpClient awsSdkAsyncHttpClient() {
    AttributeMap attributeMap =
        AttributeMap.builder().put(SdkHttpConfigurationOption.TRUST_ALL_CERTIFICATES, true).build();
    return new DefaultSdkAsyncHttpClientBuilder().buildWithDefaults(attributeMap);
  }

  @Bean
  public KinesisAsyncClient kinesisAsyncClient(
      AwsCredentialsProvider awsCredentialsProvider,
      AwsRegionProvider regionProvider,
      SdkAsyncHttpClient httpClient,
      KinesisConfig config)
      throws URISyntaxException {
    return KinesisAsyncClient.builder()
        .credentialsProvider(awsCredentialsProvider)
        .region(Region.of(regionProvider.getRegion()))
        .httpClient(httpClient)
        .endpointOverride(new URI(getKinesisEndpointURI(config)))
        .build();
  }

  @Bean
  public DynamoDbAsyncClient buildAmazonDynamoDB(
      AwsCredentialsProvider awsCredentialsProvider,
      AwsRegionProvider regionProvider,
      SdkAsyncHttpClient httpClient,
      KinesisConfig config)
      throws URISyntaxException {
    return DynamoDbAsyncClient.builder()
        .credentialsProvider(awsCredentialsProvider)
        .region(Region.of(regionProvider.getRegion()))
        .httpClient(httpClient)
        .endpointOverride(new URI(getKinesisEndpointURI(config)))
        .build();
  }

  @Bean
  public CloudWatchAsyncClient cloudWatch(
      AwsCredentialsProvider awsCredentialsProvider,
      AwsRegionProvider regionProvider,
      SdkAsyncHttpClient httpClient,
      KinesisConfig config)
      throws URISyntaxException {
    return CloudWatchAsyncClient.builder()
        .credentialsProvider(awsCredentialsProvider)
        .region(Region.of(regionProvider.getRegion()))
        .httpClient(httpClient)
        .endpointOverride(new URI(getKinesisEndpointURI(config)))
        .build();
  }

  @Bean
  public KinesisProducerConfiguration kinesisProducerConfiguration(
      AWSCredentialsProvider awsCredentialsProvider, AwsRegionProvider regionProvider) {
    return new KinesisProducerConfiguration()
        .setCredentialsProvider(awsCredentialsProvider)
        .setRegion(regionProvider.getRegion())
        .setKinesisEndpoint(LOCALSTACK_HOST)
        .setKinesisPort(LOCALSTACK_PORT)
        .setStsEndpoint(LOCALSTACK_HOST)
        .setStsPort(LOCALSTACK_PORT)
        .setCloudwatchEndpoint(LOCALSTACK_HOST)
        .setCloudwatchPort(LOCALSTACK_PORT)
        .setVerifyCertificate(false);
  }

  private static String getKinesisEndpointURI(KinesisConfig config) {
    var kinesisEndpointURI = new StringBuilder();
    kinesisEndpointURI.append("https://");
    kinesisEndpointURI.append(config.getEndpoint());

    var port = config.getPort();
    if (port.isPresent()) {
      kinesisEndpointURI.append(":");
      kinesisEndpointURI.append(port.get());
    }

    return kinesisEndpointURI.toString();
  }
}
