package com.vmware.psc.alerting.persister.aws.localstack;

import com.vmware.psc.alerting.persister.orgthreat.OrgThreatClient;
import com.vmware.psc.alerting.persister.properties.KinesisConfig;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Component;
import software.amazon.awssdk.services.dynamodb.DynamoDbAsyncClient;
import software.amazon.awssdk.services.dynamodb.model.*;
import software.amazon.awssdk.services.kinesis.KinesisAsyncClient;
import software.amazon.awssdk.services.kinesis.model.CreateStreamRequest;
import software.amazon.awssdk.services.kinesis.model.DeleteStreamRequest;
import software.amazon.awssdk.services.kinesis.model.DescribeStreamRequest;

@Component
@Profile("localstack")
public class LocalstackInfra {

  private final KinesisAsyncClient kinesis;
  private final DynamoDbAsyncClient dynamo;

  public LocalstackInfra(
      KinesisAsyncClient kinesis,
      DynamoDbAsyncClient dynamo,
      KinesisConfig kinesisConfig,
      @Value("#{{${alert-persister.org-threat.dynamo-tables}}}")
          Map<String, String> orgThreatTables)
      throws ExecutionException, InterruptedException {
    this.kinesis = kinesis;
    this.dynamo = dynamo;

    // create input stream
    createStreamIfNotExists(kinesisConfig.getClient().getStreamName());
    createTableIfNotExists(kinesisConfig.getClient().getApplicationName());

    // create output streams
    createStreamIfNotExists(kinesisConfig.getOutputStream().getStreamName());

    for (String table : orgThreatTables.values()) {
      createOrgThreatTableIfNotExists(table);
    }
  }

  public boolean streamExists(String name) {
    try {
      kinesis.describeStream(DescribeStreamRequest.builder().streamName(name).build()).get();
    } catch (ExecutionException | InterruptedException e) {
      return false;
    }
    return true;
  }

  public void createStreamIfNotExists(String name) throws InterruptedException, ExecutionException {
    if (!streamExists(name)) {
      CreateStreamRequest req =
          CreateStreamRequest.builder().streamName(name).shardCount(1).build();
      kinesis.createStream(req).get();
    }
  }

  public void deleteStream(String name) throws ExecutionException, InterruptedException {
    kinesis.deleteStream(DeleteStreamRequest.builder().streamName(name).build()).get();
  }

  public boolean tableExists(String name) {
    try {
      dynamo.describeTable(DescribeTableRequest.builder().tableName(name).build()).get();
      return true;
    } catch (InterruptedException | ExecutionException | ResourceNotFoundException e) {
      return false;
    }
  }

  public void createTableIfNotExists(String name) throws ExecutionException, InterruptedException {
    if (!tableExists(name)) {
      CreateTableRequest req =
          CreateTableRequest.builder()
              .tableName(name)
              .keySchema(
                  KeySchemaElement.builder()
                      .attributeName("leaseKey")
                      .keyType(KeyType.HASH)
                      .build())
              .attributeDefinitions(
                  AttributeDefinition.builder()
                      .attributeName("leaseKey")
                      .attributeType(ScalarAttributeType.S)
                      .build())
              .provisionedThroughput(
                  ProvisionedThroughput.builder()
                      .readCapacityUnits(100L)
                      .writeCapacityUnits(100L)
                      .build())
              .build();

      dynamo.createTable(req).get();
    }
  }

  private void createOrgThreatTableIfNotExists(String name)
      throws ExecutionException, InterruptedException {
    if (!tableExists(name)) {
      CreateTableRequest req =
          CreateTableRequest.builder()
              .tableName(name)
              .keySchema(
                  KeySchemaElement.builder()
                      .attributeName(OrgThreatClient.PARTITION_KEY_ATTRIBUTE)
                      .keyType(KeyType.HASH)
                      .build(),
                  KeySchemaElement.builder()
                      .attributeName(OrgThreatClient.SORT_KEY_ATTRIBUTE)
                      .keyType(KeyType.RANGE)
                      .build())
              .attributeDefinitions(
                  AttributeDefinition.builder()
                      .attributeName(OrgThreatClient.PARTITION_KEY_ATTRIBUTE)
                      .attributeType(ScalarAttributeType.S)
                      .build(),
                  AttributeDefinition.builder()
                      .attributeName(OrgThreatClient.SORT_KEY_ATTRIBUTE)
                      .attributeType(ScalarAttributeType.S)
                      .build())
              .provisionedThroughput(
                  ProvisionedThroughput.builder()
                      .readCapacityUnits(100L)
                      .writeCapacityUnits(100L)
                      .build())
              .build();

      dynamo.createTable(req).get();
    }
  }

  public void deleteTable(String name) throws ExecutionException, InterruptedException {
    dynamo.deleteTable(DeleteTableRequest.builder().tableName(name).build()).get();
  }
}
