package com.vmware.psc.alerting.persister.es;

import static com.vmware.psc.alerting.persister.util.CbUaeAlertUtil.*;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.protobuf.InvalidProtocolBufferException;
import com.google.protobuf.Timestamp;
import com.google.protobuf.util.JsonFormat;
import com.vmware.psc.alerting.persister.cache.metrics.CacheAlertAggregationMetrics;
import com.vmware.psc.alerting.persister.cache.service.CacheService;
import com.vmware.psc.alerting.persister.enums.AggregationUpdateType;
import com.vmware.psc.alerting.persister.es.enums.EsParamKeysType;
import com.vmware.psc.alerting.persister.exception.RetryableElasticsearchException;
import com.vmware.psc.alerting.persister.kinesis.producer.ProducerSupport;
import com.vmware.psc.alerting.persister.kinesis.utils.ProtoUtilHelper;
import com.vmware.psc.alerting.persister.metrics.PerAlertTypeCounter;
import com.vmware.psc.alerting.persister.metrics.RunStatusLogger;
import com.vmware.psc.alerting.persister.properties.EsUpdateMode;
import com.vmware.psc.alerting.persister.properties.EsWriterConfig;
import com.vmware.time.helper.TimeHelper;
import io.micrometer.core.annotation.Timed;
import io.micrometer.core.instrument.Counter;
import io.micrometer.core.instrument.DistributionSummary;
import io.micrometer.core.instrument.MeterRegistry;
import java.io.IOException;
import java.util.*;
import java.util.function.Function;
import lombok.extern.slf4j.Slf4j;
import org.opensearch.action.DocWriteRequest;
import org.opensearch.action.DocWriteResponse;
import org.opensearch.action.bulk.BulkItemResponse;
import org.opensearch.action.bulk.BulkRequest;
import org.opensearch.action.bulk.BulkResponse;
import org.opensearch.action.index.IndexRequest;
import org.opensearch.action.update.UpdateRequest;
import org.opensearch.action.update.UpdateResponse;
import org.opensearch.client.RequestOptions;
import org.opensearch.client.RestHighLevelClient;
import org.opensearch.common.unit.TimeValue;
import org.opensearch.common.xcontent.XContentType;
import org.opensearch.core.rest.RestStatus;
import org.opensearch.index.get.GetResult;
import org.opensearch.script.Script;
import org.opensearch.script.ScriptType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

@Component
@Slf4j
public class AlertWriter {
  private final RestHighLevelClient client;
  private final IPartitionStrategy partitionStrategy;
  private final JsonFormat.Printer jsonWriter;
  private final DistributionSummary batchRequestSize;
  private final Counter alertSerializationFailed;
  private final PerAlertTypeCounter alertsFailedToWrite;
  private final PerAlertTypeCounter alertsCreated;
  private final PerAlertTypeCounter alertsUpdated;
  private final PerAlertTypeCounter alertsUpdatedNoop;
  private final EsWriterConfig config;
  private final ProducerSupport producerSupport;
  private final CacheService<String, PSCAlert.CbUaeAlert> alertCacheService;

  @Value("${alert-persister.cache.aggregation.enabled}")
  private boolean isAggregationCacheEnabled;

  private final CacheAlertAggregationMetrics cacheAlertAggregationMetrics;
  private final RunStatusLogger runStatusLogger;

  @Autowired
  public AlertWriter(
      IPartitionStrategy partitionStrategy,
      MeterRegistry meterRegistry,
      EsProperties esProperties,
      Map<String, RestHighLevelClient> esClusterMap,
      ProducerSupport producerSupport,
      final CacheService<String, PSCAlert.CbUaeAlert> alertCacheService,
      RunStatusLogger runStatusLogger) {
    this.client = esClusterMap.get("alerts");
    this.partitionStrategy = partitionStrategy;
    this.alertCacheService = alertCacheService;
    this.jsonWriter =
        JsonFormat.printer().includingDefaultValueFields().omittingInsignificantWhitespace();
    this.batchRequestSize = meterRegistry.summary("es.batch_request_size");
    this.alertsCreated = new PerAlertTypeCounter(meterRegistry, "es.alerts_created");
    this.alertsUpdated = new PerAlertTypeCounter(meterRegistry, "es.alerts_updated");
    this.alertsUpdatedNoop = new PerAlertTypeCounter(meterRegistry, "es.alerts_updated_noop");
    this.alertsFailedToWrite = new PerAlertTypeCounter(meterRegistry, "es.alerts_failed_to_write");
    this.alertSerializationFailed = meterRegistry.counter("es.serialization_errors");
    this.config = esProperties.getEsConfigMap().get("alerts");
    this.producerSupport = producerSupport;
    this.cacheAlertAggregationMetrics =
        new CacheAlertAggregationMetrics(meterRegistry, "alert_aggregation_cache_metrics");
    this.runStatusLogger = runStatusLogger;
  }

  /**
   * Writes an entire list of alerts to ES using batching.
   *
   * <p>Batch size is controlled by {@link EsWriterConfig#getMaxRecordsPerRequest()}
   *
   * <p>As part of the write process, the create timestamp is set on each alert. Each batch within
   * the call will have different timestamps. This is done to make sure the create timestamp matches
   * the time an alert is written into ES as closely as possible.
   *
   * <p>After successfully writing a batch of alerts to ES, those alerts are removed from the input
   * list.
   *
   * @param alerts Alerts to write
   * @throws IOException Error writing alerts to ES
   */
  @Timed(value = "es.write_time")
  public void writeAndClearAlerts(
      List<PSCAlert.CbUaeAlert.Builder> alerts, final AggregationUpdateType aggregationUpdateType)
      throws IOException {
    if (alerts.isEmpty()) {
      log.debug("No alerts to persist");
      return;
    } else {
      log.debug("Attempting to persist {} alert(s)", alerts.size());
    }

    cacheAlertAggregationMetrics.increment(aggregationUpdateType, alerts.size());

    while (!alerts.isEmpty()) {
      List<PSCAlert.CbUaeAlert.Builder> batch =
          alerts.subList(0, Math.min(alerts.size(), config.getMaxRecordsPerRequest()));
      // Set the create timestamp on each alert as close to writing the alert to ES as possible
      Timestamp now = TimeHelper.timestampNow();

      writeAlertBatch(
          batch.stream().map(updateAlertTimestamps(now)).toList(), aggregationUpdateType);
      alerts.removeAll(batch);
    }
  }

  private Function<PSCAlert.CbUaeAlert.Builder, PSCAlert.CbUaeAlert> updateAlertTimestamps(
      Timestamp now) {
    return (PSCAlert.CbUaeAlert.Builder alert) -> {
      // create_timestamp and system_update_timestamp are both set to the same value to make
      // sorting on system_update_timestamp work.
      //
      // ES does not provide a way for sorts to fall back on a secondary field when the primary
      // field is missing. (we can't set create_time as another level of sorting after
      // system_update_timestamp because all the alerts without a system_update_timestamp will be
      // sorted *after* the ones with a system_update_timestamp)
      //
      // By making the system_update_timestamp the same as create_timestamp we ensure sorting on
      // system_update_timestamp will present alerts in a logical order.
      alert.setCreateTimestamp(now);
      alert.getAlertBuilder().setBackendTimestamp(now);
      alert.getAlertBuilder().setBackendUpdateTimestamp(now);
      alert.setSystemUpdateTimestamp(now);

      // The workflow last_update_timestamp and the create_timestamp need to match or else we'll
      // get customer questions.
      alert
          .getAlertBuilder()
          .getUserStateDetailsBuilder()
          .getWorkflowStateBuilder()
          .setLastUpdateTimestamp(now);

      alert
          .getAlertBuilder()
          .getUserStateDetailsBuilder()
          .setClassification(
              PSCRProto.CbAlertClassification.newBuilder().setLastUpdateTimestamp(now));
      if (alert.getAlert().hasMdrStateDetails()) {
        alert
            .getAlertBuilder()
            .getMdrStateDetailsBuilder()
            .getWorkflowStateBuilder()
            .setLastUpdateTimestamp(now);

        alert
            .getAlertBuilder()
            .getMdrStateDetailsBuilder()
            .getClassificationBuilder()
            .setLastUpdateTimestamp(now);
      }
      return alert.build();
    };
  }

  /**
   * Write a complete list of alerts to ES. It is the responsibility of the caller to ensure this
   * list is appropriately sized.
   *
   * <p>Consider using {@link AlertWriter#writeAndClearAlerts(List, AggregationUpdateType)} to page
   * through large collections of alerts.
   *
   * <p><strong>NOTE:</strong> this method does not update create/update timestamps before writing
   * to ES. That is only done via {@link AlertWriter#writeAndClearAlerts(List,
   * AggregationUpdateType)}. You probably want to use that method instead!
   *
   * @param alerts Alerts to write
   * @throws IOException Error persisting alerts
   */
  @Timed(value = "es.batch_write_time")
  public void writeAlertBatch(
      List<PSCAlert.CbUaeAlert> alerts, final AggregationUpdateType aggregationUpdateType)
      throws IOException {

    BulkRequest req = new BulkRequest();

    alerts.stream().map(getRequestFunc()).filter(Objects::nonNull).forEach(req::add);

    req.timeout(TimeValue.timeValueMinutes(1));
    BulkResponse bulk = client.bulk(req, RequestOptions.DEFAULT);
    batchRequestSize.record(bulk.getItems().length);

    if (config.getUpdateMode() == EsUpdateMode.SCRIPTED_UPSERT) {
      // Differentiate the different outcomes of the upsert request for better insight into what
      // sort of writes we're doing.
      for (var resp : bulk.getItems()) {
        PSCAlert.CbUaeAlert uaeAlertToUpdate = alerts.get(resp.getItemId());

        if (resp.isFailed()) {
          processFailedResponse(resp, uaeAlertToUpdate);
        } else {
          processSuccessResponse(resp, uaeAlertToUpdate, aggregationUpdateType);
        }
      }
    }
    logBulkProcessing(bulk);
  }

  /**
   * logBulkProcessing.
   *
   * @param bulk BulkResponse.
   */
  private void logBulkProcessing(final BulkResponse bulk) {
    if (bulk.hasFailures()) {
      log.error(
          "Bulk request of {} items completed in {} with failures: {}",
          bulk.getItems().length,
          bulk.getTook().toHumanReadableString(3),
          bulk.buildFailureMessage());
    } else {
      log.info(
          "Completed bulk request of {} items in {}",
          bulk.getItems().length,
          bulk.getTook().toHumanReadableString(3));

      this.runStatusLogger.incrementCountMetric(bulk.getItems().length);
    }
  }

  /**
   * Success response processing.
   *
   * @param resp BulkItemResponse.
   * @param uaeAlertToUpdate CbUaeAlert.
   */
  private void processSuccessResponse(
      final BulkItemResponse resp,
      final PSCAlert.CbUaeAlert uaeAlertToUpdate,
      final AggregationUpdateType aggregationUpdateType) {
    DocWriteResponse.Result operation = resp.getResponse().getResult();
    log.debug("ES response operation {}", operation);
    if (operation == DocWriteResponse.Result.CREATED) {
      final Optional<PSCAlert.CbUaeAlert> uaeAlertResponse = extractPersistedUpdateResponse(resp);
      syncCache(uaeAlertResponse, aggregationUpdateType);
      alertsCreated.increment(uaeAlertToUpdate);
      final PSCAlert.CbUaeAlert cbUaeAlertFromOS = uaeAlertResponse.orElseThrow();
      // publish the alert from the OS
      publishAlertToUdsStream(cbUaeAlertFromOS);
    } else if (operation == DocWriteResponse.Result.UPDATED) {
      final Optional<PSCAlert.CbUaeAlert> uaeAlertResponse = extractPersistedUpdateResponse(resp);
      syncCache(uaeAlertResponse, aggregationUpdateType);
      final PSCAlert.CbUaeAlert cbUaeAlertFromOS = uaeAlertResponse.orElseThrow();
      boolean hasNewEventAdded = updateTypeIsEventAdded(cbUaeAlertFromOS);
      if (hasNewEventAdded) {
        // When the update type is EVENT_ADDED we only update the first/last event timestamp
        // fields in OS.
        // This means that the alert from OS does not reflect the same event from kinesis that
        // triggered the update.
        // In this case, we want to publish the original alert from kinesis back to UDS after
        // updating it to include the latest timestamp details.
        PSCAlert.CbUaeAlert updatedAlert =
            mergeDetailsToIncomingEvent(uaeAlertToUpdate, cbUaeAlertFromOS);
        publishAlertToUdsStream(updatedAlert);
      } else {
        // For other update types (THREAT_CAUSE_CHANGED/THREAT_SCORE_INCREASED), the alert
        // returned from OS is the same event from kinesis that triggered the update.
        // We can publish the OS copy of the alert directly.
        publishAlertToUdsStream(cbUaeAlertFromOS);
      }

      alertsUpdated.increment(uaeAlertToUpdate);
    } else if (operation == DocWriteResponse.Result.NOOP) {
      final Optional<PSCAlert.CbUaeAlert> uaeAlertResponse = extractPersistedUpdateResponse(resp);
      // All alerts in the cache have already been sent to UDS. This means cache evict noops don't
      // need to go through UDS again since no new data is added
      if (aggregationUpdateType != AggregationUpdateType.EVICT) {
        syncCache(uaeAlertResponse, aggregationUpdateType);

        uaeAlertResponse.ifPresentOrElse(
            cbUaeAlertFromOS -> {
              // publish the alert from the OS
              // if new events added publish what ever coming from the stream
              // Merge the backendTimestamp, backendUpdateTimestamp, workflow, determination and etc
              PSCAlert.CbUaeAlert updatedAlert =
                  mergeDetailsToIncomingEvent(uaeAlertToUpdate, cbUaeAlertFromOS);
              publishAlertToUdsStream(updatedAlert);
            },
            // still want to publish back to UDS -- just not with one of those UPDATE_TYPES sent
            () -> publishAlertToUdsStream(uaeAlertToUpdate));
      }

      alertsUpdatedNoop.increment(uaeAlertToUpdate);
    } else {
      log.warn(
          "Unexpected result {} from ES operation. Index={}, ID={}",
          operation,
          resp.getIndex(),
          resp.getId());
    }
  }

  /**
   * Publish the alert to stream.
   *
   * @param cbUaeAlert CbUaeAlert.
   */
  private void publishAlertToUdsStream(final PSCAlert.CbUaeAlert cbUaeAlert) {
    producerSupport.publishUaeAlert(cbUaeAlert, getAlertId(cbUaeAlert));
  }

  /**
   * Merge alert-service-owned details from the OS alert to the incoming alert from kinesis
   *
   * @param uaeAlertToUpdate Incoming alert.
   * @param cbUaeAlertFromOS From Open search.
   */
  private PSCAlert.CbUaeAlert mergeDetailsToIncomingEvent(
      PSCAlert.CbUaeAlert uaeAlertToUpdate, PSCAlert.CbUaeAlert cbUaeAlertFromOS) {
    final PSCAlert.CbUaeAlert.Builder uaeAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(uaeAlertToUpdate);
    final PSCRProto.CbAlert.Builder alertBuilder = uaeAlertBuilder.getAlertBuilder();
    alertBuilder
        .setBackendTimestamp(getBackendTimestamp(cbUaeAlertFromOS))
        .setBackendUpdateTimestamp(getBackendUpdateTimestamp(cbUaeAlertFromOS))
        .setUserStateDetails(cbUaeAlertFromOS.getAlert().getUserStateDetails())
        .setMdrStateDetails(cbUaeAlertFromOS.getAlert().getMdrStateDetails())
        .setAlertCreationTimestamp(cbUaeAlertFromOS.getAlert().getAlertCreationTimestamp())
        .setAlertUpdateTimestamp(cbUaeAlertFromOS.getAlert().getAlertUpdateTimestamp())
        .setAlertFirstSeenTimestamp(cbUaeAlertFromOS.getAlert().getAlertFirstSeenTimestamp())
        .setAlertLastSeenTimestamp(cbUaeAlertFromOS.getAlert().getAlertLastSeenTimestamp());
    uaeAlertBuilder
        .setThreatNotesPresent(cbUaeAlertFromOS.getThreatNotesPresent())
        .setAlertNotesPresent(cbUaeAlertFromOS.getAlertNotesPresent())
        .addAllInvestigationTags(cbUaeAlertFromOS.getInvestigationTagsList())
        .setCreateTimestamp(cbUaeAlertFromOS.getCreateTimestamp())
        .setSystemUpdateTimestamp(cbUaeAlertFromOS.getSystemUpdateTimestamp())
        .setUserUpdateTimestamp(cbUaeAlertFromOS.getUserUpdateTimestamp());

    return uaeAlertBuilder.build();
  }

  /**
   * processFailedResponse.
   *
   * @param resp BulkItemResponse.
   * @param uaeAlertToUpdate CbUaeAlert.
   * @throws RetryableElasticsearchException RetryableElasticsearchException.
   * @throws InvalidProtocolBufferException InvalidProtocolBufferException.
   */
  private void processFailedResponse(
      final BulkItemResponse resp, final PSCAlert.CbUaeAlert uaeAlertToUpdate)
      throws RetryableElasticsearchException, InvalidProtocolBufferException {
    alertsFailedToWrite.increment(uaeAlertToUpdate);
    BulkItemResponse.Failure failure = resp.getFailure();
    if (failure.getStatus() == RestStatus.FORBIDDEN) {
      // Response status FORBIDDEN means the index is blocking writes so the operation will
      // need to be retried later
      throw new RetryableElasticsearchException(resp.getIndex(), failure.getCause());
    } else if (failure.getStatus() == RestStatus.BAD_REQUEST
        && (failure.getMessage().contains("type=index_closed_exception")
            || failure.getMessage().contains("resource_not_found_exception"))) {
      // There doesn't seem to be a better way to check for the index being closed or the
      // upsert script being missing.
      // When the response is BAD_REQUEST and we find one of these known exceptions it makes
      // sense to retry.
      // Otherwise, the BAD_REQUEST likely means bad data, which will never work, so we
      // should just move on.
      throw new RetryableElasticsearchException(resp.getIndex(), failure.getCause());
    } else if (failure.getStatus() == RestStatus.TOO_MANY_REQUESTS) {
      log.warn(
          "Got a TOO_MANY_REQUESTS - 429 failure while writing alert {}. Status={}, Error={}, body={}",
          getAlertId(uaeAlertToUpdate),
          failure.getStatus(),
          failure.getMessage(),
          jsonWriter.print(uaeAlertToUpdate));
      throw new RetryableElasticsearchException(resp.getIndex(), failure.getCause());
    } else {
      log.warn(
          "Got a non-retryable error writing alert {}. Dropping it. Status={}, Error={}, body={}",
          getAlertId(uaeAlertToUpdate),
          failure.getStatus(),
          failure.getMessage(),
          jsonWriter.print(uaeAlertToUpdate));
    }
  }

  /**
   * Sync Cache with alert coming from the ES response.
   *
   * @param uaeAlertResponse CbUaeAlert.
   */
  private void syncCache(
      final Optional<PSCAlert.CbUaeAlert> uaeAlertResponse,
      final AggregationUpdateType aggregationUpdateType) {
    log.debug(
        "Sync Cache aggregation enabled {} and update type {}",
        isAggregationCacheEnabled,
        aggregationUpdateType);
    if (isAggregationCacheEnabled
        && (aggregationUpdateType == AggregationUpdateType.NEW
            || aggregationUpdateType == AggregationUpdateType.SCORE
            || aggregationUpdateType == AggregationUpdateType.JAVA_ANALYTICS_ALERT_ELEVATED
            || aggregationUpdateType == AggregationUpdateType.FLUSH_INTERVAL
            || aggregationUpdateType == AggregationUpdateType.JAVA_ANALYTICS_PRIMARY)) {
      uaeAlertResponse.ifPresent(
          alert -> {
            final String alertId = getAlertId(alert);
            log.debug("Update alert back to cache for alert Id {}", alertId);
            PSCAlert.CbUaeAlert.Builder builder =
                PSCAlert.CbUaeAlert.newBuilder(alert).setUpdateCount(0L);
            builder.setCreateTimestamp(TimeHelper.timestampNow());
            alertCacheService.put(alertId, builder.build());
          });
    }
  }

  private Function<PSCAlert.CbUaeAlert, DocWriteRequest<?>> getRequestFunc() {
    if (config.getUpdateMode() == EsUpdateMode.SCRIPTED_UPSERT) {
      return this::scriptedUpsertForAlert;
    }
    return this::indexRequestForAlert;
  }

  private IndexRequest indexRequestForAlert(PSCAlert.CbUaeAlert alert) {
    try {
      String index = partitionStrategy.getIndexForDocument(alert);
      String id = getAlertId(alert);
      String json = jsonWriter.print(alert);
      return new IndexRequest(index).id(id).source(json, XContentType.JSON);
    } catch (InvalidProtocolBufferException e) {
      // Definitely no sense retrying here, just mark a counter and drop the record.
      alertSerializationFailed.increment();
      log.error("Error serializing alert {} to json. contents: {}", getAlertId(alert), alert, e);
      return null;
    }
  }

  private UpdateRequest scriptedUpsertForAlert(PSCAlert.CbUaeAlert alert) {
    try {
      String index = partitionStrategy.getIndexForDocument(alert);
      String id = getAlertId(alert);
      String json = jsonWriter.print(alert);
      ObjectMapper mapper = new ObjectMapper();
      final Map<String, Object> params =
          Map.of(EsParamKeysType.UPDATE.key(), mapper.readValue(json, HashMap.class));
      return new UpdateRequest(index, id)
          .scriptedUpsert(true)
          .upsert(Collections.emptyMap())
          .retryOnConflict(config.getUpsertConflictRetries())
          .fetchSource(config.isShouldFetchSource())
          .script(new Script(ScriptType.STORED, null, "alert-upsert", params));

    } catch (InvalidProtocolBufferException | JsonProcessingException e) {
      // Definitely no sense retrying here, just mark a counter and drop the record.
      alertSerializationFailed.increment();
      log.error("Error serializing alert {} to json. contents: {}", getAlertId(alert), alert, e);
      return null;
    } catch (IllegalArgumentException e) {
      if (e.getMessage().contains("Timestamp is not valid")) {
        log.error(
            "Error serializing alert {} to json. Contains an illegal timestamp (message: {}). contents: {}",
            getAlertId(alert),
            e.getMessage(),
            alert);
        return null;
      } else {
        throw e;
      }
    }
  }

  /**
   * Extract the Alert from the ES Response.
   *
   * @param resp BulkItemResponse.
   * @return Optional.
   */
  private static Optional<PSCAlert.CbUaeAlert> extractPersistedUpdateResponse(
      final BulkItemResponse resp) {
    return Optional.ofNullable(resp)
        .map(BulkItemResponse::<UpdateResponse>getResponse)
        .map(UpdateResponse::getGetResult)
        .map(GetResult::sourceAsString)
        .map(
            alertJson -> {
              try {
                return ProtoUtilHelper.toProto(alertJson, PSCAlert.CbUaeAlert.getDefaultInstance());
              } catch (InvalidProtocolBufferException e) {
                throw new IllegalStateException(e);
              }
            });
  }
}
