package com.vmware.psc.alerting.persister.es;

import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.AlertType.*;
import static com.carbonblack.protocols.psc.response.PSCRProto.ThreatScore.CbThreatSource.THREAT_SOURCE_DEFENSE_ANALYTICS;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.AlertType;
import com.google.protobuf.Timestamp;
import com.vmware.time.helper.TimeHelper;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.EnumSet;
import java.util.Set;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Component;

/**
 * Daily Partitioning Strategy.
 *
 * <p>This strategy produces a single <strong>write index</strong> per day. All alerts created in
 * the same UTC day will end up in the same index.
 *
 * <p>Alerts that are <strong>never updated</strong> are partitioned by ingress time. This ensures
 * that, in the case of kinesis reprocessing, these alerts will be re-written to the original index.
 *
 * <p>Alerts that are <strong>updated</strong> are partitioned by a "create timestamp" provided by
 * the producer. This timestamp must be from a known-good source (ie, not based on sensor time as
 * that could be anything). Check the logic below to see how each specific alert type is handled.
 */
@Component
@Slf4j
public class DailyFixedRetentionPartitionStrategy implements IPartitionStrategy {

  DateTimeFormatter alertIndexFormatter =
      DateTimeFormatter.ISO_LOCAL_DATE.withZone(ZoneId.of("UTC"));

  /**
   * Alert types in this list guarantee that the producer will never send any updates. All alert
   * types in this list will partition by ingress timestamp.
   */
  Set<AlertType> IMMUTABLE_ALERT_TYPES =
      EnumSet.of(
          ALERT_TYPE_DEVICE_CONTROL,
          ALERT_TYPE_QUERY_WATCHLIST,
          ALERT_TYPE_INGRESS_WATCHLIST,
          ALERT_TYPE_HASHBAN);

  @Override
  public String getIndexForDocument(PSCAlert.CbUaeAlertOrBuilder alert) {
    Timestamp partitionTimestamp = getPartitionTimestamp(alert);
    Instant partitionInstant = TimeHelper.toInstant(partitionTimestamp);
    return String.format("alert-%s-r180", alertIndexFormatter.format(partitionInstant));
  }

  Timestamp getPartitionTimestamp(PSCAlert.CbUaeAlertOrBuilder alert) {
    // Immutable alerts (alerts that are never updated) use the ingress timestamp.
    // This accounts for kinesis reprocessing situations where we see the same alert multiple times.
    // This way, we'll write to the original index and no-op.
    if (IMMUTABLE_ALERT_TYPES.contains(alert.getAlert().getAlertType())) {
      return alert.getEventBatchHeader().getIngressTimestamp();
    }

    // Container Runtime alerts use the alert creation timestamp for partitioning.
    // This timestamp is generated by the CR backend -- rather than by the sensor itself -- so this
    // timestamp can be trusted to be the real UTC time.
    // This alert type should also use the aggregation pattern, but that's a different effort.
    if (alert.getAlert().getAlertType() == ALERT_TYPE_K8S_RUNTIME) {
      return alert.getAlert().getAlertCreationTimestamp();
    }

    // Aggregated alerts use the ingress timestamp.
    // The aggregation service generates the aggregate GUID based on the UTC date from the ingress
    // timestamp, so we can use it for partitioning.
    if (alert.getAlert().getAlertAggregation().getIsAggregated()) {
      return alert.getEventBatchHeader().getIngressTimestamp();
    }

    // Threat alerts vary based on whether they're from java analytics or DRE.
    if (alert.getAlert().getAlertType() == ALERT_TYPE_THREAT) {
      if (alert.getAlert().getAlertThreatScore().getThreatSource()
          == THREAT_SOURCE_DEFENSE_ANALYTICS) {
        // Threat alerts from analytics use the alert creation timestamp for partitioning.
        // This timestamp is generated by analytics and tracked across updates. Analytics can be
        // trusted to provide a real UTC time.
        return alert.getAlert().getAlertCreationTimestamp();
      } else {
        // All other threat alerts should be treated as one-offs and use the ingress timestamp like
        // the other immutable alert types.
        return alert.getEventBatchHeader().getIngressTimestamp();
      }
    }

    log.warn(
        "Falling back to CbAlert.alert_creation_timestamp for CbAlert(id={}, type={}, timestamp={})",
        alert.getAlert().getAlertIdGuid(),
        alert.getAlert().getAlertType(),
        alert.getAlert().getAlertCreationTimestamp());
    return alert.getAlert().getAlertCreationTimestamp();
  }
}
