package com.vmware.psc.alerting.persister.es;

import com.amazonaws.regions.AwsRegionProvider;
import com.vmware.psc.alerting.escommon.AwsRequestSigningApacheInterceptor;
import com.vmware.psc.alerting.persister.properties.EsWriterConfig;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.HttpHost;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.opensearch.client.RestClient;
import org.opensearch.client.RestClientBuilder;
import org.opensearch.client.RestHighLevelClient;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Profile;
import org.springframework.util.StringUtils;
import software.amazon.awssdk.auth.credentials.AwsCredentialsProvider;
import software.amazon.awssdk.auth.signer.Aws4Signer;

/**
 * Elasticsearch Config Customization
 *
 * <p>To use IAM auth on our AWS-managed ES clusters, we need to sign all requests. These signatures
 * are added via a custom request interceptor.
 *
 * <p>In switching to OpenSearch we need to wire our own ES client as we can no longer use the
 * spring-default autoconfiguration for elasticsearch. Someday we can likely move back to the
 * auto-configuration, but until then we can use this forked version.
 */
@Configuration
@Slf4j
@RequiredArgsConstructor
public class ElasticsearchConfig {

  private final Map<String, RestHighLevelClient> esClientMap = new HashMap<>();

  /**
   * Only create the Signing Interceptor for aws deployments.
   *
   * <p>This interceptor properly signs our requests for aws iam auth.
   */
  @Bean
  @Profile("aws")
  @ConditionalOnProperty(
      value = "alert-persister.elasticsearch-connect-with-basic-credentials",
      havingValue = "false")
  AwsRequestSigningApacheInterceptor requestSigningInterceptor(
      AwsRegionProvider regionProvider, AwsCredentialsProvider credentialsProvider) {
    return new AwsRequestSigningApacheInterceptor(
        "es", Aws4Signer.create(), credentialsProvider, regionProvider.getRegion());
  }

  /**
   * High level es client map
   *
   * @param optInterceptor AwsRequestSigningApacheInterceptor
   * @param esProperties EsProperties
   * @return The client for use in our apps
   */
  @Bean
  @ConditionalOnProperty(
      value = "alert-persister.elasticsearch-connect-with-basic-credentials",
      havingValue = "false")
  Map<String, RestHighLevelClient> esClusterMapRequestSigning(
      Optional<AwsRequestSigningApacheInterceptor> optInterceptor, EsProperties esProperties) {
    for (Map.Entry<String, EsWriterConfig> entries : esProperties.getEsConfigMap().entrySet()) {
      log.info("es cluster map: {}:{}", entries.getKey(), entries.getValue().getUris().get(0));
      esClientMap.put(
          entries.getKey(),
          new RestHighLevelClient(restClientBuilder(optInterceptor, entries.getValue())));
    }
    return esClientMap;
  }

  @Bean
  @ConditionalOnProperty(
      value = "alert-persister.elasticsearch-connect-with-basic-credentials",
      havingValue = "true")
  Map<String, RestHighLevelClient> esClusterMapBasicCred(EsProperties esProperties) {
    for (Map.Entry<String, EsWriterConfig> entries : esProperties.getEsConfigMap().entrySet()) {
      log.info(
          "es cluster map (basic cred): {}:{}",
          entries.getKey(),
          entries.getValue().getUris().get(0));
      esClientMap.put(
          entries.getKey(),
          new RestHighLevelClient(restClientBuilderWithBasicCredentials(entries.getValue())));
    }
    return esClientMap;
  }

  /**
   * Configure the rest client builder. The interceptor is optional as localdev does not need this
   * configuration.
   *
   * @param interceptor AWS Auth interceptor (aws-deployed envs only)
   * @param esWriterConfig Elasticsearch config for one cluster
   * @return RestClientBuilder
   */
  private static RestClientBuilder restClientBuilder(
      Optional<AwsRequestSigningApacheInterceptor> interceptor, EsWriterConfig esWriterConfig) {
    HttpHost[] hosts =
        esWriterConfig.getUris().stream()
            .map(ElasticsearchConfig::createHttpHost)
            .toArray(HttpHost[]::new);
    RestClientBuilder builder = RestClient.builder(hosts);

    interceptor.ifPresent(
        awsRequestSigningApacheInterceptor ->
            builder.setHttpClientConfigCallback(
                httpClientBuilder -> {
                  httpClientBuilder.addInterceptorLast(awsRequestSigningApacheInterceptor);
                  return httpClientBuilder;
                }));

    builder.setRequestConfigCallback(
        requestConfigBuilder -> {
          if (esWriterConfig.getConnectionTimeout() != null) {
            requestConfigBuilder.setConnectTimeout(
                (int) esWriterConfig.getConnectionTimeout().toMillis());
          }

          if (esWriterConfig.getSocketTimeout() != null) {
            requestConfigBuilder.setSocketTimeout(
                (int) esWriterConfig.getSocketTimeout().toMillis());
          }

          return requestConfigBuilder;
        });

    return builder;
  }

  private static RestClientBuilder restClientBuilderWithBasicCredentials(
      EsWriterConfig esWriterConfig) {
    HttpHost[] hosts =
        esWriterConfig.getUris().stream()
            .map(ElasticsearchConfig::createHttpHost)
            .toArray(HttpHost[]::new);

    RestClientBuilder builder = RestClient.builder(hosts);

    final CredentialsProvider credentialsProvider = new BasicCredentialsProvider();
    credentialsProvider.setCredentials(
        AuthScope.ANY,
        new UsernamePasswordCredentials(esWriterConfig.getUser(), esWriterConfig.getPass()));

    builder.setHttpClientConfigCallback(
        httpClientBuilder -> httpClientBuilder.setDefaultCredentialsProvider(credentialsProvider));

    builder.setRequestConfigCallback(
        requestConfigBuilder -> {
          if (esWriterConfig.getConnectionTimeout() != null) {
            requestConfigBuilder.setConnectTimeout(
                (int) esWriterConfig.getConnectionTimeout().toMillis());
          }

          if (esWriterConfig.getSocketTimeout() != null) {
            requestConfigBuilder.setSocketTimeout(
                (int) esWriterConfig.getSocketTimeout().toMillis());
          }

          return requestConfigBuilder;
        });

    return builder;
  }

  /**
   * Helper to convert a string into a HttpHost
   *
   * <p>Copied from {@link
   * org.springframework.boot.autoconfigure.elasticsearch.ElasticsearchRestClientConfigurations.RestClientBuilderConfiguration#createHttpHost(String)}.
   */
  private static HttpHost createHttpHost(String uri) {
    try {
      return createHttpHost(URI.create(uri));
    } catch (IllegalArgumentException ex) {
      return HttpHost.create(uri);
    }
  }

  private static HttpHost createHttpHost(URI uri) {
    if (!StringUtils.hasLength(uri.getUserInfo())) {
      return HttpHost.create(uri.toString());
    }
    try {
      return HttpHost.create(
          new URI(
                  uri.getScheme(),
                  null,
                  uri.getHost(),
                  uri.getPort(),
                  uri.getPath(),
                  uri.getQuery(),
                  uri.getFragment())
              .toString());
    } catch (URISyntaxException ex) {
      throw new IllegalStateException(ex);
    }
  }
}
