package com.vmware.psc.alerting.persister.es.notes;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.vmware.psc.alerting.persister.exception.RetryableElasticsearchException;
import com.vmware.psc.alerting.persister.model.notes.Notes;
import io.micrometer.core.annotation.Timed;
import io.micrometer.core.instrument.Counter;
import io.micrometer.core.instrument.MeterRegistry;
import java.io.IOException;
import java.time.Instant;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Collectors;
import lombok.extern.slf4j.Slf4j;
import org.opensearch.action.DocWriteRequest;
import org.opensearch.action.bulk.BulkRequest;
import org.opensearch.action.bulk.BulkResponse;
import org.opensearch.action.index.IndexRequest;
import org.opensearch.client.RequestOptions;
import org.opensearch.client.RestHighLevelClient;
import org.opensearch.common.unit.TimeValue;
import org.opensearch.common.xcontent.XContentType;
import org.opensearch.core.rest.RestStatus;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

@Component
@Slf4j
public class NotesWriter {

  private final RestHighLevelClient client;
  private final NotesPartitionStrategy partitionStrategy;
  private final ObjectMapper objectMapper;
  private final MeterRegistry meterRegistry;
  private final Counter failedCount;
  private final Counter versionConflictCount;

  @Value("${alert-persister.notes-config.chunk-size:200}")
  private int chunkSize;

  public NotesWriter(
      Map<String, RestHighLevelClient> esClusterMap,
      NotesPartitionStrategy partitionStrategy,
      ObjectMapper objectMapper,
      MeterRegistry meterRegistry) {
    this.client = esClusterMap.get("notes");
    this.partitionStrategy = partitionStrategy;
    this.objectMapper = objectMapper;
    this.meterRegistry = meterRegistry;
    failedCount = meterRegistry.counter("es.notes_error");
    versionConflictCount = meterRegistry.counter("es.notes_version_conflict");
  }

  /**
   * The method takes a list of notes and write into elasticsearch. It forces
   * "DocWriteRequest.OpType.CREATE" operation so that we can dedupe things by using the same hashid
   *
   * @param notes A list of notes to flush
   */
  @Timed(value = "es.note.batch_write_time")
  public void writeNotesBatch(List<Notes> notes) throws IOException {
    meterRegistry.gauge("es.notes_list_size", notes.size());
    int start = 0;
    int end = 0;
    while (end < notes.size()) {
      end = Math.min(start + chunkSize, notes.size());
      List<Notes> subList = notes.subList(start, end);
      Instant now = Instant.now();
      writeNotesInBatch(
          subList.stream().map(updateNotesTimestamp(now)).collect(Collectors.toList()));
      start = end;
    }
  }

  private Function<Notes, Notes> updateNotesTimestamp(Instant now) {
    return (Notes note) -> {
      note.setLastUpdateTimestamp(now);
      note.setCreateTimestamp(now);
      return note;
    };
  }

  void writeNotesInBatch(List<Notes> notes) throws IOException {
    BulkRequest req = new BulkRequest();
    notes.stream().map(this::indexRequestForNotes).filter(Objects::nonNull).forEach(req::add);
    req.timeout(TimeValue.timeValueMinutes(1));

    if (req.requests().isEmpty()) {
      return;
    }

    BulkResponse bulk = client.bulk(req, RequestOptions.DEFAULT);

    for (var resp : bulk.getItems()) {
      Notes original = notes.get(resp.getItemId());
      if (resp.isFailed()) {
        // Fail
        failedCount.count();
        var failure = resp.getFailure();
        log.debug("writeNotesBatch: {}", resp.getFailure().getStatus());
        if (failure.getStatus() == RestStatus.CONFLICT) {
          versionConflictCount.count();
          log.debug("writeNotesBatch: Document already exist for notes={}", original);
        } else {
          throw new RetryableElasticsearchException(resp.getIndex(), failure.getCause());
        }
      }
    }
  }

  private IndexRequest indexRequestForNotes(Notes notes) {
    try {
      String index = partitionStrategy.getIndexForNotes(notes);
      String id = notes.getId();
      if (id == null) {
        log.warn("indexRequestForNotes: notes id is not set {}", notes);
        id =
            buildNotesId(
                notes.getAlertId(),
                notes.getSuppressionRuleId(),
                notes.getSuppressionNoteVersion(),
                notes.isInternal());
      }
      String json = objectMapper.writeValueAsString(notes);
      return new IndexRequest(index)
          .id(id)
          .source(json, XContentType.JSON)
          .opType(DocWriteRequest.OpType.CREATE);
    } catch (JsonProcessingException e) {
      // TODO Definitely no sense retrying here, just mark a counter and drop the record.
      log.error("Error serializing to json. alertId={}, content={}", notes.getAlertId(), notes, e);
      return null;
    }
  }

  public static String buildNotesId(
      String alertId, String suppressionRuleId, long suppressionVersion, boolean isInternal) {
    String decodedId =
        String.format("%s-%s-%s-%s", alertId, suppressionRuleId, suppressionVersion, isInternal);
    return UUID.nameUUIDFromBytes(decodedId.getBytes()).toString();
  }
}
