package com.vmware.psc.alerting.persister.kinesis.client;

import com.vmware.psc.alerting.persister.kinesis.client.health.KinesisApplicationHealthIndicator;
import com.vmware.psc.alerting.persister.kinesis.client.properties.KinesisClientProperties;
import java.util.Optional;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.lang.Nullable;
import software.amazon.awssdk.services.cloudwatch.CloudWatchAsyncClient;
import software.amazon.awssdk.services.dynamodb.DynamoDbAsyncClient;
import software.amazon.awssdk.services.kinesis.KinesisAsyncClient;
import software.amazon.kinesis.common.ConfigsBuilder;
import software.amazon.kinesis.coordinator.Scheduler;
import software.amazon.kinesis.leases.LeaseManagementConfig;
import software.amazon.kinesis.lifecycle.LifecycleConfig;
import software.amazon.kinesis.metrics.MetricsConfig;
import software.amazon.kinesis.processor.ProcessorConfig;
import software.amazon.kinesis.processor.ShardRecordProcessorFactory;
import software.amazon.kinesis.retrieval.RetrievalConfig;
import software.amazon.kinesis.retrieval.polling.PollingConfig;

/** Service that queues behavior events to be augmented and persisted via Kinesis. */
@Slf4j
@Setter
public class KinesisApplication implements InitializingBean, DisposableBean {

  protected final KinesisClientProperties properties;
  private final KinesisAsyncClient kinesis;
  private final DynamoDbAsyncClient dynamo;
  private final CloudWatchAsyncClient cloudwatch;
  private final ShardRecordProcessorFactory recordProcessorFactory;
  private final KinesisApplicationHealthIndicator healthIndicator;

  private final String workerId;
  private Scheduler scheduler;

  public KinesisApplication(
      KinesisClientProperties properties,
      KinesisAsyncClient kinesis,
      DynamoDbAsyncClient dynamo,
      CloudWatchAsyncClient cloudwatch,
      ShardRecordProcessorFactory recordProcessorFactory) {
    this(properties, kinesis, dynamo, cloudwatch, recordProcessorFactory, null);
  }

  public KinesisApplication(
      KinesisClientProperties properties,
      KinesisAsyncClient kinesis,
      DynamoDbAsyncClient dynamo,
      CloudWatchAsyncClient cloudwatch,
      ShardRecordProcessorFactory recordProcessorFactory,
      @Nullable KinesisApplicationHealthIndicator healthIndicator) {
    properties.validate();
    this.properties = properties;
    this.kinesis = kinesis;
    this.dynamo = dynamo;
    this.cloudwatch = cloudwatch;
    this.recordProcessorFactory = recordProcessorFactory;
    this.healthIndicator = healthIndicator;

    workerId = properties.generateWorkerId();
  }

  @Override
  public void afterPropertiesSet() throws Exception {

    ConfigsBuilder configsBuilder =
        new ConfigsBuilder(
            properties.getStreamName(),
            properties.getApplicationName(),
            kinesis,
            dynamo,
            cloudwatch,
            workerId,
            recordProcessorFactory);

    var coordinatorConfig = configsBuilder.coordinatorConfig();
    if (healthIndicator != null) {
      coordinatorConfig.workerStateChangeListener(healthIndicator);
    }

    LeaseManagementConfig leaseManagementConfig =
        configsBuilder
            .leaseManagementConfig()
            .maxLeasesForWorker(properties.getMaxLeasesForWorker())
            .failoverTimeMillis(properties.getLeaseFailoverTimeMs());

    MetricsConfig metricsConfig =
        configsBuilder.metricsConfig().metricsLevel(properties.getMetricsLevel());
    // TODO: can configure micrometer as a metrics factory to publish directly to wavefront instead
    // of cloudwatch

    LifecycleConfig lifecycleConfig =
        configsBuilder
            .lifecycleConfig()
            .logWarningForTaskAfterMillis(
                Optional.of(properties.getLogWarningForTaskAfterMillis()));

    ProcessorConfig processorConfig =
        configsBuilder
            .processorConfig()
            .callProcessRecordsEvenForEmptyRecordList(
                properties.isCallProcessRecordsEvenForEmptyRecordList());

    // NOTE: using PollingConfig is essentially the same as 1.x KCL apps.
    PollingConfig pollingConfig =
        new PollingConfig(properties.getStreamName(), kinesis)
            .maxRecords(properties.getMaxRecordsPerRequest())
            .idleTimeBetweenReadsInMillis(properties.getIdleTimeBetweenReadsMs());

    RetrievalConfig retrievalConfig =
        configsBuilder
            .retrievalConfig()
            .initialPositionInStreamExtended(properties.initialPositionInStreamExtended())
            .retrievalSpecificConfig(pollingConfig);

    try {
      scheduler =
          new Scheduler(
              configsBuilder.checkpointConfig(),
              coordinatorConfig,
              leaseManagementConfig,
              lifecycleConfig,
              metricsConfig,
              processorConfig,
              retrievalConfig);
      log.info(
          "Started kinesis worker {} for app {} on stream {}",
          workerId,
          properties.getApplicationName(),
          properties.getStreamName());
      new Thread(scheduler).start();
    } catch (Exception ex) {
      log.error(
          "Exception initializing Kinesis client for application {} on stream {}: {}",
          properties.getApplicationName(),
          properties.getStreamName(),
          ex.getMessage(),
          ex);

      // Destroy the worker if it got created by any chance.
      destroy();
    }
  }

  @Override
  public void destroy() throws Exception {
    if (scheduler != null) {
      log.warn("destroy: Closing the KCL application");
      try {
        Future<Boolean> shutdownStatus = scheduler.startGracefulShutdown();
        // wait max of 30 seconds for shutdown
        if (shutdownStatus != null) {
          Boolean success =
              shutdownStatus.get(properties.getShutdownWaitSeconds(), TimeUnit.SECONDS);
          if (Boolean.TRUE.equals(success)) {
            log.warn("destroy: KCL Application Shutdown successful");
          } else {
            log.warn(
                "destroy: KCL Application - a non-exception case caused the shutdown process to terminate early");
          }
        }
      } catch (Exception ex) {
        log.warn("destroy: Exception on shutdown: " + ex.getMessage(), ex);
      }
    }
  }
}
