package com.vmware.psc.alerting.persister.kinesis.client;

import com.carbonblack.psc.backoff.ExponentialBackoff;
import com.carbonblack.psc.backoff.exception.AbortedExecutionException;
import com.carbonblack.psc.backoff.exception.RetriesExceededException;
import com.vmware.psc.alerting.persister.kinesis.client.properties.RecordProcessorProperties;
import com.vmware.psc.alerting.persister.kinesis.exception.AbandonWorkerException;
import com.vmware.psc.alerting.persister.kinesis.exception.RecordParseException;
import io.micrometer.core.instrument.DistributionSummary;
import io.micrometer.core.instrument.MeterRegistry;
import io.micrometer.core.instrument.Tags;
import io.micrometer.core.instrument.Timer;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.slf4j.MDC;
import org.xerial.snappy.Snappy;
import software.amazon.kinesis.lifecycle.events.*;
import software.amazon.kinesis.processor.RecordProcessorCheckpointer;
import software.amazon.kinesis.processor.ShardRecordProcessor;
import software.amazon.kinesis.retrieval.KinesisClientRecord;

/**
 * Base Kinesis Record processor
 *
 * <p>Handles checkpoints and retries and collects related metrics
 */
@Slf4j
public abstract class RecordProcessorImpl<T> implements ShardRecordProcessor {

  private final MeterRegistry registry;
  private final RecordProcessorMetrics metrics;

  private final int maxEventsBeforeCheckpoint;
  private final long checkpointIntervalMs;
  private final boolean useCompression;

  private final ExponentialBackoff processBackoff;
  private final ExponentialBackoff checkpointBackoff;

  @Getter private String kinesisShardId = null;
  private DistributionSummary perShardMillisBehindLatest;

  private long nextCheckpointTimeMs = 0;
  private int eventsSinceCheckpoint = 0;
  private String lastCheckpointedRecord;
  private String lastProcessedRecord;

  private static final String SHARD_ID_KEY = "shardId";

  protected RecordProcessorImpl(
      MeterRegistry registry,
      RecordProcessorProperties processorProperties,
      ExponentialBackoff processBackoff,
      ExponentialBackoff checkpointBackoff) {
    this.registry = registry;
    this.metrics = new RecordProcessorMetrics(registry, "metrics");

    this.maxEventsBeforeCheckpoint = processorProperties.getMaxEventsBeforeCheckpoint();
    this.checkpointIntervalMs = processorProperties.getCheckpointIntervalMs();
    this.useCompression = processorProperties.isUseCompression();

    this.processBackoff = processBackoff;
    this.checkpointBackoff = checkpointBackoff;
  }

  @Override
  public void initialize(final InitializationInput input) {
    log.info(
        "initialize: Initializing record processor for shard: {} at Sequence Number: {}. "
            + "Last processed: {}, last checkpointed: {}",
        input.shardId(),
        input.extendedSequenceNumber().sequenceNumber(),
        lastProcessedRecord,
        lastCheckpointedRecord);

    this.kinesisShardId = input.shardId();
    this.perShardMillisBehindLatest =
        registry.summary("STREAMING.metrics.latency_ms", Tags.of("shard_id", input.shardId()));
  }

  @Override
  public void processRecords(final ProcessRecordsInput input) {
    // Unfortunately can't put this in initialize() since initialize runs in a separate thread
    MDC.put(SHARD_ID_KEY, this.getKinesisShardId());

    List<ParsedRecord<T>> records = parseRecords(input.records());
    Timer.Sample timer = Timer.start(registry);
    try {
      // Process records and perform all exception handling.
      processParsedRecords(records, input.checkpointer());
    } catch (AbandonWorkerException ex) {
      throw ex;
    } catch (Throwable ex) {
      throw new AbandonWorkerException("Error during processing of " + kinesisShardId, ex);
    } finally {
      timer.stop(metrics.recordBatchProcessingTime);
      // Record our running latency, or 0 if there are no records in the batch so the latency
      // doesn't appear to grow when there's no data.
      long latency = input.records().isEmpty() ? 0 : input.millisBehindLatest();
      metrics.latencyMs.record(latency);
      perShardMillisBehindLatest.record(latency);
    }
  }

  /**
   * Parse the list of {@link KinesisClientRecord}s from kinesis, returning a list of {@code T}
   * objects for procesing
   *
   * @param records Records from kinesis to decompress/parse
   * @return POJOs of the records
   */
  protected List<ParsedRecord<T>> parseRecords(List<KinesisClientRecord> records) {
    if (records.isEmpty()) {
      log.trace("parseRecords: Parsing 0 records from {}", kinesisShardId);
    } else {
      log.debug("parseRecords: Parsing {} records from {}", records.size(), kinesisShardId);
    }

    Timer.Sample timer = Timer.start(registry);

    long totalBytes = 0;
    List<ParsedRecord<T>> items = new ArrayList<>(records.size());

    for (KinesisClientRecord r : records) {
      try {
        ParsedRecord<T> parsed = parseRecord(r);
        totalBytes += parsed.getRecordSizeInBytes();
        items.add(parsed);
      } catch (RecordParseException ex) {
        // TODO: add to DLQ
        metrics.recordParsingErrors.increment();
        log.error("Error parsing record from raw bytes, skipping it!", ex);
      }
    }
    timer.stop(metrics.recordBatchParsingTime);

    metrics.recordBatchSize.record(records.size());
    metrics.recordBatchBytes.record(totalBytes);

    return items;
  }

  /**
   * Parse a single {@link KinesisClientRecord} from kineis.
   *
   * <p>Decompresses the byte buffer (if necessary) and calls {@link
   * RecordProcessorImpl#deserialize(byte[])} on the resulting byte array.
   *
   * @param record Kinesis record to parse
   * @return POJO the record contained
   * @throws RecordParseException in case of serialization error or bad data
   */
  private ParsedRecord<T> parseRecord(KinesisClientRecord record) throws RecordParseException {
    ByteBuffer data = record.data();
    int dataSize = data.limit();
    byte[] serializedData = new byte[dataSize];
    data.get(serializedData);

    metrics.recordBytes.record(dataSize);

    try {
      if (useCompression) {
        serializedData = Snappy.uncompress(serializedData);
      }
      return new ParsedRecord<>(record, deserialize(serializedData), dataSize);
    } catch (IOException ex) {
      throw new RecordParseException(serializedData, ex);
    }
  }

  /** Translate the raw {@code byte[]} into the POJO */
  protected abstract T deserialize(final byte[] serializedData) throws IOException;

  /**
   * Process a parsed kinesis record batch.
   *
   * <p>Calls {@link RecordProcessorImpl#process(Object)} for each record and handles any exceptions
   * using the {@link RecordProcessorImpl#processBackoff}
   *
   * @param records List of records to process
   * @param checkpointer Handle of the checkpointer, for marking records as fully processed.
   */
  protected void processParsedRecords(
      final List<ParsedRecord<T>> records, final RecordProcessorCheckpointer checkpointer) {
    for (ParsedRecord<T> record : records) {
      Timer.Sample timer = Timer.start(registry);

      try {
        processBackoff.execute(
            () -> {
              process(record.getItem());
              eventsSinceCheckpoint++;
              lastProcessedRecord = record.record.sequenceNumber();
            },
            metrics.recordProcessingRetries::increment);
      } catch (AbortedExecutionException ex) {
        log.error(
            "Non-retryable error during process for record {} in shard {}, skipping it!",
            record,
            kinesisShardId,
            ex);
        metrics.recordProcessingErrors.increment();
      } catch (RetriesExceededException ex) {
        log.error(
            "Retryable error reached retry limit during process for record {} in shard {}",
            record,
            kinesisShardId,
            ex);
        metrics.recordProcessingErrors.increment();
        throw new AbandonWorkerException(
            "Abandoning worker after retries exceeded during record processing", ex);
      }

      timer.stop(metrics.recordProcessingTime);
    }

    if (records.isEmpty()) {
      processCheckpoint(null, checkpointer);
    } else {
      processCheckpoint(records.get(records.size() - 1).getRecord(), checkpointer);
    }
  }

  /**
   * Process a single parsed record from kinesis.
   *
   * <p>Exceptions thrown here will trigger retries in {@link
   * RecordProcessorImpl#processParsedRecords(List, RecordProcessorCheckpointer)}.
   *
   * @param data Instance of class T
   * @return boolean true if the processing was successful and false if we had some kind of error
   */
  protected abstract boolean process(final T data) throws Exception;

  /**
   * Implement checkpoint logic here.
   *
   * <p>Called before each kinesis checkpoint operation. Exceptions here will be retried using
   * {@link RecordProcessorImpl#checkpointBackoff}
   */
  protected abstract void beforeCheckpoint() throws Exception;

  /**
   * Determine whether a checkpoint is necessary, and if so kick it off
   *
   * @param record Last processed record
   * @param checkpointer Checkpointer
   */
  private void processCheckpoint(
      final KinesisClientRecord record, final RecordProcessorCheckpointer checkpointer) {
    final long curTime = System.currentTimeMillis();
    final boolean shouldCheckpoint =
        eventsSinceCheckpoint >= maxEventsBeforeCheckpoint || curTime > nextCheckpointTimeMs;

    if (!shouldCheckpoint) {
      return;
    }

    Timer.Sample timer = Timer.start(registry);

    checkpoint(record, checkpointer);
    nextCheckpointTimeMs = System.currentTimeMillis() + checkpointIntervalMs;
    eventsSinceCheckpoint = 0;

    if (record != null) {
      lastCheckpointedRecord = record.sequenceNumber();
    }

    timer.stop(metrics.checkpointTime);
  }

  /**
   * Checkpoint with retries.
   *
   * @param record Record
   * @param checkpointer IRecordProcessorCheckpointer
   */
  private void checkpoint(
      final KinesisClientRecord record, final RecordProcessorCheckpointer checkpointer)
      throws AbandonWorkerException {
    // NOTE: this is needed for the lambda, not for multithreading
    AtomicBoolean beforeCheckpointSuccess = new AtomicBoolean(false);
    try {
      checkpointBackoff.execute(
          () -> {
            // Ensure we only call the beforeCheckpoint once.
            // checkpointer.checkpoint() can fail (eg. ThrottlingException), and in those cases we
            // do not want to
            // call beforeCheckpoint again. That work was already done successfully so running it
            // again can cause
            // errors.
            if (!beforeCheckpointSuccess.get()) {
              beforeCheckpoint();
              beforeCheckpointSuccess.set(true);
            }

            if (record != null) {
              checkpointer.checkpoint(record.sequenceNumber());
            } else {
              checkpointer.checkpoint();
            }
          },
          metrics.checkpointRetryCount::increment);
    } catch (AbortedExecutionException e) {
      log.error("checkpoint: Non-retryable exception: {}. Will not checkpoint.", e.getMessage(), e);
      metrics.checkpointErrorCount.increment();
    } catch (RetriesExceededException e) {
      log.error(
          "checkpoint: Exceeded retries on retryable exception: {}. Will not checkpoint.",
          e.getMessage(),
          e);
      throw new AbandonWorkerException(
          "Abandon worker during checkpoint due to exceeded retries", e);
    }
  }

  @Override
  public void leaseLost(LeaseLostInput leaseLostInput) {
    log.info("Lost our lease for shard {}", kinesisShardId);
    MDC.remove(SHARD_ID_KEY);
  }

  @Override
  public void shardEnded(ShardEndedInput shardEndedInput) {
    log.warn("Shard {} ended. Invoking checkpoint.", kinesisShardId);
    checkpoint(null, shardEndedInput.checkpointer());
    MDC.remove(SHARD_ID_KEY);
  }

  @Override
  public void shutdownRequested(ShutdownRequestedInput shutdownRequestedInput) {
    log.warn("Shutdown requested for shard {}. Invoking checkpoint.", kinesisShardId);
    checkpoint(null, shutdownRequestedInput.checkpointer());
    MDC.remove(SHARD_ID_KEY);
  }
}
