package com.vmware.psc.alerting.persister.kinesis.client.properties;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.UUID;
import joptsimple.internal.Strings;
import lombok.Getter;
import lombok.Setter;
import org.springframework.boot.context.properties.NestedConfigurationProperty;
import software.amazon.awssdk.services.kinesis.KinesisAsyncClient;
import software.amazon.awssdk.services.kinesis.model.GetRecordsRequest;
import software.amazon.kinesis.common.InitialPositionInStream;
import software.amazon.kinesis.common.InitialPositionInStreamExtended;
import software.amazon.kinesis.metrics.MetricsLevel;

@Getter
@Setter
public class KinesisClientProperties {

  private boolean enabled = true;

  /** Name of the kinesis stream the application will consume */
  private String streamName;
  /** Name of the KCL application (corresponds to the dynamodb table used) */
  private String applicationName;

  /**
   * Fail over time in milliseconds. A worker which does not renew it's lease within this time
   * interval will be regarded as having problems and it's shards will be assigned to other workers.
   * For applications that have a large number of shards, this may be set to a higher number to
   * reduce the number of DynamoDB IOPS required for tracking leases.
   *
   * <p>Default value: 10000L
   */
  private long leaseFailoverTimeMs = 10000L;

  /**
   * Logs a warning when kinesis tasks are stuck processing for longer that this specified time
   * limit. Tasks are allowed to continue processing beyond this time and will continue logging
   * until the task completes.
   *
   * <p>Sample log: Previous PROCESS task still pending for shard shardId-000000015590 since 78018
   * ms ago.
   */
  private long logWarningForTaskAfterMillis = 60000; // 1 minute

  /**
   * The max number of leases (shards) this worker should process. This can be useful to avoid
   * overloading (and thrashing) a worker when a host has resource constraints or during deployment.
   *
   * <p>NOTE: Setting this to a low value can cause data loss if workers are not able to pick up all
   * shards in the stream due to the max limit.
   *
   * <p>Default value: {@link Integer#MAX_VALUE}
   */
  private int maxLeasesForWorker = Integer.MAX_VALUE;

  /**
   * Call the RecordProcessor::processParsedRecords() API even if {@code GetRecords} returned an
   * empty record list
   */
  private boolean callProcessRecordsEvenForEmptyRecordList = true;

  /**
   * The value for how long the ShardConsumer should sleep if no records are returned from the call
   * to {@link KinesisAsyncClient#getRecords(GetRecordsRequest)}.
   *
   * <p>Default value: 1000L
   */
  private long idleTimeBetweenReadsMs = 1000L;

  /**
   * Max records to fetch from Kinesis in a single GetRecords call.
   *
   * <p>Default value: 10000
   */
  private int maxRecordsPerRequest = 10000;

  /** Timeout for waiting for the consumer to shut down gracefully during destroy() */
  private int shutdownWaitSeconds = 30;

  /**
   * Used to specify the position in the stream where a new application should start from. This is
   * used during initial application bootstrap (when a checkpoint doesn't exist for a shard or its
   * parents).
   */
  private InitialPositionInStream initialPositionInStream = InitialPositionInStream.TRIM_HORIZON;

  /**
   * When initialPositionInStream is AT_TIMESTAMP, this field specifies the position where the
   * application should start reading
   */
  private Date initialPositionInStreamTimestamp;

  public InitialPositionInStreamExtended initialPositionInStreamExtended() {
    if (initialPositionInStream == InitialPositionInStream.AT_TIMESTAMP) {
      return InitialPositionInStreamExtended.newInitialPositionAtTimestamp(
          initialPositionInStreamTimestamp);
    }
    return InitialPositionInStreamExtended.newInitialPosition(initialPositionInStream);
  }

  /** Level of detail to include in the kinesis metrics sent to cloudwatch */
  private MetricsLevel metricsLevel = MetricsLevel.NONE;

  @NestedConfigurationProperty
  private ClientBackoffProperties backoff = new ClientBackoffProperties();

  @NestedConfigurationProperty
  private RecordProcessorProperties recordProcessor = new RecordProcessorProperties();

  public String generateWorkerId() {
    try {
      return InetAddress.getLocalHost().getCanonicalHostName() + ":" + UUID.randomUUID();
    } catch (UnknownHostException e) {
      return "error-getting-host:" + UUID.randomUUID();
    }
  }

  public void validate() {
    checkArgument(!Strings.isNullOrEmpty(streamName), "Stream name is required");
    checkArgument(!Strings.isNullOrEmpty(applicationName), "Application name is required");
    if (initialPositionInStream == InitialPositionInStream.AT_TIMESTAMP) {
      checkNotNull(
          initialPositionInStreamTimestamp,
          "Timestamp at initial position in stream is required when initial-position-in-stream is AT_TIMESTAMP");
    }
  }
}
