package com.vmware.psc.alerting.persister.kinesis.producer;

import static com.google.common.base.Preconditions.checkNotNull;

import com.carbonblack.psc.backoff.ExponentialBackoff;
import com.carbonblack.psc.backoff.retrylocker.IAbortRetryLocker;
import com.carbonblack.psc.backoff.sleep.ISleeper;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.vmware.psc.alerting.persister.kinesis.producer.properties.KinesisProducerProperties;
import io.micrometer.core.instrument.MeterRegistry;
import io.micrometer.core.instrument.Timer;
import lombok.extern.slf4j.Slf4j;
import org.xerial.snappy.Snappy;
import software.amazon.awssdk.core.SdkBytes;
import software.amazon.awssdk.services.kinesis.KinesisAsyncClient;
import software.amazon.awssdk.services.kinesis.model.PutRecordRequest;
import software.amazon.awssdk.services.kinesis.model.PutRecordResponse;

@Slf4j
public class KinesisBasicProducer implements IKinesisProducer {
  private final KinesisAsyncClient kinesisClient;
  private final KinesisProducerProperties properties;
  private final RecordProducerMetrics metrics;
  private final MeterRegistry registry;
  private final ObjectMapper objectMapper;
  private final ExponentialBackoff addRecordBackoff;

  public KinesisBasicProducer(
      KinesisAsyncClient kinesis,
      KinesisProducerProperties properties,
      ISleeper sleeper,
      IAbortRetryLocker retryLocker,
      MeterRegistry meterRegistry,
      ObjectMapper objectMapper) {
    checkNotNull(properties.getBasic(), "Creating basic producer without required configuration");
    this.kinesisClient = kinesis;
    this.properties = properties;
    this.metrics = new RecordProducerMetrics(meterRegistry, properties.getMetricPrefix());
    this.addRecordBackoff =
        new ExponentialBackoff(properties.getBasic().getBackoff(), sleeper, retryLocker);
    this.registry = meterRegistry;
    this.objectMapper = objectMapper;
  }

  @Override
  public void addRecord(final String streamName, final Object data, final String partitionKey)
      throws Exception {
    byte[] serializedData = objectMapper.writeValueAsBytes(data);
    addRawRecord(streamName, serializedData, partitionKey);
  }

  @Override
  public void addRawRecord(final String streamName, final byte[] data, final String partitionKey)
      throws Exception {
    final Timer.Sample timer = Timer.start(registry);

    final byte[] payload = properties.isUseCompression() ? Snappy.compress(data) : data;

    int recordSize = payload.length;
    metrics.getPushSize().record(recordSize);

    addRecordBackoff.execute(
        () -> {
          PutRecordResponse putRecordResult =
              kinesisClient
                  .putRecord(
                      PutRecordRequest.builder()
                          .streamName(streamName)
                          .data(SdkBytes.fromByteArray(payload))
                          .partitionKey(partitionKey)
                          .build())
                  .get();
          log.debug(
              "Successfully pushed one record to Kinesis stream: {}, partitionKey: {}",
              streamName,
              partitionKey);

          if (properties.isEnableShardMetrics()) {
            String shardId = putRecordResult.shardId();
            metrics.getShardPushSize(shardId).record(recordSize);
          }
        },
        metrics.getPushRetries()::increment);

    timer.stop(metrics.getPushTime());
  }
}
