package com.vmware.psc.alerting.persister.kinesis.producer;

import static com.google.common.base.Preconditions.checkNotNull;

import com.amazonaws.services.kinesis.producer.KinesisProducer;
import com.amazonaws.services.kinesis.producer.KinesisProducerConfiguration;
import com.amazonaws.services.kinesis.producer.UserRecordResult;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.util.concurrent.ListenableFuture;
import com.vmware.psc.alerting.persister.kinesis.producer.properties.KinesisProducerProperties;
import io.micrometer.core.instrument.MeterRegistry;
import io.micrometer.core.instrument.Tags;
import io.micrometer.core.instrument.Timer;
import java.nio.ByteBuffer;
import java.nio.charset.StandardCharsets;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ForkJoinPool;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.DisposableBean;
import org.xerial.snappy.Snappy;

@Slf4j
public class KinesisKPLProducer implements IKinesisProducer, DisposableBean {
  private final ObjectMapper objectMapper;

  private final KinesisProducer producer;
  private final KinesisProducerProperties properties;
  private final MeterRegistry registry;
  private final RecordProducerMetrics metrics;

  private volatile boolean terminated = false;

  private static final Pattern SHARD_ID_PATTERN = Pattern.compile("shardId-\\d+");

  public KinesisKPLProducer(
      KinesisProducerConfiguration config,
      KinesisProducerProperties properties,
      MeterRegistry meterRegistry,
      ObjectMapper objectMapper) {
    checkNotNull(properties.getKpl(), "Creating kpl producer without required configuration");
    this.properties = properties;
    this.objectMapper = objectMapper;
    this.registry = meterRegistry;
    this.metrics = new RecordProducerMetrics(meterRegistry, properties.getMetricPrefix());

    config
        .setRecordMaxBufferedTime(properties.getKpl().getMaxQueueTimeMs())
        .setMetricsLevel(properties.getKpl().getMetricsLevel())
        .setLogLevel(log.isInfoEnabled() ? "info" : "warning")
        .setRateLimit(properties.getKpl().getRateLimit())
        .setMaxConnections(properties.getKpl().getMaxConnections())
        .setRecordTtl(properties.getKpl().getRecordTtl());

    if (properties.getKpl().getThreadPoolSize() > 0) {
      config.setThreadPoolSize(properties.getKpl().getThreadPoolSize());
      config.setThreadingModel(KinesisProducerConfiguration.ThreadingModel.POOLED);
    } else {
      config.setThreadingModel(KinesisProducerConfiguration.ThreadingModel.PER_REQUEST);
    }

    producer = new KinesisProducer(config);
    registry.gauge(
        metrics.getQueueSize(),
        Tags.empty(),
        producer,
        KinesisProducer::getOutstandingRecordsCount);
  }

  @Override
  public void addRecord(final String streamName, final Object data, final String partitionKey)
      throws Exception {
    byte[] serializedData = objectMapper.writeValueAsString(data).getBytes(StandardCharsets.UTF_8);
    addRawRecord(streamName, serializedData, partitionKey);
  }

  @Override
  public void addRawRecord(final String streamName, final byte[] data, final String partitionKey)
      throws Exception {
    while (producer.getOutstandingRecordsCount() > properties.getKpl().getMaxQueueSize()) {
      producerWait();
    }
    if (terminated) {
      throw new RuntimeException("addRecord called after KPL has already been terminated");
    }

    final Timer.Sample timer = Timer.start(registry);
    try {
      final ByteBuffer payload =
          ByteBuffer.wrap(properties.isUseCompression() ? Snappy.compress(data) : data);

      int recordSize = payload.remaining();
      metrics.getPushSize().record(recordSize);

      ListenableFuture<UserRecordResult> recordFuture =
          producer.addUserRecord(streamName, partitionKey, payload);

      if (properties.isEnableShardMetrics()) {
        recordFuture.addListener(
            markShardRates(recordSize, recordFuture), ForkJoinPool.commonPool());
      }

    } catch (Exception ex) {
      metrics.getErrorCounter().increment();

      if (properties.isEnableShardMetrics()) {
        Matcher matcher = SHARD_ID_PATTERN.matcher(ex.getMessage());
        if (matcher.find()) {
          String shardId = matcher.group();
          metrics.getShardErrorCount(shardId).increment();
        }
      }

      throw ex;
    } finally {
      timer.stop(metrics.getPushTime());
    }
  }

  private Runnable markShardRates(int recordSize, ListenableFuture<UserRecordResult> recordFuture) {
    return () -> {
      try {
        UserRecordResult result = recordFuture.get();
        String shardId = result.getShardId();
        metrics.getShardPushSize(shardId).record(recordSize);
      } catch (InterruptedException | ExecutionException e) {
        log.debug("Ignoring interrupt calculating shard metrics", e);
      }
    };
  }

  private void producerWait() throws InterruptedException {
    Thread.sleep(1); // TODO: improve this
    metrics.getWaitCounter().increment();
  }

  public void terminate() {
    terminated = true;
    log.warn("Closing the KPL application");
    producer.flushSync();
    producer.destroy();
  }

  public void flush() {
    producer.flushSync();
  }

  @Override
  public void destroy() {
    terminate();
  }
}
