package com.vmware.psc.alerting.persister.kinesis.producer;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRAugmentedEventBatch;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.vmware.psc.alerting.data.uds.UdsEventCreator;
import com.vmware.psc.alerting.persister.exception.PublishException;
import com.vmware.psc.alerting.persister.kinesis.utils.ProtoUtilHelper;
import com.vmware.psc.alerting.persister.metrics.PerAlertTypeCounter;
import com.vmware.psc.alerting.persister.properties.KinesisConfig;
import com.vmware.time.helper.TimeHelper;
import io.micrometer.core.instrument.MeterRegistry;
import java.util.Optional;
import java.util.UUID;
import lombok.extern.slf4j.Slf4j;
import org.jetbrains.annotations.NotNull;
import org.springframework.stereotype.Component;

@Slf4j
@Component
public class ProducerSupport {
  private IKinesisProducer kinesisProducer;
  private final KinesisConfig config;

  private final PerAlertTypeCounter alertsPublishedToUds;
  private final PerAlertTypeCounter alertsFailedToPublishToUds;
  private final PerAlertTypeCounter alertsSkippedToPublishToUds;

  public ProducerSupport(
      Optional<IKinesisProducer> kinesisProducer,
      KinesisConfig config,
      MeterRegistry meterRegistry) {
    // disabled if the producer.kpl.enabled and producer.basic.enabled properties are set to false
    kinesisProducer.ifPresent(producer -> this.kinesisProducer = producer);

    this.config = config;
    this.alertsPublishedToUds = new PerAlertTypeCounter(meterRegistry, "uds.alerts_published");
    this.alertsFailedToPublishToUds =
        new PerAlertTypeCounter(meterRegistry, "uds.alerts_failed_to_publish");
    this.alertsSkippedToPublishToUds =
        new PerAlertTypeCounter(meterRegistry, "uds.alerts_skipped_to_publish");
  }

  public void publishUaeAlert(PSCAlert.CbUaeAlert cbUaeAlert, String alertId) {
    try {

      final PSCAlert.CbUaeAlert cbUaeAlertWithDispatchTags = addCloudDispatchTags(cbUaeAlert);
      PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg eventBatchMsgWithProducerInfo =
          UdsEventCreator.convertToUds(
              cbUaeAlertWithDispatchTags,
              PSCRAugmentedEventBatch.CbBatchProducerMetadata.ProducerType.ALERT_PERSISTER,
              TimeHelper.timestampNow());

      // Publish to UDS stream
      if (kinesisProducer != null) {
        kinesisProducer.addRawRecord(
            config.getOutputStream().getStreamName(),
            eventBatchMsgWithProducerInfo.toByteArray(),
            String.format("%s-%s", UUID.randomUUID(), "proto"));

        log.debug(
            "publishUaeAlert: Successfully published event batch to stream [{}] for alertId [{}]",
            config.getOutputStream().getStreamName(),
            alertId);
        if (log.isTraceEnabled()) {
          log.trace(
              "publishUaeAlert: Published data: {}",
              ProtoUtilHelper.toString(eventBatchMsgWithProducerInfo));
        }
        alertsPublishedToUds.increment(cbUaeAlert);
      } else {
        log.debug(
            "publishUaeAlert: Kinesis publishing is disabled - "
                + "skip publishing event batch to stream [{}] for alertId [{}]",
            config.getOutputStream().getStreamName(),
            alertId);
        alertsSkippedToPublishToUds.increment(cbUaeAlert);
      }
    } catch (Exception ex) {
      alertsFailedToPublishToUds.increment(cbUaeAlert);
      throw new PublishException(
          String.format(
              "Failed to publish event batch to stream [%s] for alertId [%s]",
              config.getOutputStream().getStreamName(), alertId),
          ex);
    }
  }

  /**
   * Add the cloud dispatch tags
   *
   * @param cbUaeAlert CbUaeAlert
   * @return CbUaeAlert.
   */
  @NotNull
  private PSCAlert.CbUaeAlert addCloudDispatchTags(final PSCAlert.CbUaeAlert cbUaeAlert) {
    boolean hasAlertPersisterDispatchTag = hasAlertPersisterDispatchTag(cbUaeAlert);
    if (!hasAlertPersisterDispatchTag) {
      PSCAlert.CbUaeAlert.Builder cbUaeBuilder = PSCAlert.CbUaeAlert.newBuilder(cbUaeAlert);
      cbUaeBuilder
          .getEventBuilder()
          .getHeaderBuilder()
          .addCloudDispatchTags(
              PSCRAugmentedEventBatch.CbBatchProducerMetadata.ProducerType.ALERT_PERSISTER.name());
      return cbUaeBuilder.build();
    }
    return cbUaeAlert;
  }

  /**
   * Check dispatch tags contains ALERT_PERSISTER
   *
   * @param cbUaeAlert CbUaeAlert.
   * @return boolean.
   */
  private static boolean hasAlertPersisterDispatchTag(final PSCAlert.CbUaeAlert cbUaeAlert) {
    return Optional.ofNullable(cbUaeAlert)
        .filter(PSCAlert.CbUaeAlert::hasEvent)
        .map(PSCAlert.CbUaeAlert::getEvent)
        .filter(PSCRProto.CbEventMsg::hasHeader)
        .map(PSCRProto.CbEventMsg::getHeader)
        .map(PSCRProto.CbEventHeader::getCloudDispatchTagsList)
        .filter(
            dispatchTags ->
                dispatchTags.contains(
                    PSCRAugmentedEventBatch.CbBatchProducerMetadata.ProducerType.ALERT_PERSISTER
                        .name()))
        .isPresent();
  }
}
