package com.vmware.psc.alerting.persister.kinesis.producer;

import static com.vmware.psc.alerting.persister.kinesis.utils.MeterName.join;

import com.vmware.psc.alerting.persister.kinesis.utils.MeterName;
import io.micrometer.core.instrument.Counter;
import io.micrometer.core.instrument.DistributionSummary;
import io.micrometer.core.instrument.MeterRegistry;
import io.micrometer.core.instrument.Timer;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import lombok.Getter;

public class RecordProducerMetrics {
  private static final String WAIT_COUNT = "push_wait_count";
  private static final String ERROR_COUNT = "push_error_count";
  private static final String PUSHED_TIME = "push_time";
  private static final String PUSHED_SIZE = "push_size";
  private static final String QUEUE_SIZE = "queue_size";
  private static final String SHARD_PUSH = "shard_push";
  private static final String SHARD_PUSHED_ERROR = "shard_push_error";
  private static final String PUSH_RETRIES = "push_retries";

  private final MeterRegistry registry;
  private final String prefix;

  @Getter private final Counter waitCounter;
  @Getter private final Counter errorCounter;
  @Getter private final Timer pushTime;
  @Getter private final DistributionSummary pushSize;
  @Getter private final String queueSize;
  @Getter private final Counter pushRetries;

  private final ConcurrentMap<String, Counter> errorCountByShard;
  private final ConcurrentMap<String, DistributionSummary> pushSizeByShard;

  public RecordProducerMetrics(MeterRegistry registry, String prefix) {
    this.registry = registry;
    this.prefix = join("STREAMING", prefix);

    waitCounter = registry.counter(MeterName.join(prefix, WAIT_COUNT));
    errorCounter = registry.counter(MeterName.join(prefix, ERROR_COUNT));

    pushTime = registry.timer(MeterName.join(prefix, PUSHED_TIME));
    pushSize = registry.summary(MeterName.join(prefix, PUSHED_SIZE));
    queueSize = MeterName.join(prefix, QUEUE_SIZE);
    pushRetries = registry.counter(MeterName.join(prefix, PUSH_RETRIES));

    errorCountByShard = new ConcurrentHashMap<>();
    pushSizeByShard = new ConcurrentHashMap<>();
  }

  public Counter getShardErrorCount(String shardId) {
    return errorCountByShard.computeIfAbsent(
        shardId, id -> registry.counter(MeterName.join(prefix, SHARD_PUSHED_ERROR, shardId)));
  }

  public DistributionSummary getShardPushSize(String shardId) {
    return pushSizeByShard.computeIfAbsent(
        shardId, id -> registry.summary(MeterName.join(prefix, SHARD_PUSH, shardId)));
  }
}
