package com.vmware.psc.alerting.persister.metrics;

import java.time.Duration;
import java.time.Instant;
import java.util.concurrent.atomic.AtomicReference;
import lombok.extern.slf4j.Slf4j;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

@Slf4j
@Component
/*
 A simple logging service to increment, log and reset a count value based on a given rate.
 For better accuracy of count calculation, a time-decaying counter or circular buffer could have been used to bucket
 the counters and filter the results. However, since it has the potential to impact performance and increase
 complexity, it is better to use a single value counter and use lock-free operations to update/reset the counter.
*/
public class RunStatusLogger {
  // use lockless for thread-safety and for efficiency
  private final AtomicReference<CountMetric> countMetricRef = new AtomicReference<>();

  public RunStatusLogger() {
    countMetricRef.set(new CountMetric(0, Instant.now()));
  }

  // fixedRateString is used since this a simple/quick task to reset the metrics based on the given
  // rate.
  // The rate is expected to be in minutes and big enough that the scheduled tasks are not expected
  // to impact the thread-pool size.
  @Scheduled(
      fixedRateString = "${alert-persister.scheduled.log-run-status.fixed-rate-in-milliseconds}")
  public void logRunStatus() {
    resetCountMetric();
  }

  public void incrementCountMetric() {
    // use the lockless CAS operation for metric update
    countMetricRef.getAndUpdate(
        current -> new CountMetric(current.getCount() + 1, current.getLastUpdateTime()));
  }

  public void incrementCountMetric(int amount) {
    // use the lockless CAS operation for metric update
    countMetricRef.getAndUpdate(
        current -> new CountMetric(current.getCount() + amount, current.getLastUpdateTime()));
  }

  private void resetCountMetric() {
    var countMetric = countMetricRef.get();
    // time elapsed since last update
    var timeElapsed = Duration.between(countMetric.getLastUpdateTime(), Instant.now());

    log.info(
        "Alert-persister run-status: wrote {} alerts in the last {} minutes",
        countMetric.getCount(),
        roundHalfUpNearestTen(timeElapsed.toMinutes()));

    // reset
    countMetricRef.set(new CountMetric(0, Instant.now()));
  }

  private static long roundHalfUpNearestTen(long value) {
    return (value + 5) / 10 * 10;
  }
}
