package com.vmware.psc.alerting.persister.orgthreat;

import com.carbonblack.psc.orgkey.IOrgKeyService;
import com.carbonblack.psc.orgkey.OrgKey;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.vmware.psc.alerting.persister.exception.RetryableOrgThreatException;
import java.io.IOException;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.HashMap;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Component;
import software.amazon.awssdk.core.SdkBytes;
import software.amazon.awssdk.services.dynamodb.DynamoDbAsyncClient;
import software.amazon.awssdk.services.dynamodb.model.*;

@Profile({"aws", "localstack"})
@Component
@Slf4j
public class OrgThreatClient {
  private final DynamoDbAsyncClient dynamoClient;
  private final IOrgKeyService orgKeyService;
  private final ObjectMapper objectMapper = new ObjectMapper();
  private final Map<String, String> orgThreatTables;
  public static final String PARTITION_KEY_ATTRIBUTE = "partitionKey";
  public static final String SORT_KEY_ATTRIBUTE = "sortKey";
  public static final String VALUE_ATTRIBUTE = "objectValue";
  public static final String VERSION_ATTRIBUTE = "objectVersion";
  public static final String TTL_ATTRIBUTE = "objectTTL";

  public OrgThreatClient(
      DynamoDbAsyncClient dynamoClient,
      IOrgKeyService orgKeyService,
      @Value("#{{${alert-persister.org-threat.dynamo-tables}}}")
          Map<String, String> orgThreatTables) {
    this.dynamoClient = dynamoClient;
    this.orgKeyService = orgKeyService;
    this.orgThreatTables = orgThreatTables;

    orgThreatTables.forEach((env, table) -> log.info("org-threat-table: {}:{}", env, table));
  }

  /**
   * Lookup the orgthreat entry for a given org/threat, if one is available.
   *
   * @param orgKeyStr Org that has the potential orgthreat
   * @param threatId Threat to find
   * @return The OrgThreat, or null if none is available
   * @throws RetryableOrgThreatException Error during dynamo lookup - feel free to retry. NOTE:
   *     Dynamo's get() method internally retries indefinitely. In order to prevent a failed call
   *     from hanging forever we specify a timeout. In the event of a timeout or other exception, it
   *     is safe to retry the operation. Seeing these exceptions strongly suggests a major dynamo
   *     outage / throttling event
   */
  @Cacheable(value = "orgthreat")
  public OrgThreat getById(String orgKeyStr, String threatId) throws RetryableOrgThreatException {
    OrgKey orgKey = orgKeyService.decodeOrgKey(orgKeyStr);

    Map<String, AttributeValue> keyToGet = buildItemKey(orgKey, threatId);

    String table = getTableForOrgKey(orgKey);

    try {
      GetItemResponse item =
          dynamoClient
              .getItem(GetItemRequest.builder().key(keyToGet).tableName(table).build())
              .get(60, TimeUnit.SECONDS);

      if (item.hasItem()) {
        SdkBytes raw = item.item().get(VALUE_ATTRIBUTE).b();
        try {
          return objectMapper.readValue(raw.asInputStream(), OrgThreat.class);
        } catch (IOException e) {
          log.warn(
              "Error parsing OrgThreat(orgkey={}, threat_id={}) from dynamo: {}",
              orgKeyStr,
              threatId,
              raw.asUtf8String());
          return null;
        }
      }

      return null;
    } catch (ExecutionException e) {
      log.warn(
          "Failed to execute getById: OrgThreat(orgkey={}, threat_id={}), ex: {}",
          orgKeyStr,
          threatId,
          e.getCause().getMessage());
      return null;
    } catch (InterruptedException | TimeoutException e) {
      throw new RetryableOrgThreatException(orgKey, threatId, e);
    }
  }

  /**
   * Write an orgthreat entry for a given org/threat
   *
   * @param orgKeyStr Org that owns the orgthreat
   * @param threat orgthreat to write
   * @throws JsonProcessingException Error serializing {@link OrgThreat} to JSON
   * @throws RetryableOrgThreatException Error writing to dynamo, feel free to retry
   */
  public void writeOrgThreat(String orgKeyStr, OrgThreat threat)
      throws RetryableOrgThreatException, JsonProcessingException {
    OrgKey orgKey = orgKeyService.decodeOrgKey(orgKeyStr);
    String table = getTableForOrgKey(orgKey);

    Map<String, AttributeValue> item = buildWriteItem(orgKey, threat);
    PutItemRequest.Builder builder = PutItemRequest.builder().item(item);
    if (threat.getVersion() > 0) {
      String version = String.valueOf(threat.getVersion());
      Map<String, AttributeValue> expressionAttributeValues = new HashMap<>();
      expressionAttributeValues.put(":ver", AttributeValue.builder().n(version).build());
      builder
          .expressionAttributeValues(expressionAttributeValues)
          .conditionExpression(
              "attribute_not_exists(" + VERSION_ATTRIBUTE + ") or " + VERSION_ATTRIBUTE + "< :ver");
    }
    try {
      dynamoClient.putItem(builder.item(item).tableName(table).build()).get(60, TimeUnit.SECONDS);
    } catch (InterruptedException | ExecutionException | TimeoutException e) {
      throw new RetryableOrgThreatException(orgKey, threat.getThreatId(), e.getCause());
    }
  }

  private String getTableForOrgKey(OrgKey orgKey) {
    String orgTable = orgThreatTables.get(orgKey.getEnv());

    if (orgTable == null) {
      throw new NoSuchElementException(
          String.format(
              "Failed to find org-threat table for env: %s orgKey: %s",
              orgKey.getEnv(), orgKey.getKey()));
    }

    return orgTable;
  }

  private Map<String, AttributeValue> buildItemKey(OrgKey orgKey, String threatId) {
    String partitionKey = String.valueOf(orgKey.getOrgId());
    String sortKey = buildSortKey(orgKey.getOrgId(), threatId);

    Map<String, AttributeValue> keyToGet = new HashMap<>();
    keyToGet.put(PARTITION_KEY_ATTRIBUTE, AttributeValue.builder().s(partitionKey).build());
    keyToGet.put(SORT_KEY_ATTRIBUTE, AttributeValue.builder().s(sortKey).build());
    return keyToGet;
  }

  private Map<String, AttributeValue> buildWriteItem(OrgKey orgKey, OrgThreat orgThreat)
      throws JsonProcessingException {
    Map<String, AttributeValue> item = buildItemKey(orgKey, orgThreat.getThreatId());
    SdkBytes raw = SdkBytes.fromByteArray(objectMapper.writeValueAsBytes(orgThreat));
    String version = String.valueOf(orgThreat.getVersion());
    String ttl = String.valueOf(Instant.now().plus(365, ChronoUnit.DAYS).getEpochSecond());

    item.put(VALUE_ATTRIBUTE, AttributeValue.builder().b(raw).build());
    item.put(VERSION_ATTRIBUTE, AttributeValue.builder().n(version).build());
    item.put(TTL_ATTRIBUTE, AttributeValue.builder().n(ttl).build());

    return item;
  }

  private static String buildSortKey(long orgId, String threatId) {
    return String.format("%d-0-%s", orgId, threatId);
  }
}
