package com.vmware.psc.alerting.persister.service;

import static com.carbonblack.protocols.psc.response.PSCRProto.ThreatScore.CbThreatSource.THREAT_SOURCE_DEFENSE_ANALYTICS;
import static com.vmware.psc.alerting.persister.util.CbUaeAlertUtil.*;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.vmware.psc.alerting.persister.enums.AggregationUpdateType;
import com.vmware.psc.alerting.persister.util.CbUaeAlertUtil;
import com.vmware.time.helper.TimeHelper;
import java.time.Duration;
import java.time.Instant;
import java.util.Optional;
import java.util.function.BiPredicate;
import lombok.experimental.UtilityClass;
import lombok.extern.slf4j.Slf4j;

@Slf4j
@UtilityClass
public class AggregationService {

  /**
   * Checks for the income alert is new or income alert score against cached alert score.
   *
   * @param oldAlert from cache
   * @param newAlert from kinesis
   * @param flushInterval flush interval duration
   * @return AggregationUpdateType.
   */
  public static AggregationUpdateType findAlertUpdateType(
      final PSCAlert.CbUaeAlert oldAlert,
      final PSCAlert.CbUaeAlert newAlert,
      final Duration flushInterval) {
    if (oldAlert == null) return AggregationUpdateType.NEW;
    if (isJavaAnalyticsPrimaryAlert(oldAlert, newAlert)) {
      return AggregationUpdateType.JAVA_ANALYTICS_PRIMARY;
    }
    if (isThreatScoreIncreased(oldAlert, newAlert)) {
      return AggregationUpdateType.SCORE;
    }
    if (isAnalyticsAlertElevated(oldAlert, newAlert)) {
      return AggregationUpdateType.JAVA_ANALYTICS_ALERT_ELEVATED;
    }
    if (shouldFlushAlert(oldAlert, flushInterval)) {
      return AggregationUpdateType.FLUSH_INTERVAL;
    }
    return AggregationUpdateType.NO_CHANGE;
  }

  private static boolean isJavaAnalyticsPrimaryAlert(
      PSCAlert.CbUaeAlert oldAlert, PSCAlert.CbUaeAlert newAlert) {
    if (!isAnalyticsPrimaryAlert(newAlert)) {
      // Incoming alert is not primary
      return false;
    }

    if (!isAnalyticsPrimaryAlert(oldAlert)) {
      // The cached alert is not primary so this copy must be JAVA_ANALYTICS_PRIMARY even if it's
      // older than the cached one.
      return true;
    }

    // Only mark as JAVA_ANALYTICS_PRIMARY if our timestamps are outside the range of the cached
    // first/last event times
    return isNewEventBeforeCachedFirstEventTime(oldAlert, newAlert)
        || isNewEventAfterCachedLastEventTime(oldAlert, newAlert);
  }

  private boolean shouldFlushAlert(PSCAlert.CbUaeAlert oldAlert, Duration flushInterval) {
    final Instant instantLastFlushTime = TimeHelper.toInstant(oldAlert.getCreateTimestamp());
    final Instant flushIntervalCreateInstant = instantLastFlushTime.plus(flushInterval);
    return flushIntervalCreateInstant.isBefore(Instant.now());
  }

  private boolean isThreatScoreIncreased(
      PSCAlert.CbUaeAlert oldAlert, PSCAlert.CbUaeAlert newAlert) {
    int oldThreatScore = getThreatScore(oldAlert);
    int incomeAlertThreatScore = getThreatScore(newAlert);
    return incomeAlertThreatScore > oldThreatScore;
  }

  private boolean isAnalyticsAlertElevated(
      PSCAlert.CbUaeAlert oldAlert, PSCAlert.CbUaeAlert newAlert) {
    return newAlert.hasAlert()
        && isCbAnalytics(newAlert)
        && isJavaAnalyticsAlert(newAlert)
        && isAlertElevated(oldAlert, newAlert);
  }

  private boolean isAlertElevated(PSCAlert.CbUaeAlert oldAlert, PSCAlert.CbUaeAlert newAlert) {
    String oldAlertProducer = getAlertProducer(oldAlert);
    String newAlertProducer = getAlertProducer(newAlert);
    return newAlertProducer != null
        && !newAlertProducer.equals(oldAlertProducer)
        && newAlertProducer.endsWith("_ELEVATED");
  }

  /**
   * Check the alert is JA primary alert
   *
   * @param newAlert CbUaeAlert
   * @return boolean.
   */
  private static boolean isAnalyticsPrimaryAlert(final PSCAlert.CbUaeAlert newAlert) {
    return newAlert.hasAlert()
        && isCbAnalytics(newAlert)
        && isJavaAnalyticsAlert(newAlert)
        && isDefenseAnalyticsPrimaryEvent(newAlert);
  }

  /**
   * Check the alert is CbAnalytics.
   *
   * @param alert PSCAlert.CbUaeAlert.
   * @return boolean return the whether the alerts are Cb_Analytics.
   */
  private boolean isCbAnalytics(final PSCAlert.CbUaeAlert alert) {
    BiPredicate<PSCAlert.CbUaeAlert, PSCRProto.CbAlert.AlertType> alertTypePredicate =
        (uaeAlert, alertType) ->
            uaeAlert.hasAlert() && uaeAlert.getAlert().getAlertType().equals(alertType);
    return alertTypePredicate.test(alert, PSCRProto.CbAlert.AlertType.ALERT_TYPE_THREAT)
        || alertTypePredicate.test(alert, PSCRProto.CbAlert.AlertType.ALERT_TYPE_HASHBAN);
  }

  /**
   * Check whether the alert is DRE/Java analytics alert
   *
   * @param alert PSCAlert.CbUaeAlert.
   * @return boolean.
   */
  public static boolean isJavaAnalyticsAlert(final PSCAlert.CbUaeAlert alert) {
    return Optional.ofNullable(alert)
        .filter(PSCAlert.CbUaeAlert::hasAlert)
        .map(PSCAlert.CbUaeAlert::getAlert)
        .map(PSCRProto.CbAlert::getAlertThreatScore)
        .map(PSCRProto.ThreatScore::getThreatSource)
        .filter(THREAT_SOURCE_DEFENSE_ANALYTICS::equals)
        .isPresent();
  }

  /**
   * Check the alert has the primary event.
   *
   * @param alert PSCAlert.CbUaeAlert.
   * @return boolean.
   */
  public static boolean isDefenseAnalyticsPrimaryEvent(final PSCAlert.CbUaeAlert alert) {
    return Optional.ofNullable(alert)
        .filter(PSCAlert.CbUaeAlert::hasAlert)
        .map(PSCAlert.CbUaeAlert::getAlert)
        .map(PSCRProto.CbAlert::getDefenseAnalyticsPrimaryEvent)
        .orElse(Boolean.FALSE);
  }

  /**
   * Prepare the alert data based on the AggregateUpdateType and update the same in cache.
   *
   * @param incomingCbUaeAlert PSCAlert.CbUaeAlert.
   * @param updateType AggregationUpdateType.
   * @return PSCAlert.CbUaeAlert.Builder.
   */
  public static PSCAlert.CbUaeAlert.Builder updateIncomingAlertFromCache(
      final PSCAlert.CbUaeAlert.Builder incomingCbUaeAlert,
      final PSCAlert.CbUaeAlert cachedAlert,
      final AggregationUpdateType updateType) {
    final String docId = getAlertId(incomingCbUaeAlert);

    if (AggregationUpdateType.SCORE == updateType
        || AggregationUpdateType.JAVA_ANALYTICS_PRIMARY == updateType
        || AggregationUpdateType.FLUSH_INTERVAL == updateType
        || AggregationUpdateType.JAVA_ANALYTICS_ALERT_ELEVATED == updateType) {
      log.debug("Update Type: {} for the alert id {}", updateType, docId);
      preserveFields(cachedAlert, incomingCbUaeAlert);
    }
    incomingCbUaeAlert.setUpdateCount(0L);
    incomingCbUaeAlert.setCreateTimestamp(TimeHelper.timestampNow());
    return incomingCbUaeAlert;
  }

  /**
   * preserveFields.
   *
   * @param cachedAlert CbUaeAlert.
   * @param incomingAlert CbUaeAlert.
   * @return CbUaeAlert.Builder.
   */
  public static PSCAlert.CbUaeAlert.Builder preserveFields(
      final PSCAlert.CbUaeAlert cachedAlert, final PSCAlert.CbUaeAlert.Builder incomingAlert) {
    preserveFirstSeenTime(cachedAlert, incomingAlert);
    preserveLastEventSeenTime(cachedAlert, incomingAlert);
    preserveHighestAlertExecState(cachedAlert, incomingAlert);
    return incomingAlert;
  }

  /**
   * preserveHighestAlertExecState.
   *
   * @param cachedAlert Cached CbUaeAlert
   * @param incomingAlert Incoming CbUaeAlert.
   */
  private static void preserveHighestAlertExecState(
      PSCAlert.CbUaeAlert cachedAlert, PSCAlert.CbUaeAlert.Builder incomingAlert) {
    final PSCRProto.CbAlert.ThreatExecutionState cachedThreatExecutionState =
        CbUaeAlertUtil.getHighestAlertExecutionState(cachedAlert);
    if (cachedThreatExecutionState
        == PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_PREVENTED) {
      incomingAlert.getAlertBuilder().setHighestAlertExecutionState(cachedThreatExecutionState);
    }
  }

  /**
   * preserve some of the details from the cached alert in the incoming alert
   *
   * @param cachedAlert Cached CbUaeAlert.
   * @param incomingAlert Incoming CbUaeAlert.
   */
  public static PSCAlert.CbUaeAlert.Builder preserveDetailsFromCache(
      final PSCAlert.CbUaeAlert cachedAlert, final PSCAlert.CbUaeAlert.Builder incomingAlert) {
    // backend timestamp should always be the initial timestamp set during alert creation.
    // this can come as a different value in the incoming alert,so overwrite with cached
    // value
    incomingAlert
        .getAlertBuilder()
        .setBackendTimestamp(CbUaeAlertUtil.getBackendTimestamp(cachedAlert));
    // timestamps from new alert can be different so overwrite with cached
    // value
    incomingAlert
        .getAlertBuilder()
        .setBackendUpdateTimestamp(CbUaeAlertUtil.getBackendUpdateTimestamp(cachedAlert));
    incomingAlert
        .getAlertBuilder()
        .setUserUpdateTimestamp(cachedAlert.getAlert().getUserUpdateTimestamp());
    incomingAlert
        .getAlertBuilder()
        .setAlertFirstSeenTimestamp(cachedAlert.getAlert().getAlertFirstSeenTimestamp());
    incomingAlert
        .getAlertBuilder()
        .setAlertLastSeenTimestamp(cachedAlert.getAlert().getAlertLastSeenTimestamp());
    incomingAlert
        .getAlertBuilder()
        .setMdrAlertNotesPresent(cachedAlert.getAlert().getMdrAlertNotesPresent());
    incomingAlert
        .getAlertBuilder()
        .setAlertNotesPresent(cachedAlert.getAlert().getAlertNotesPresent());
    return incomingAlert;
  }

  /**
   * preserveLastEventSeenTime.
   *
   * @param cachedAlert Cached CbUaeAlert.
   * @param incomingAlert Incoming CbUaeAlert.
   */
  private static void preserveLastEventSeenTime(
      final PSCAlert.CbUaeAlert cachedAlert, final PSCAlert.CbUaeAlert.Builder incomingAlert) {
    boolean shouldUpdateFromCache = !isNewEventAfterCachedLastEventTime(cachedAlert, incomingAlert);
    if (shouldUpdateFromCache) {
      // last seen timestamp from new alert is before cached last seen time -- overwrite with cached
      // value
      incomingAlert.setLastEventTimestamp(CbUaeAlertUtil.getLastEventTimestamp(cachedAlert));
    }
  }

  /**
   * Check Cache have the latest last event timestamp.
   *
   * @param cachedAlert CbUaeAlert.
   * @param incomingAlert CbUaeAlertOrBuilder.
   * @return boolean.
   */
  private static boolean isNewEventAfterCachedLastEventTime(
      PSCAlert.CbUaeAlert cachedAlert, PSCAlert.CbUaeAlertOrBuilder incomingAlert) {
    final Instant cachedLastEventTime =
        TimeHelper.toInstant(CbUaeAlertUtil.getLastEventTimestamp(cachedAlert));
    final Instant incomingLastEventTime =
        TimeHelper.toInstant(CbUaeAlertUtil.getLastEventTimestamp(incomingAlert));
    return incomingLastEventTime.isAfter(cachedLastEventTime);
  }

  /**
   * preserveFirstSeenTime.
   *
   * @param cachedAlert Cached CbUaeAlert.
   * @param incomingAlert Incoming CbUaeAlert.
   */
  private static void preserveFirstSeenTime(
      PSCAlert.CbUaeAlert cachedAlert, PSCAlert.CbUaeAlert.Builder incomingAlert) {

    boolean shouldUpdateFromCache =
        !isNewEventBeforeCachedFirstEventTime(cachedAlert, incomingAlert);
    if (shouldUpdateFromCache) {
      // first seen timestamp from new alert is after cached first seen time -- overwrite with
      // cached value
      incomingAlert.setFirstEventTimestamp(CbUaeAlertUtil.getFirstEventTimestamp(cachedAlert));
    }
  }

  /**
   * Check Cache have the latest first event timestamp.
   *
   * @param cachedAlert CbUaeAlert
   * @param incomingAlert CbUaeAlert
   * @return boolean.
   */
  private static boolean isNewEventBeforeCachedFirstEventTime(
      final PSCAlert.CbUaeAlert cachedAlert, final PSCAlert.CbUaeAlertOrBuilder incomingAlert) {
    final Instant cachedFirstEventTime =
        TimeHelper.toInstant(CbUaeAlertUtil.getFirstEventTimestamp(cachedAlert));
    final Instant incomingFirstEventTime =
        TimeHelper.toInstant(CbUaeAlertUtil.getFirstEventTimestamp(incomingAlert));
    return incomingFirstEventTime.isBefore(cachedFirstEventTime);
  }

  /**
   * Update Alert Timestamp when there is no score change.
   *
   * @param cachedAlert PSCAlert.CbUaeAlert.
   * @param incomingAlert PSCAlert.CbUaeAlert.
   * @return PSCAlert.CbUaeAlert.Builder.
   */
  public static PSCAlert.CbUaeAlert.Builder updateTimestamp(
      final PSCAlert.CbUaeAlert cachedAlert, final PSCAlert.CbUaeAlert.Builder incomingAlert) {
    incomingAlert.setUpdateCount(cachedAlert.getUpdateCount() + 1);
    return preserveDetailsFromCache(cachedAlert, incomingAlert);
  }
}
