package com.vmware.psc.alerting.persister.util;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.google.protobuf.Timestamp;
import com.vmware.time.helper.TimeHelper;
import java.util.Optional;
import lombok.experimental.UtilityClass;
import org.jetbrains.annotations.NotNull;

/** Alert Util Methods. */
@UtilityClass
public class CbUaeAlertUtil {

  public static String getAlertProducer(final PSCAlert.CbUaeAlertOrBuilder alert) {
    return getCbAlert(alert).map(PSCRProto.CbAlert::getAlertProducer).orElse(null);
  }

  /**
   * Get the Alert Threat Score.
   *
   * @param alert PSCAlert.CbUaeAlert.
   * @return int.
   */
  public static int getThreatScore(final PSCAlert.CbUaeAlertOrBuilder alert) {
    return getCbAlert(alert)
        .map(PSCRProto.CbAlert::getAlertThreatScore)
        .map(PSCRProto.ThreatScore::getThreatScore)
        .orElse(0);
  }

  /**
   * use aggregate_guid if it exists, otherwise fall back to alert_id_guid
   *
   * @param alert CbUaeAlert.
   * @return String
   */
  public static String getAlertId(final PSCAlert.CbUaeAlertOrBuilder alert) {
    final Optional<PSCRProto.CbAlert> cbAlert = getCbAlert(alert);
    return cbAlert
        .filter(PSCRProto.CbAlert::hasAlertAggregation)
        .map(PSCRProto.CbAlert::getAlertAggregation)
        .map(PSCRProto.CbAlertAggregation::getAggregateGuid)
        .orElseGet(() -> cbAlert.map(PSCRProto.CbAlert::getAlertIdGuid).orElseThrow());
  }

  /**
   * Get alert Update Timestamp
   *
   * @param alert PSCAlert.CbUaeAlert.
   * @return Timestamp.
   */
  public static Timestamp getAlertUpdateTimestamp(final PSCAlert.CbUaeAlertOrBuilder alert) {
    return getCbAlert(alert)
        .map(PSCRProto.CbAlert::getAlertUpdateTimestamp)
        .orElse(TimeHelper.timestampNow());
  }

  /**
   * getUpdateCount.
   *
   * @param alert CbUaeAlert.
   * @return Long.
   */
  public static Long getUpdateCount(final PSCAlert.CbUaeAlertOrBuilder alert) {
    return Optional.ofNullable(alert).map(PSCAlert.CbUaeAlertOrBuilder::getUpdateCount).orElse(0L);
  }

  /**
   * Current Timestamp.
   *
   * @return Timestamp.
   */
  public static Timestamp getCurrentProtoBufTimeStamp() {
    return TimeHelper.timestampNow();
  }

  /**
   * getFirstEventTimestamp.
   *
   * @param alert CbUaeAlert.
   * @return Timestamp.
   */
  public static Timestamp getFirstEventTimestamp(final PSCAlert.CbUaeAlertOrBuilder alert) {
    return Optional.ofNullable(alert)
        .map(PSCAlert.CbUaeAlertOrBuilder::getFirstEventTimestamp)
        .orElseThrow();
  }

  /**
   * getLastEventTimestamp.
   *
   * @param alert CbUaeAlert.
   * @return Timestamp.
   */
  public static Timestamp getLastEventTimestamp(final PSCAlert.CbUaeAlertOrBuilder alert) {
    return Optional.ofNullable(alert)
        .map(PSCAlert.CbUaeAlertOrBuilder::getLastEventTimestamp)
        .orElseThrow();
  }

  /**
   * getBackendTimestamp.
   *
   * @param alert CbUaeAlert.
   * @return Timestamp.
   */
  public static Timestamp getBackendTimestamp(final PSCAlert.CbUaeAlertOrBuilder alert) {
    return Optional.ofNullable(alert)
        .map(PSCAlert.CbUaeAlertOrBuilder::getAlert)
        .map(PSCRProto.CbAlert::getBackendTimestamp)
        .orElse(null);
  }

  /**
   * getBackendUpdateTimestamp.
   *
   * @param alert CbUaeAlert.
   * @return Timestamp.
   */
  public static Timestamp getBackendUpdateTimestamp(final PSCAlert.CbUaeAlertOrBuilder alert) {
    return Optional.ofNullable(alert)
        .map(PSCAlert.CbUaeAlertOrBuilder::getAlert)
        .map(PSCRProto.CbAlert::getBackendUpdateTimestamp)
        .orElse(null);
  }

  /**
   * getHighestAlertExecutionState.
   *
   * @param alert CbUaeAlert.
   * @return ThreatExecutionState.
   */
  public static PSCRProto.CbAlert.ThreatExecutionState getHighestAlertExecutionState(
      final PSCAlert.CbUaeAlertOrBuilder alert) {
    return Optional.ofNullable(alert)
        .map(PSCAlert.CbUaeAlertOrBuilder::getAlert)
        .map(PSCRProto.CbAlert::getHighestAlertExecutionState)
        .orElseThrow();
  }

  @NotNull
  private static Optional<PSCRProto.CbAlert> getCbAlert(PSCAlert.CbUaeAlertOrBuilder alert) {
    return Optional.ofNullable(alert).map(PSCAlert.CbUaeAlertOrBuilder::getAlert);
  }

  /**
   * Check for the events added
   *
   * @param uaeAlert CbUaeAlert
   * @return boolean.
   */
  public static boolean updateTypeIsEventAdded(final PSCAlert.CbUaeAlertOrBuilder uaeAlert) {
    return getCbAlert(uaeAlert)
        .map(PSCRProto.CbAlert::getAlertUpdateDetails)
        .map(PSCRProto.CbAlertUpdateDetails::getUpdateTypesList)
        .filter(
            updateTypes ->
                !updateTypes.isEmpty()
                    && updateTypes.contains(
                        PSCRProto.CbAlertUpdateDetails.AlertUpdateType.EVENT_ADDED))
        .isPresent();
  }
}
