package com.vmware.psc.alerting.persister;

import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.AlertType.ALERT_TYPE_THREAT;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.CbAlertConfidence.ALERT_CONFIDENCE_LOW;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.CbAlertConfidence.ALERT_CONFIDENCE_MEDIUM;
import static com.carbonblack.protocols.psc.response.PSCRProto.ThreatScore.CbThreatSource.THREAT_SOURCE_DEFENSE_ANALYTICS;
import static com.vmware.psc.alerting.persister.PersisterRecordProcessor.CLOSED;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRAugmentedEventBatch;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.carbonblack.protocols.psc.response.PSCRProto.CbAlertOperator.CbAlertOperatorType;
import com.carbonblack.psc.orgkey.IOrgKeyService;
import com.google.common.collect.ImmutableList;
import com.vmware.psc.alerting.data.uae.UaeAlertCreator;
import com.vmware.psc.alerting.data.uds.UdsEventCreator;
import com.vmware.psc.alerting.exception.NonRetryableDecompressionException;
import com.vmware.psc.alerting.persister.cache.properties.CacheAlertAggProperties;
import com.vmware.psc.alerting.persister.cache.service.CacheService;
import com.vmware.psc.alerting.persister.enums.AggregationUpdateType;
import com.vmware.psc.alerting.persister.es.AlertWriter;
import com.vmware.psc.alerting.persister.es.notes.NotesWriter;
import com.vmware.psc.alerting.persister.exception.RetryableOrgThreatException;
import com.vmware.psc.alerting.persister.kinesis.producer.ProducerSupport;
import com.vmware.psc.alerting.persister.orgthreat.OrgThreat;
import com.vmware.psc.alerting.persister.orgthreat.OrgThreatClient;
import com.vmware.psc.alerting.persister.orgthreat.ThreatWorkflow;
import com.vmware.psc.alerting.persister.orgthreat.WorkflowState;
import com.vmware.psc.alerting.persister.util.RecordProcessingSupport;
import com.vmware.psc.alerting.sampledata.UaeAlertSamples;
import com.vmware.psc.alerting.sampledata.UdsInputSamples;
import com.vmware.time.helper.TimeHelper;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;

class PersisterRecordProcessorTest extends BaseAlertPersisterSpringBootTest {
  @Autowired IOrgKeyService orgKeyService;

  @Mock OrgThreatClient orgThreatClient;

  @Mock AlertWriter alertWriter;
  @Mock NotesWriter notesWriter;

  @Captor private ArgumentCaptor<List<PSCAlert.CbUaeAlert.Builder>> writtenAlertsCaptor;

  @Captor private ArgumentCaptor<AggregationUpdateType> updateTypeArgumentCaptor;

  @Autowired private CacheService<String, PSCAlert.CbUaeAlert> alertCacheService;

  @Autowired private RecordProcessingSupport recordProcessingSupport;

  @Autowired private ProducerSupport producerSupport;

  @Test
  void testProcessThenCheckpoint() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(false);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT1M"));
    ProducerSupport producerSupportMock = mock(ProducerSupport.class);
    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupportMock);
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.hostBasedFirewallEvent().withOrgKey(orgKey).build();

    when(orgThreatClient.getById(any(), any())).thenReturn(null);

    processor.process(batch);
    processor.beforeCheckpoint();
    verify(producerSupportMock, times(0)).publishUaeAlert(any(), any());
    verify(alertWriter)
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());
    assertEquals(orgKey, writtenAlertsCaptor.getValue().get(0).getDeviceMetadata().getOrgKey());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());
  }

  @Test
  void testOrgThreatWorkflowHandling() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(false);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT1M"));
    ProducerSupport producerSupportMock = mock(ProducerSupport.class);

    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupportMock);
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.hashbanDenyEvent().withOrgKey(orgKey).build();

    // First alert will have no OrgThreat - Expect standard OPEN state
    when(orgThreatClient.getById(any(), any())).thenReturn(null);
    processor.process(batch);

    // Second alert will have a DISMISSED OrgThreat - Expect CLOSED state
    batch = UdsInputSamples.hashbanTerminateEvent().withOrgKey(orgKey).build();
    List<String> list = new ArrayList<>();
    list.add("Juno");
    list.add("Tuna");

    OrgThreat expectedThreat =
        OrgThreat.builder()
            .tags(list)
            .notePresent(true)
            .orgId(666)
            .threatId("10511b05131a15dfc3cf399b1aa1f2930ee6aedbc62163c594482d2067c324d6")
            .workFlow(
                ThreatWorkflow.builder()
                    .state(WorkflowState.DISMISSED)
                    .remediation("FALSE_POSITIVE")
                    .changedBy("Someone")
                    .build())
            .build();
    when(orgThreatClient.getById(any(), any())).thenReturn(expectedThreat);
    processor.process(batch);

    processor.beforeCheckpoint();

    verify(producerSupportMock, times(0)).publishUaeAlert(any(), any());
    verify(alertWriter)
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(2, writtenAlertsCaptor.getValue().size());
    assertEquals(orgKey, writtenAlertsCaptor.getValue().get(0).getDeviceMetadata().getOrgKey());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());

    assertTrue(writtenAlertsCaptor.getValue().get(1).getThreatNotesPresent());
    assertEquals(list, writtenAlertsCaptor.getValue().get(1).getInvestigationTagsList());

    PSCAlert.CbUaeAlert.Builder open = writtenAlertsCaptor.getValue().get(0);
    PSCAlert.CbUaeAlert.Builder closed = writtenAlertsCaptor.getValue().get(1);

    PSCRProto.CbAlertWorkflowState openWorkflowState =
        open.getAlert().getUserStateDetails().getWorkflowState();
    assertEquals("OPEN", openWorkflowState.getStatus());
    assertEquals(CbAlertOperatorType.SYSTEM, openWorkflowState.getChangedBy().getType());
    assertEquals("ALERT_CREATION", openWorkflowState.getChangedBy().getName());

    PSCRProto.CbAlertClassification classification =
        open.getAlert().getUserStateDetails().getClassification();
    assertEquals(
        PSCRProto.CbAlertClassification.CbAlertClassificationType.ALERT_CLASSIFICATION_UNKNOWN,
        classification.getAlertClassification());
    assertEquals(CbAlertOperatorType.SYSTEM, classification.getChangedBy().getType());
    assertEquals("ALERT_CREATION", classification.getChangedBy().getName());

    PSCRProto.CbAlertWorkflowState closedWorkflowState =
        closed.getAlert().getUserStateDetails().getWorkflowState();
    assertEquals("CLOSED", closedWorkflowState.getStatus());
    assertEquals(CbAlertOperatorType.SYSTEM, closedWorkflowState.getChangedBy().getType());
    assertEquals("AUTO_DISMISSAL", closedWorkflowState.getChangedBy().getName());
  }

  @Test
  void testOrgThreatWorkflowHandlingWithClosedWorkFlow() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(false);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT1M"));
    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupport);
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);

    // prepare the closed workflow event
    PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    builder.getDeviceMetadataBuilder().setOrgKey(orgKey);
    builder
        .getAlertBuilder()
        .getUserStateDetailsBuilder()
        .getWorkflowStateBuilder()
        .setStatus(CLOSED);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsEventCreator.convertToUds(builder.build());

    when(orgThreatClient.getById(any(), any())).thenReturn(null);

    List<String> list = new ArrayList<>();
    list.add("Juno");
    list.add("Tuna");

    OrgThreat persistedOrgThreat =
        OrgThreat.builder()
            .tags(list)
            .notePresent(true)
            .orgId(666)
            .threatId("10511b05131a15dfc3cf399b1aa1f2930ee6aedbc62163c594482d2067c324d6")
            .workFlow(
                ThreatWorkflow.builder()
                    .state(WorkflowState.OPEN)
                    .remediation("FALSE_POSITIVE")
                    .changedBy("Someone")
                    .build())
            .build();
    when(orgThreatClient.getById(any(), any())).thenReturn(persistedOrgThreat);
    processor.process(batch);

    processor.beforeCheckpoint();

    verify(alertWriter)
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());
    final PSCAlert.CbUaeAlert.Builder closedAlert =
        writtenAlertsCaptor.getValue().stream()
            .findFirst()
            .orElse(PSCAlert.CbUaeAlert.newBuilder());
    assertEquals(orgKey, closedAlert.getDeviceMetadata().getOrgKey());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());

    PSCRProto.CbAlertWorkflowState closedWorkflowState =
        closedAlert.getAlert().getUserStateDetails().getWorkflowState();

    // Assert closed Workflow state is not updated
    assertEquals(CLOSED, closedWorkflowState.getStatus());
    assertTrue(closedAlert.getThreatNotesPresent());
    assertEquals(list, closedAlert.getInvestigationTagsList());
  }

  @Test
  void testProcessThenCheckpointWithSameData() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(true);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT10M"));
    ProducerSupport producerSupportMock = mock(ProducerSupport.class);
    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupportMock);
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.hashbanTerminateEvent().withOrgKey(orgKey).build();

    when(orgThreatClient.getById(any(), any())).thenReturn(null);

    processor.process(batch);

    processor.process(batch);
    processor.beforeCheckpoint();

    verify(producerSupportMock, times(1)).publishUaeAlert(any(), any());
    verify(alertWriter)
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());
    assertEquals(orgKey, writtenAlertsCaptor.getValue().get(0).getDeviceMetadata().getOrgKey());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());
  }

  @Test
  void testProcessNewAndScoreChangeAlertThenCheckpoint() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(true);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT10M"));
    ProducerSupport producerSupportMock = mock(ProducerSupport.class);
    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupportMock);
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.hostBasedFirewallEvent()
            .withOrgKey(orgKey)
            .withTriggeredAlertId("hbfw")
            .withAlertThreatScore(3)
            .build();

    when(orgThreatClient.getById(any(), any())).thenReturn(null);

    // Process alert first time.
    processor.process(batch);
    processor.beforeCheckpoint();

    verify(producerSupportMock, times(0)).publishUaeAlert(any(), any());
    verify(alertWriter, times(1))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());

    // Process alert second time with no score change.
    processor.process(batch);
    processor.beforeCheckpoint();

    verify(alertWriter, times(2))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());

    // Process alert third time with score changes.
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batchWithScoreChange =
        UdsInputSamples.hostBasedFirewallEvent()
            .withTriggeredAlertId("hbfw")
            .withOrgKey(orgKey)
            .withAlertThreatScore(5)
            .build();

    processor.process(batchWithScoreChange);
    processor.beforeCheckpoint();

    verify(alertWriter, times(4))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());
    assertEquals(orgKey, writtenAlertsCaptor.getValue().get(0).getDeviceMetadata().getOrgKey());
    assertEquals(AggregationUpdateType.SCORE, updateTypeArgumentCaptor.getValue());

    alertCacheService.deleteAll();
  }

  @Test
  void testNonPrimaryEventComesWithElevatedAlertIsUpdatingInES() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(true);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT10M"));
    ProducerSupport producerSupportMock = mock(ProducerSupport.class);
    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupportMock);

    String alertId = UUID.randomUUID().toString();
    String REASON_CODE = "R_BOX_WEB_RUN";
    String REASON_CODE_ELEVATED = "T_BOX_WEB_RUN";
    var alertTimestamp = Instant.now().minus(Duration.ofDays(31));
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.hostBasedFirewallEvent()
            .withOrgKey(orgKey)
            .withTriggeredAlertId(alertId)
            .withAlertCreationTimestamp(TimeHelper.toTimestamp(alertTimestamp))
            .backdatedTo(alertTimestamp)
            .withAlertThreatScore(5)
            .build();

    when(orgThreatClient.getById(any(), any())).thenReturn(null);

    List<PSCAlert.CbUaeAlert> alerts = UaeAlertCreator.convertAlerts(batch);
    PSCAlert.CbUaeAlert original = alerts.get(0);
    var alertThreatScoreBuilder =
        original.getAlert().getAlertThreatScore().toBuilder()
            .setThreatSource(THREAT_SOURCE_DEFENSE_ANALYTICS);
    PSCAlert.CbUaeAlert firstAlert =
        original.toBuilder()
            .setAlert(
                original.getAlert().toBuilder()
                    .setAlertConfidence(ALERT_CONFIDENCE_LOW)
                    .setAlertType(ALERT_TYPE_THREAT)
                    .setDefenseAnalyticsPrimaryEvent(true)
                    .setAlertProducer("F745EEE6-706F-3DDD-8852-85D943D8F230")
                    .setAlertThreatScore(alertThreatScoreBuilder)
                    .setLegacyV6AlertDetails(
                        original.getAlert().getLegacyV6AlertDetails().toBuilder()
                            .setThreatCause(
                                original
                                    .getAlert()
                                    .getLegacyV6AlertDetails()
                                    .getThreatCause()
                                    .toBuilder()
                                    .setReasonCode(REASON_CODE)))
                    .addAllThreatIndicators(
                        ImmutableList.of(
                            PSCRProto.CbThreatIndicator.newBuilder()
                                .setAnalyticsReasonCode(REASON_CODE)
                                .build()))
                    .setAlertThreatScore(alertThreatScoreBuilder))
            .build();
    processor.process(UdsEventCreator.convertToUds(firstAlert));
    processor.beforeCheckpoint();

    verify(producerSupportMock, times(0)).publishUaeAlert(any(), any());
    verify(alertWriter, times(1))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());

    var secondEventTimestamp = TimeHelper.toTimestamp(alertTimestamp.plus(Duration.ofMinutes(2)));
    PSCAlert.CbUaeAlert updatedAlert =
        original.toBuilder()
            .setAlert(
                original.getAlert().toBuilder()
                    .setAlertUpdateTimestamp(secondEventTimestamp)
                    .setAlertConfidence(ALERT_CONFIDENCE_MEDIUM)
                    .setAlertProducer("F745EEE6-706F-3DDD-8852-85D943D8F230_ELEVATED")
                    .setAlertType(ALERT_TYPE_THREAT)
                    .setDefenseAnalyticsPrimaryEvent(false)
                    .setAlertThreatScore(alertThreatScoreBuilder)
                    .setLegacyV6AlertDetails(
                        original.getAlert().getLegacyV6AlertDetails().toBuilder()
                            .setThreatCause(
                                original
                                    .getAlert()
                                    .getLegacyV6AlertDetails()
                                    .getThreatCause()
                                    .toBuilder()
                                    .setReasonCode(REASON_CODE_ELEVATED)))
                    .addAllThreatIndicators(
                        ImmutableList.of(
                            PSCRProto.CbThreatIndicator.newBuilder()
                                .setAnalyticsReasonCode(REASON_CODE_ELEVATED)
                                .build()))
                    .setAlertThreatScore(alertThreatScoreBuilder))
            .build();
    processor.process(UdsEventCreator.convertToUds(updatedAlert));
    processor.beforeCheckpoint();

    // Verify secondary events updated properly
    verify(alertWriter, times(3))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());
    assertEquals(
        AggregationUpdateType.JAVA_ANALYTICS_ALERT_ELEVATED, updateTypeArgumentCaptor.getValue());
    var fromEs = writtenAlertsCaptor.getValue().get(0);
    assertEquals(orgKey, fromEs.getDeviceMetadata().getOrgKey());
    assertEquals(ALERT_CONFIDENCE_MEDIUM, fromEs.getAlert().getAlertConfidence());
    assertEquals(
        "F745EEE6-706F-3DDD-8852-85D943D8F230_ELEVATED", fromEs.getAlert().getAlertProducer());
    assertEquals(
        REASON_CODE_ELEVATED,
        fromEs.getAlert().getLegacyV6AlertDetails().getThreatCause().getReasonCode());
    assertEquals(
        REASON_CODE_ELEVATED, fromEs.getAlert().getThreatIndicators(0).getAnalyticsReasonCode());
    assertEquals(REASON_CODE_ELEVATED, fromEs.getPrecomputed().getReasonCode());
    alertCacheService.deleteAll();
  }

  @Test
  void testProcessNewAndScoreChangeAlertThenCheckpointWithCacheDisabled() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(false);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT1M"));
    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupport);
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.hostBasedFirewallEvent()
            .withOrgKey(orgKey)
            .withTriggeredAlertId("hbfw")
            .withAlertThreatScore(3)
            .build();

    when(orgThreatClient.getById(any(), any())).thenReturn(null);

    // Process alert first time.
    processor.process(batch);
    processor.beforeCheckpoint();

    verify(alertWriter, times(1))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(1, writtenAlertsCaptor.getValue().size());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());

    // Process alert second time with no score change.
    processor.process(batch);
    processor.beforeCheckpoint();

    verify(alertWriter, times(2))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(2, writtenAlertsCaptor.getValue().size());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());

    // Process alert third time with score changes.
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batchWithScoreChange =
        UdsInputSamples.hostBasedFirewallEvent().withOrgKey(orgKey).withAlertThreatScore(5).build();

    processor.process(batchWithScoreChange);
    processor.beforeCheckpoint();

    verify(alertWriter, times(3))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(3, writtenAlertsCaptor.getValue().size());
    assertEquals(orgKey, writtenAlertsCaptor.getValue().get(0).getDeviceMetadata().getOrgKey());
    assertEquals(0L, writtenAlertsCaptor.getValue().get(0).getUpdateCount());
    assertEquals(0L, writtenAlertsCaptor.getValue().get(1).getUpdateCount());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());
  }

  @Test
  void testProcessNewAndNoneThenFireFlushInterval() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(true);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT5M"));
    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupport);
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.hostBasedFirewallEvent()
            .withOrgKey(orgKey)
            .withTriggeredAlertId("hbfw1")
            .withAlertThreatScore(3)
            .build();

    when(orgThreatClient.getById(any(), any())).thenReturn(null);

    // Process alert first time.
    processor.process(batch);
    processor.beforeCheckpoint();

    final PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    final Instant instant = Instant.now().minusSeconds(10 * 60);
    builder.setCreateTimestamp(TimeHelper.toTimestamp(instant));
    alertCacheService.put("hbfw1", builder.build());

    verify(alertWriter, times(1))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());

    processor.process(batch);
    processor.beforeCheckpoint();

    verify(alertWriter, times(3))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(AggregationUpdateType.FLUSH_INTERVAL, updateTypeArgumentCaptor.getValue());
  }

  @Test
  void testProcessNewAndNoneThenNotFireFlushInterval() throws Exception {
    CacheAlertAggProperties cacheAlertAggProperties = new CacheAlertAggProperties();
    cacheAlertAggProperties.setEnabled(true);
    cacheAlertAggProperties.setFlushInterval(Duration.parse("PT5M"));
    PersisterRecordProcessor processor =
        new PersisterRecordProcessor(
            recordProcessingSupport,
            alertWriter,
            notesWriter,
            orgThreatClient,
            alertCacheService,
            cacheAlertAggProperties,
            producerSupport);
    String orgKey = orgKeyService.encodeOrgKey("unittest", 666);
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.hostBasedFirewallEvent()
            .withOrgKey(orgKey)
            .withTriggeredAlertId("hbfw2")
            .withAlertThreatScore(3)
            .build();

    when(orgThreatClient.getById(any(), any())).thenReturn(null);

    // Process alert first time.
    processor.process(batch);
    processor.beforeCheckpoint();

    verify(alertWriter, times(1))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());

    final PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    builder.setCreateTimestamp(TimeHelper.timestampNow());
    alertCacheService.put("hbfw2", builder.build());

    processor.process(batch);
    processor.beforeCheckpoint();
    verify(alertWriter, times(2))
        .writeAndClearAlerts(writtenAlertsCaptor.capture(), updateTypeArgumentCaptor.capture());
    assertEquals(AggregationUpdateType.NEW, updateTypeArgumentCaptor.getValue());
  }

  @Test
  void testSkipAlertsWithMultipleThreatIndicator()
      throws RetryableOrgThreatException,
          IOException,
          NoSuchMethodException,
          InvocationTargetException,
          IllegalAccessException,
          NonRetryableDecompressionException {
    when(orgThreatClient.getById(any(), any())).thenReturn(null);
    var alertBuilder =
        UaeAlertCreator.convertAlertsAsMutable(UdsInputSamples.networkIdsEvent().build()).get(0);
    alertBuilder
        .getAlertBuilder()
        .addAllThreatIndicators(
            ImmutableList.of(
                PSCRProto.CbThreatIndicator.newBuilder().addTtpTags("SOME_TTP").build(),
                PSCRProto.CbThreatIndicator.newBuilder().addTtpTags("ANOTHER_TTP").build()));
    final Method method =
        PersisterRecordProcessor.class.getDeclaredMethod(
            "shouldSkipAlertRecord", PSCAlert.CbUaeAlert.Builder.class);
    method.setAccessible(true);
    boolean isSkipAlert = (boolean) method.invoke(null, alertBuilder);
    assertTrue(isSkipAlert);
  }

  @Test
  void testSkipAlertsWithThreatIndicator()
      throws RetryableOrgThreatException,
          IOException,
          NoSuchMethodException,
          InvocationTargetException,
          IllegalAccessException {
    when(orgThreatClient.getById(any(), any())).thenReturn(null);
    PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    Method method =
        PersisterRecordProcessor.class.getDeclaredMethod(
            "shouldSkipAlertRecord", PSCAlert.CbUaeAlert.Builder.class);
    method.setAccessible(true);
    boolean isSkipAlert = (boolean) method.invoke(null, builder);
    assertFalse(isSkipAlert);
  }
}
