package com.vmware.psc.alerting.persister.cache.listeners;

import static org.awaitility.Awaitility.await;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRAugmentedEventBatch;
import com.google.protobuf.Timestamp;
import com.vmware.psc.alerting.data.uae.UaeAlertCreator;
import com.vmware.psc.alerting.exception.NonRetryableDecompressionException;
import com.vmware.psc.alerting.persister.BaseAlertPersisterSpringBootTest;
import com.vmware.psc.alerting.persister.cache.service.CacheService;
import com.vmware.psc.alerting.persister.es.AlertEsClient;
import com.vmware.psc.alerting.persister.es.AlertWriter;
import com.vmware.psc.alerting.sampledata.UaeAlertSamples;
import com.vmware.psc.alerting.sampledata.UdsInputSamples;
import com.vmware.time.helper.TimeHelper;
import java.io.IOException;
import java.time.Duration;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import lombok.extern.slf4j.Slf4j;
import org.ehcache.event.CacheEvent;
import org.junit.jupiter.api.Test;
import org.opensearch.client.RestHighLevelClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.support.NullValue;
import org.springframework.test.util.ReflectionTestUtils;

@Slf4j
class AlertCacheListenerTest extends BaseAlertPersisterSpringBootTest {

  @Autowired CacheService<String, PSCAlert.CbUaeAlert> alertCacheService;

  @Autowired AlertEsClient alertEsClient;

  @Autowired Map<String, RestHighLevelClient> esClusterMap;

  @Test
  void onEventWithUpdateCountGreaterZero() throws IOException, NonRetryableDecompressionException {
    final Timestamp timestamp = TimeHelper.timestampNow();
    final PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.deviceControlEvent()
            .withTriggeredAlertId("cache-alert-" + System.currentTimeMillis())
            .withAlertCreationTimestamp(timestamp)
            .withIngressTimestamp(timestamp)
            .build();

    List<PSCAlert.CbUaeAlert> alerts = UaeAlertCreator.convertAlerts(batch);
    PSCAlert.CbUaeAlert cbUaeAlert = alerts.stream().findFirst().orElseThrow();
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder(cbUaeAlert);
    builder.setUpdateCount(1);
    PSCAlert.CbUaeAlert response =
        alertEsClient.findAlertById(builder.getAlert().getAlertIdGuid()).orElse(null);
    assertNull(response);

    alertCacheService.put(builder.getAlert().getAlertIdGuid(), builder.build());
    alertCacheService.delete(cbUaeAlert.getAlert().getAlertIdGuid());
    PSCAlert.CbUaeAlert writtenAlert =
        await()
            .atMost(Duration.ofSeconds(60))
            .until(
                () -> alertEsClient.findAlertById(builder.getAlert().getAlertIdGuid()).orElse(null),
                Objects::nonNull);
    assertNotNull(writtenAlert);
    assertNotNull(builder.getAlert().getAlertUpdateTimestamp());
    //    assertEquals(1, writtenAlert.getUpdateCount());
  }

  @Test
  void onEventWithUpdateCountZero() throws IOException, NonRetryableDecompressionException {
    final Timestamp timestamp = TimeHelper.timestampNow();
    final PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.deviceControlEvent()
            .withTriggeredAlertId("cache-alert-" + System.currentTimeMillis())
            .withAlertCreationTimestamp(timestamp)
            .withIngressTimestamp(timestamp)
            .build();

    List<PSCAlert.CbUaeAlert> alerts = UaeAlertCreator.convertAlerts(batch);
    PSCAlert.CbUaeAlert cbUaeAlert = alerts.stream().findFirst().orElseThrow();
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder(cbUaeAlert);
    builder.setUpdateCount(0);
    PSCAlert.CbUaeAlert response =
        alertEsClient.findAlertById(builder.getAlert().getAlertIdGuid()).orElse(null);
    assertNull(response);

    alertCacheService.put(builder.getAlert().getAlertIdGuid(), builder.build());
    alertCacheService.delete(cbUaeAlert.getAlert().getAlertIdGuid());

    await().atMost(Duration.ofSeconds(10));
    PSCAlert.CbUaeAlert writtenAlert =
        alertEsClient.findAlertById(builder.getAlert().getAlertIdGuid()).orElse(null);

    // Verify no alerts written as update count is zero
    assertNull(writtenAlert);
  }

  @Test
  void instanceOfCheckWithSpringNullValue() throws IOException {
    AlertWriter alertWriter = mock(AlertWriter.class);
    AlertCacheListener alertCacheListener = new AlertCacheListener();
    ReflectionTestUtils.setField(alertCacheListener, "alertWriter", alertWriter);
    final CacheEvent<? extends String, ? extends PSCAlert.CbUaeAlert> cacheEvent =
        mock(CacheEvent.class);
    doReturn(NullValue.INSTANCE).when(cacheEvent).getOldValue();
    alertCacheListener.onEvent(cacheEvent);
    verify(alertWriter, times(0)).writeAndClearAlerts(any(), any());
  }

  @Test
  void instanceOfCheckWithValidAlert() throws IOException {
    PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    builder.setUpdateCount(2);
    PSCAlert.CbUaeAlert alert = builder.build();
    AlertWriter alertWriter = mock(AlertWriter.class);
    AlertCacheListener alertCacheListener = new AlertCacheListener();
    ReflectionTestUtils.setField(alertCacheListener, "alertWriter", alertWriter);
    final CacheEvent<? extends String, ? extends PSCAlert.CbUaeAlert> cacheEvent =
        mock(CacheEvent.class);
    doReturn(alert).when(cacheEvent).getOldValue();
    alertCacheListener.onEvent(cacheEvent);
    verify(alertWriter, times(1)).writeAndClearAlerts(any(), any());
  }
}
