package com.vmware.psc.alerting.persister.es;

import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.AlertType.ALERT_TYPE_THREAT;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.CbAlertConfidence.ALERT_CONFIDENCE_LOW;
import static com.carbonblack.protocols.psc.response.PSCRProto.CbAlert.CbAlertConfidence.ALERT_CONFIDENCE_MEDIUM;
import static com.carbonblack.protocols.psc.response.PSCRProto.ThreatScore.CbThreatSource.THREAT_SOURCE_DEFENSE_ANALYTICS;
import static com.vmware.psc.alerting.persister.util.CbUaeAlertUtil.getAlertId;
import static org.awaitility.Awaitility.await;
import static org.junit.jupiter.api.Assertions.*;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRAugmentedEventBatch;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.google.common.collect.ImmutableList;
import com.google.protobuf.Timestamp;
import com.google.protobuf.util.JsonFormat;
import com.google.protobuf.util.Timestamps;
import com.vmware.psc.alerting.data.uae.UaeAlertCreator;
import com.vmware.psc.alerting.exception.NonRetryableDecompressionException;
import com.vmware.psc.alerting.persister.BaseAlertPersisterSpringBootTest;
import com.vmware.psc.alerting.persister.cache.service.CacheService;
import com.vmware.psc.alerting.persister.enums.AggregationUpdateType;
import com.vmware.psc.alerting.persister.exception.RetryableElasticsearchException;
import com.vmware.psc.alerting.persister.util.CbUaeAlertUtil;
import com.vmware.psc.alerting.sampledata.UaeAlertSamples;
import com.vmware.psc.alerting.sampledata.UdsInputSamples;
import com.vmware.time.helper.TimeHelper;
import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;
import org.jetbrains.annotations.NotNull;
import org.junit.jupiter.api.*;
import org.opensearch.action.admin.indices.delete.DeleteIndexRequest;
import org.opensearch.action.admin.indices.settings.put.UpdateSettingsRequest;
import org.opensearch.action.get.GetRequest;
import org.opensearch.action.get.GetResponse;
import org.opensearch.client.RequestOptions;
import org.opensearch.client.RestHighLevelClient;
import org.opensearch.client.indices.CloseIndexRequest;
import org.opensearch.client.indices.CreateIndexRequest;
import org.opensearch.common.settings.Settings;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.util.ReflectionTestUtils;

class AlertWriterTest extends BaseAlertPersisterSpringBootTest {
  @Autowired AlertWriter writer;
  @Autowired Map<String, RestHighLevelClient> esClusterMap;
  @Autowired IPartitionStrategy partitionStrategy;

  PSCAlert.CbUaeAlert readOnlyIndexAlert;
  PSCAlert.CbUaeAlert closedIndexAlert;
  RestHighLevelClient client;

  @Autowired AlertEsClient alertEsClient;

  @Autowired CacheService<String, PSCAlert.CbUaeAlert> cacheService;

  @BeforeAll
  void createIndices() throws IOException, NonRetryableDecompressionException {
    client = esClusterMap.get("notes");
    Instant readOnlyAlertTime = Instant.parse("2020-01-01T00:00:00.000Z");
    Instant closedIndexTime = Instant.parse("2020-01-02T00:00:00.000Z");
    readOnlyIndexAlert =
        UaeAlertCreator.convertAlerts(
                UdsInputSamples.deviceControlEvent()
                    .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
                    .backdatedTo(readOnlyAlertTime)
                    .withAlertCreationTimestamp(TimeHelper.toTimestamp(readOnlyAlertTime))
                    .build())
            .get(0);
    closedIndexAlert =
        UaeAlertCreator.convertAlerts(
                UdsInputSamples.securityeventEvent()
                    .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
                    .backdatedTo(closedIndexTime)
                    .withAlertCreationTimestamp(TimeHelper.toTimestamp(closedIndexTime))
                    .build())
            .get(0);

    String readOnlyIndex = partitionStrategy.getIndexForDocument(readOnlyIndexAlert);
    String closedIndex = partitionStrategy.getIndexForDocument(closedIndexAlert);

    Settings blocked = Settings.builder().put("index.blocks.read_only", true).build();
    client
        .indices()
        .create(new CreateIndexRequest(readOnlyIndex).settings(blocked), RequestOptions.DEFAULT);

    client.indices().create(new CreateIndexRequest(closedIndex), RequestOptions.DEFAULT);
    client.indices().close(new CloseIndexRequest(closedIndex), RequestOptions.DEFAULT);
  }

  @AfterAll
  void cleanIndices() throws IOException {
    String readOnlyIndex = partitionStrategy.getIndexForDocument(readOnlyIndexAlert);
    String closedIndex = partitionStrategy.getIndexForDocument(closedIndexAlert);
    Settings unblocked = Settings.builder().put("index.blocks.read_only", false).build();
    client
        .indices()
        .putSettings(
            new UpdateSettingsRequest(readOnlyIndex).settings(unblocked), RequestOptions.DEFAULT);
    client.indices().delete(new DeleteIndexRequest(readOnlyIndex), RequestOptions.DEFAULT);
    client.indices().delete(new DeleteIndexRequest(closedIndex), RequestOptions.DEFAULT);
  }

  @Test
  void testWriter() throws NonRetryableDecompressionException, IOException {
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.deviceControlEvent()
            .withTriggeredAlertId("some-alert-" + System.currentTimeMillis())
            .build();

    List<PSCAlert.CbUaeAlert> alerts = UaeAlertCreator.convertAlerts(batch);
    writer.writeAlertBatch(alerts, AggregationUpdateType.NEW);

    PSCAlert.CbUaeAlert alert = alerts.get(0);

    PSCAlert.CbUaeAlert.Builder alertBuilder = PSCAlert.CbUaeAlert.newBuilder(alert);
    // AlertUpdateType will be updated by OS
    PSCRProto.CbAlertUpdateDetails cbAlertUpdateDetails =
        PSCRProto.CbAlertUpdateDetails.newBuilder()
            .addUpdateTypes(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.ALERT_CREATED)
            .build();
    alertBuilder.getAlertBuilder().setAlertUpdateDetails(cbAlertUpdateDetails);
    copyUaeFieldsToUds(alertBuilder);
    PSCAlert.CbUaeAlert response = lookupAlertFromOpenSearch(alertBuilder.build());

    assertEquals(alertBuilder.build(), response);
  }

  @Test
  void testBatchedWriting() throws NonRetryableDecompressionException, IOException {

    List<PSCAlert.CbUaeAlert> alerts = new ArrayList<>();
    String prefix = "bulk-alert-" + System.currentTimeMillis() + "-";
    for (int i = 0; i < 25; i++) {
      PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
          UdsInputSamples.deviceControlEvent().withTriggeredAlertId(prefix + i).build();
      alerts.addAll(UaeAlertCreator.convertAlerts(batch));
    }

    // flushList will be, well, flushed by writeAndClearAlerts
    List<PSCAlert.CbUaeAlert.Builder> flushList =
        alerts.stream().map(PSCAlert.CbUaeAlert::toBuilder).collect(Collectors.toList());
    // builders will retain a copy of each builder (with timestamps updated by the flush process)
    List<PSCAlert.CbUaeAlert.Builder> builders = new ArrayList<>(flushList);

    Timestamp beforeWrites = TimeHelper.timestampNow();

    writer.writeAndClearAlerts(flushList, AggregationUpdateType.NEW);
    assertEquals(0, flushList.size());

    for (var alert : builders) {
      // Will be updated by the OS.
      PSCRProto.CbAlertUpdateDetails cbAlertUpdateDetails =
          PSCRProto.CbAlertUpdateDetails.newBuilder()
              .addUpdateTypes(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.ALERT_CREATED)
              .build();
      alert.getAlertBuilder().setAlertUpdateDetails(cbAlertUpdateDetails);
      copyUaeFieldsToUds(alert);
      PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(alert.build());
      assertEquals(alert.build(), fromEs);

      // Test timestamps are set properly
      assertTrue(Timestamps.compare(beforeWrites, alert.getCreateTimestamp()) <= 0);
      assertEquals(
          0,
          Timestamps.compare(
              alert.getAlert().getUserStateDetails().getWorkflowState().getLastUpdateTimestamp(),
              alert.getCreateTimestamp()));
    }
  }

  @Test
  void testScriptedUpdateWriterPreservesCreateTime()
      throws NonRetryableDecompressionException, IOException {
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.deviceControlEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .build();

    List<PSCAlert.CbUaeAlert> alerts = UaeAlertCreator.convertAlerts(batch);
    writer.writeAlertBatch(alerts, AggregationUpdateType.NEW);

    PSCAlert.CbUaeAlert original = alerts.get(0);

    PSCAlert.CbUaeAlert.Builder originalBuilder = PSCAlert.CbUaeAlert.newBuilder(original);
    // AlertUpdateType will be updated by OS
    PSCRProto.CbAlertUpdateDetails cbAlertUpdateDetails =
        PSCRProto.CbAlertUpdateDetails.newBuilder()
            .addUpdateTypes(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.ALERT_CREATED)
            .build();
    originalBuilder.getAlertBuilder().setAlertUpdateDetails(cbAlertUpdateDetails);
    copyUaeFieldsToUds(originalBuilder);

    PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(original);
    assertEquals(originalBuilder.build(), fromEs);

    Timestamp now = TimeHelper.timestampNow();
    // Update with new events
    PSCAlert.CbUaeAlert updated =
        original.toBuilder()
            .setCreateTimestamp(now)
            .setFirstEventTimestamp(
                TimeHelper.toTimestamp(TimeHelper.toInstant(now).minus(Duration.ofMinutes(2))))
            .setAlert(
                original.getAlert().toBuilder()
                    .setAlertThreatScore(
                        PSCRProto.ThreatScore.newBuilder()
                            .setThreatSource(
                                PSCRProto.ThreatScore.CbThreatSource.THREAT_SOURCE_PSC_RULES)
                            .build())
                    .setAlertUpdateTimestamp(now))
            .build();

    writer.writeAlertBatch(ImmutableList.of(updated), AggregationUpdateType.NEW);

    fromEs = lookupAlertFromOpenSearch(original);

    assertTrue(
        fromEs
            .getAlert()
            .getAlertUpdateDetails()
            .getUpdateTypesList()
            .contains(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.EVENT_ADDED));
    // This is an update - what we set as the new create time should become the update timestamp,
    // and the original create time should be preserved
    assertEquals(updated.getCreateTimestamp(), fromEs.getSystemUpdateTimestamp());
    assertEquals(original.getCreateTimestamp(), fromEs.getCreateTimestamp());
  }

  @Test
  void testNoopWhenUpdateTimestampIsOlder() throws NonRetryableDecompressionException, IOException {
    Instant time = Instant.now();
    Timestamp now = TimeHelper.timestampNow();

    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.deviceControlEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .withAlertUpdateTimestamp(now)
            .build();

    List<PSCAlert.CbUaeAlert> alerts = UaeAlertCreator.convertAlerts(batch);
    writer.writeAlertBatch(alerts, AggregationUpdateType.NEW);

    PSCAlert.CbUaeAlert original = alerts.get(0);

    time = time.minus(Duration.ofSeconds(100));
    Timestamp earlier = TimeHelper.toTimestamp(time);
    PSCAlert.CbUaeAlert updated =
        original.toBuilder()
            .setCreateTimestamp(now)
            .setAlert(
                original.getAlert().toBuilder()
                    .setAlertThreatScore(
                        PSCRProto.ThreatScore.newBuilder().setThreatScore(100).build())
                    .setAlertUpdateTimestamp(earlier))
            .build();

    writer.writeAlertBatch(ImmutableList.of(updated), AggregationUpdateType.NEW);

    PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(original);
    assertTrue(fromEs.hasSystemUpdateTimestamp());
    assertEquals(original.getCreateTimestamp(), fromEs.getSystemUpdateTimestamp());
    assertEquals(
        original.getAlert().getAlertUpdateTimestamp(), fromEs.getAlert().getAlertUpdateTimestamp());
    assertEquals(
        original.getAlert().getAlertThreatScore(), fromEs.getAlert().getAlertThreatScore());
  }

  @Test
  void testUpdateWhenNonPrimaryEventComesWithElevatedAlert()
      throws NonRetryableDecompressionException, IOException {
    String alertId = UUID.randomUUID().toString();
    var alertTimestamp = Instant.now().minus(Duration.ofDays(31));
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.deviceControlEvent()
            .withTriggeredAlertId(alertId)
            .withTriggeredAlertId(alertId)
            .withAlertCreationTimestamp(TimeHelper.toTimestamp(alertTimestamp))
            .backdatedTo(alertTimestamp)
            .withAlertThreatScore(5)
            .build();

    List<PSCAlert.CbUaeAlert> alerts = UaeAlertCreator.convertAlerts(batch);
    PSCAlert.CbUaeAlert original = alerts.get(0);
    var alertThreatScoreBuilder =
        original.getAlert().getAlertThreatScore().toBuilder()
            .setThreatSource(THREAT_SOURCE_DEFENSE_ANALYTICS);
    String REASON_CODE = "R_BOX_WEB_RUN";
    String REASON_CODE_ELEVATED = "T_BOX_WEB_RUN";
    PSCAlert.CbUaeAlert firstAlert =
        original.toBuilder()
            .setAlert(
                original.getAlert().toBuilder()
                    .setAlertConfidence(ALERT_CONFIDENCE_LOW)
                    .setAlertType(ALERT_TYPE_THREAT)
                    .setDefenseAnalyticsPrimaryEvent(true)
                    .setAlertProducer("F745EEE6-706F-3DDD-8852-85D943D8F230")
                    .setLegacyV6AlertDetails(
                        original.getAlert().getLegacyV6AlertDetails().toBuilder()
                            .setThreatCause(
                                original
                                    .getAlert()
                                    .getLegacyV6AlertDetails()
                                    .getThreatCause()
                                    .toBuilder()
                                    .setReasonCode(REASON_CODE)))
                    .addAllThreatIndicators(
                        ImmutableList.of(
                            PSCRProto.CbThreatIndicator.newBuilder()
                                .setAnalyticsReasonCode(REASON_CODE)
                                .build()))
                    .setAlertThreatScore(alertThreatScoreBuilder))
            .setPrecomputed(original.getPrecomputed().toBuilder().setReasonCode(REASON_CODE))
            .build();
    writer.writeAlertBatch(ImmutableList.of(firstAlert), AggregationUpdateType.NEW);

    var secondEventTimestamp = TimeHelper.toTimestamp(alertTimestamp.plus(Duration.ofMinutes(2)));
    PSCAlert.CbUaeAlert updatedAlert =
        original.toBuilder()
            .setAlert(
                original.getAlert().toBuilder()
                    .setAlertUpdateTimestamp(secondEventTimestamp)
                    .setAlertConfidence(ALERT_CONFIDENCE_MEDIUM)
                    .setAlertProducer("F745EEE6-706F-3DDD-8852-85D943D8F230_ELEVATED")
                    .setLegacyV6AlertDetails(
                        original.getAlert().getLegacyV6AlertDetails().toBuilder()
                            .setThreatCause(
                                original
                                    .getAlert()
                                    .getLegacyV6AlertDetails()
                                    .getThreatCause()
                                    .toBuilder()
                                    .setReasonCode(REASON_CODE_ELEVATED)))
                    .setAlertType(ALERT_TYPE_THREAT)
                    .setDefenseAnalyticsPrimaryEvent(false)
                    .addAllThreatIndicators(
                        ImmutableList.of(
                            PSCRProto.CbThreatIndicator.newBuilder()
                                .setAnalyticsReasonCode(REASON_CODE_ELEVATED)
                                .build()))
                    .setAlertThreatScore(alertThreatScoreBuilder))
            .setPrecomputed(
                original.getPrecomputed().toBuilder().setReasonCode(REASON_CODE_ELEVATED))
            .build();
    writer.writeAlertBatch(
        ImmutableList.of(updatedAlert), AggregationUpdateType.JAVA_ANALYTICS_ALERT_ELEVATED);

    // Verify secondary events updated properly
    PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(original);
    assertTrue(fromEs.getAlert().getDefenseAnalyticsPrimaryEvent());
    assertEquals(ALERT_CONFIDENCE_MEDIUM, fromEs.getAlert().getAlertConfidence());
    assertEquals(
        "F745EEE6-706F-3DDD-8852-85D943D8F230_ELEVATED", fromEs.getAlert().getAlertProducer());
    assertEquals(
        REASON_CODE_ELEVATED,
        fromEs.getAlert().getLegacyV6AlertDetails().getThreatCause().getReasonCode());
    assertEquals(
        REASON_CODE_ELEVATED, fromEs.getAlert().getThreatIndicators(0).getAnalyticsReasonCode());
    assertEquals(REASON_CODE_ELEVATED, fromEs.getPrecomputed().getReasonCode());
  }

  @Test
  void testNonPrimaryEventNotPersistedInOs()
      throws NonRetryableDecompressionException, IOException {
    Timestamp now = TimeHelper.timestampNow();

    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.mappedThreatProcessEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .withAlertUpdateTimestamp(now)
            .build();

    PSCAlert.CbUaeAlert.Builder b = UaeAlertCreator.convertAlerts(batch).get(0).toBuilder();
    b.getAlertBuilder().setDefenseAnalyticsPrimaryEvent(false);
    PSCAlert.CbUaeAlert original = b.build();
    List<PSCAlert.CbUaeAlert> alerts = ImmutableList.of(original);

    writer.writeAlertBatch(alerts, AggregationUpdateType.NEW);

    String index = partitionStrategy.getIndexForDocument(original);

    GetResponse doc =
        client.get(
            new GetRequest(index, original.getAlert().getAlertIdGuid()), RequestOptions.DEFAULT);
    assertFalse(
        doc.isExists(),
        "Doc should not exist. If it doesn't that means there was an error writing it. Non Primary JA alerts will not saved");
  }

  @Test
  void testEventNotUpdatedForLegacyAnalyticsNonPrimaryEvent()
      throws NonRetryableDecompressionException, IOException {
    Instant time = Instant.now();
    Timestamp now = TimeHelper.timestampNow();

    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.mappedThreatProcessEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .withAlertUpdateTimestamp(now)
            .build();

    PSCAlert.CbUaeAlert.Builder b = UaeAlertCreator.convertAlerts(batch).get(0).toBuilder();
    PSCAlert.CbUaeAlert original = b.build();
    List<PSCAlert.CbUaeAlert> alerts = ImmutableList.of(original);

    writer.writeAlertBatch(alerts, AggregationUpdateType.NEW);

    time = time.plus(Duration.ofSeconds(100));
    Timestamp later = TimeHelper.toTimestamp(time);
    // Ensure the update is considered
    b.getAlertBuilder().setAlertUpdateTimestamp(later);
    // This timestamp should get put into the update timestamp during upsert
    b.setCreateTimestamp(now);
    // This earlier timestamp should become the new first activity timestamp during upsert
    b.getFirstEventTimestampBuilder().setSeconds(b.getFirstEventTimestamp().getSeconds() - 100);
    // This updated document does not have an event, but rather than clear the event the upsert
    // should preserve the original event as defenseAnalyticsPrimaryEvent is false
    b.clearEvent();
    b.getAlertBuilder().setDefenseAnalyticsPrimaryEvent(false);

    writer.writeAlertBatch(ImmutableList.of(b.build()), AggregationUpdateType.NEW);

    PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(original);

    assertEquals(b.getCreateTimestamp(), fromEs.getSystemUpdateTimestamp());
    assertEquals(
        original.getAlert().getAlertUpdateTimestamp(), fromEs.getAlert().getAlertUpdateTimestamp());
    assertEquals(b.getFirstEventTimestamp(), fromEs.getFirstEventTimestamp());
    assertEquals(original.getEvent(), fromEs.getEvent());
  }

  @Test
  void testEventUpdateForLegacyAnalyticsPrimaryEvent()
      throws NonRetryableDecompressionException, IOException {
    Timestamp now = TimeHelper.timestampNow();

    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.deviceControlEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .withAlertUpdateTimestamp(now)
            .withAlertCreationTimestamp(now)
            .build();

    PSCAlert.CbUaeAlert.Builder originalEventWithPrimaryFalse =
        UaeAlertCreator.convertAlerts(batch).get(0).toBuilder();
    originalEventWithPrimaryFalse
        .getAlertBuilder()
        .setAlertThreatScore(
            PSCRProto.ThreatScore.newBuilder()
                .setThreatScore(3)
                .setThreatSource(
                    PSCRProto.ThreatScore.CbThreatSource.THREAT_SOURCE_DEFENSE_ANALYTICS)
                .build());
    originalEventWithPrimaryFalse.getAlertBuilder().setDefenseAnalyticsPrimaryEvent(false);
    originalEventWithPrimaryFalse
        .getEventBuilder()
        .getHeaderBuilder()
        .setSensorAction(PSCRProto.CbEventHeader.CbSensorAction.ACTION_ALLOW);
    PSCAlert.CbUaeAlert original = originalEventWithPrimaryFalse.build();
    List<PSCAlert.CbUaeAlert> alerts = ImmutableList.of(original);

    writer.writeAlertBatch(alerts, AggregationUpdateType.NEW);

    // Using same alert for the date & time and updating the sensor action, defense primary true.
    PSCAlert.CbUaeAlert.Builder updateEventWithPrimaryTrue =
        PSCAlert.CbUaeAlert.newBuilder(originalEventWithPrimaryFalse.build());

    updateEventWithPrimaryTrue
        .getEventBuilder()
        .getHeaderBuilder()
        .setSensorAction(PSCRProto.CbEventHeader.CbSensorAction.ACTION_TERMINATE);

    // Update the event with Defense Primary True
    updateEventWithPrimaryTrue.getAlertBuilder().setDefenseAnalyticsPrimaryEvent(true);

    writer.writeAlertBatch(
        ImmutableList.of(updateEventWithPrimaryTrue.build()), AggregationUpdateType.NEW);

    PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(original);

    assertTrue(
        fromEs
            .getAlert()
            .getAlertUpdateDetails()
            .getUpdateTypesList()
            .contains(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.ALERT_CREATED));

    assertTrue(fromEs.getAlert().getDefenseAnalyticsPrimaryEvent());
    assertEquals(
        PSCRProto.CbEventHeader.CbSensorAction.ACTION_TERMINATE,
        fromEs.getEvent().getHeader().getSensorAction());
  }

  @Test
  void testThreatCauseChange() throws NonRetryableDecompressionException, IOException {
    Timestamp now = TimeHelper.timestampNow();

    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.mappedThreatProcessEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .withAlertUpdateTimestamp(now)
            .withAlertCreationTimestamp(now)
            .build();

    PSCAlert.CbUaeAlert.Builder originalEvent =
        UaeAlertCreator.convertAlerts(batch).get(0).toBuilder();

    // Update the alert with the highest alert execution state and assert its prevented
    originalEvent
        .getAlertBuilder()
        .setHighestAlertExecutionState(
            PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_PREVENTED);

    originalEvent
        .getEventBuilder()
        .getHeaderBuilder()
        .setSensorAction(PSCRProto.CbEventHeader.CbSensorAction.ACTION_ALLOW);
    PSCAlert.CbUaeAlert original = originalEvent.build();
    List<PSCAlert.CbUaeAlert> alerts = ImmutableList.of(original);

    writer.writeAlertBatch(alerts, AggregationUpdateType.NEW);

    // Using same alert for the date & time and updating the sensor action, defense primary true.
    PSCAlert.CbUaeAlert.Builder updateEventWithPrimaryTrue =
        PSCAlert.CbUaeAlert.newBuilder(originalEvent.build());

    updateEventWithPrimaryTrue
        .getAlertBuilder()
        .setHighestAlertExecutionState(
            PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_ALLOWED);

    updateEventWithPrimaryTrue
        .getEventBuilder()
        .getHeaderBuilder()
        .setSensorAction(PSCRProto.CbEventHeader.CbSensorAction.ACTION_TERMINATE);

    Instant time =
        TimeHelper.toInstant(updateEventWithPrimaryTrue.getAlertBuilder().getAlertUpdateTimestamp())
            .plus(Duration.ofSeconds(300));

    updateEventWithPrimaryTrue
        .getAlertBuilder()
        .setAlertUpdateTimestamp(TimeHelper.toTimestamp(time));

    writer.writeAlertBatch(
        ImmutableList.of(updateEventWithPrimaryTrue.build()), AggregationUpdateType.NEW);

    PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(original);

    assertTrue(
        fromEs
            .getAlert()
            .getAlertUpdateDetails()
            .getUpdateTypesList()
            .contains(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.THREAT_CAUSE_CHANGED));

    assertTrue(fromEs.getAlert().getDefenseAnalyticsPrimaryEvent());
    assertEquals(
        PSCRProto.CbEventHeader.CbSensorAction.ACTION_TERMINATE,
        fromEs.getEvent().getHeader().getSensorAction());
    // assert to check highest threat execution state is prevented
    assertEquals(
        PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_PREVENTED,
        fromEs.getAlert().getHighestAlertExecutionState());
  }

  @Test
  void testMissingDeviceIpDoesntBreakIndexing()
      throws NonRetryableDecompressionException, IOException {
    Timestamp now = TimeHelper.timestampNow();

    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.deviceControlEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .withAlertUpdateTimestamp(now)
            .build();

    PSCAlert.CbUaeAlert.Builder alert = UaeAlertCreator.convertAlerts(batch).get(0).toBuilder();
    alert.getDeviceMetadataBuilder().clearSelfReportedDeviceIp();

    writer.writeAlertBatch(ImmutableList.of(alert.build()), AggregationUpdateType.NEW);

    String index = partitionStrategy.getIndexForDocument(alert);

    // AlertUpdateType will be updated by OS
    PSCRProto.CbAlertUpdateDetails cbAlertUpdateDetails =
        PSCRProto.CbAlertUpdateDetails.newBuilder()
            .addUpdateTypes(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.ALERT_CREATED)
            .build();
    alert.getAlertBuilder().setAlertUpdateDetails(cbAlertUpdateDetails);

    copyUaeFieldsToUds(alert);

    GetResponse doc =
        client.get(
            new GetRequest(index, alert.getAlert().getAlertIdGuid()), RequestOptions.DEFAULT);
    assertTrue(
        doc.isExists(),
        "Doc should exist. If it doesn't that means there was an error writing it.");
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder();
    JsonFormat.parser().merge(doc.getSourceAsString(), builder);
    assertEquals(alert.build(), builder.build());
  }

  @Test
  void testReadOnlyIndexThrowsDuringWrite() {
    String readOnlyIndex = partitionStrategy.getIndexForDocument(readOnlyIndexAlert);

    RetryableElasticsearchException ex =
        assertThrows(
            RetryableElasticsearchException.class,
            () ->
                writer.writeAlertBatch(
                    Collections.singletonList(readOnlyIndexAlert), AggregationUpdateType.NEW));

    assertEquals(readOnlyIndex, ex.getIndex());
  }

  @Test
  void testClosedIndexThrowsDuringWrite() {
    String closedIndex = partitionStrategy.getIndexForDocument(closedIndexAlert);

    RetryableElasticsearchException ex =
        assertThrows(
            RetryableElasticsearchException.class,
            () ->
                writer.writeAlertBatch(
                    Collections.singletonList(closedIndexAlert), AggregationUpdateType.NEW));

    assertEquals(closedIndex, ex.getIndex());
  }

  @Test
  void testBatchWithTheSameAlertIdMultipleTimes()
      throws NonRetryableDecompressionException, IOException {
    Instant time = Instant.now();
    Timestamp now = TimeHelper.timestampNow();

    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.mappedThreatProcessEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .build();

    var b = UaeAlertCreator.convertAlerts(batch).get(0).toBuilder();
    b.getAlertBuilder()
        .setAlertCreationTimestamp(now)
        .clearAlertUpdateTimestamp()
        .getLegacyV6AlertDetailsBuilder()
        .setSensorAction("");

    var original = b.build();

    List<PSCAlert.CbUaeAlert> alerts = ImmutableList.of(original);

    writer.writeAlertBatch(alerts, AggregationUpdateType.SCORE);

    var updateOneTime = time.plus(Duration.ofSeconds(100));
    var updateTwoTime = time.plus(Duration.ofSeconds(200));

    var updateOne = original.toBuilder();
    updateOne
        .getAlertBuilder()
        .setAlertUpdateTimestamp(TimeHelper.toTimestamp(updateOneTime))
        .getLegacyV6AlertDetailsBuilder()
        .setSensorAction("from update one");

    var updateTwo = original.toBuilder();
    updateTwo
        .getAlertBuilder()
        .setAlertUpdateTimestamp(TimeHelper.toTimestamp(updateTwoTime))
        .getLegacyV6AlertDetailsBuilder()
        .setSensorAction("from update two");

    writer.writeAlertBatch(
        ImmutableList.of(updateTwo.build(), updateOne.build()), AggregationUpdateType.NEW);

    PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(original);

    assertTrue(
        fromEs
            .getAlert()
            .getAlertUpdateDetails()
            .getUpdateTypesList()
            .contains(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.THREAT_CAUSE_CHANGED));

    assertEquals(
        updateTwo.getAlert().getAlertUpdateTimestamp(),
        fromEs.getAlert().getAlertUpdateTimestamp());

    assertEquals("from update two", fromEs.getAlert().getLegacyV6AlertDetails().getSensorAction());
  }

  @Test
  void testNewAlerts() throws IOException, NonRetryableDecompressionException {
    final var alert =
        UaeAlertCreator.convertAlertsAsMutable(
                UdsInputSamples.hostBasedFirewallEvent()
                    .withRandomIdAndHash()
                    .withAlertCreationTimestamp(TimeHelper.timestampNow())
                    .build())
            .get(0);
    final List<PSCAlert.CbUaeAlert.Builder> cbUaeAlerts = new ArrayList<>();
    cbUaeAlerts.add(alert);
    ReflectionTestUtils.setField(writer, "isAggregationCacheEnabled", false);
    cacheService.deleteAll();
    // Publishing new alert
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.NEW);

    // verification
    final PSCAlert.CbUaeAlert response = lookupAlertFromOpenSearch(alert.build());
    assertNotNull(response);
    assertNull(cacheService.get(getAlertId(response)));

    // Even new alerts should show system_update_timestamp as the same as create_timestamp
    // This allows more logical sorting on last_update_time
    assertEquals(response.getCreateTimestamp(), response.getSystemUpdateTimestamp());
    assertEquals(
        response.getCreateTimestamp(),
        response.getAlert().getUserStateDetails().getWorkflowState().getLastUpdateTimestamp());

    assertTrue(
        response
            .getAlert()
            .getAlertUpdateDetails()
            .getUpdateTypesList()
            .contains(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.ALERT_CREATED));
  }

  @Test
  void testFirstEventGreatAndLastEventTimeLesser() throws IOException {
    final PSCAlert.CbUaeAlert original = UaeAlertSamples.hbfwAlert();
    final List<PSCAlert.CbUaeAlert.Builder> cbUaeAlerts = new ArrayList<>();
    final Timestamp timestamp = CbUaeAlertUtil.getCurrentProtoBufTimeStamp();
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder(original);
    String id = UUID.randomUUID().toString();
    builder
        .getAlertBuilder()
        .setAlertIdGuid(id)
        .setAlertCreationTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    builder.getAlertBuilder().getAlertAggregationBuilder().setAggregateGuid(id);
    cbUaeAlerts.add(builder);
    ReflectionTestUtils.setField(writer, "isAggregationCacheEnabled", true);
    cacheService.deleteAll();
    // Publishing the original new alert
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.NEW);
    assertNotNull(cacheService.get(getAlertId(builder)));

    // Update Alert with event times
    builder = PSCAlert.CbUaeAlert.newBuilder(original);
    builder
        .getAlertBuilder()
        .setAlertIdGuid(id)
        .setAlertCreationTimestamp(timestamp)
        .setAlertUpdateTimestamp(timestamp);
    builder.getAlertBuilder().getAlertAggregationBuilder().setAggregateGuid(id);

    var firstEventTime =
        TimeHelper.toInstant(builder.getFirstEventTimestamp()).plus(Duration.ofSeconds(100));
    var lastEventTime =
        TimeHelper.toInstant(builder.getLastEventTimestamp()).minus(Duration.ofSeconds(100));
    builder.setFirstEventTimestamp(TimeHelper.toTimestamp(firstEventTime));
    builder.setLastEventTimestamp(TimeHelper.toTimestamp(lastEventTime));
    builder.setUpdateCount(5L);
    builder.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(8);
    cbUaeAlerts.add(builder);
    cacheService.deleteAll();
    // Publishing with updated event timestamp with in the range
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.SCORE);

    // verification
    final PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(builder);
    assertNotNull(fromEs);
    assertNotNull(cacheService.get(getAlertId(builder)));

    // check first and last event timestamp is not updated
    assertEquals(original.getFirstEventTimestamp(), fromEs.getFirstEventTimestamp());
    assertEquals(original.getLastEventTimestamp(), fromEs.getLastEventTimestamp());
  }

  @Test
  void testEventsAdded() throws IOException {
    PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    String id = UUID.randomUUID().toString();
    builder.getAlertBuilder().setAlertIdGuid(id);
    builder.getAlertBuilder().getAlertAggregationBuilder().setAggregateGuid(id);
    final PSCAlert.CbUaeAlert alert = builder.build();
    final List<PSCAlert.CbUaeAlert.Builder> cbUaeAlerts = new ArrayList<>();
    final Timestamp timestamp = CbUaeAlertUtil.getCurrentProtoBufTimeStamp();
    PSCAlert.CbUaeAlert.Builder originalBuilder = PSCAlert.CbUaeAlert.newBuilder(alert);
    originalBuilder
        .getAlertBuilder()
        .setAlertCreationTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    cbUaeAlerts.add(originalBuilder);
    ReflectionTestUtils.setField(writer, "isAggregationCacheEnabled", true);
    cacheService.deleteAll();

    // Writing to the OS
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.NEW);
    assertNotNull(cacheService.get(getAlertId(alert)));

    // Prepare the alert for second pass
    // Update eventTimestamp
    PSCAlert.CbUaeAlert.Builder updatedBuilder = PSCAlert.CbUaeAlert.newBuilder(alert);
    updatedBuilder
        .getAlertBuilder()
        .setAlertCreationTimestamp(timestamp)
        .setAlertUpdateTimestamp(timestamp);

    var firstEventTime =
        TimeHelper.toInstant(updatedBuilder.getFirstEventTimestamp())
            .minus(Duration.ofSeconds(100));
    var lastEventTime =
        TimeHelper.toInstant(updatedBuilder.getLastEventTimestamp()).plus(Duration.ofSeconds(100));
    updatedBuilder.setFirstEventTimestamp(TimeHelper.toTimestamp(firstEventTime));
    updatedBuilder.setLastEventTimestamp(TimeHelper.toTimestamp(lastEventTime));
    updatedBuilder.setUpdateCount(5L);
    updatedBuilder.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(8);
    cbUaeAlerts.add(updatedBuilder);
    cacheService.deleteAll();

    // Writing to the OS
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.SCORE);

    // Verification
    final PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(originalBuilder);
    assertNotNull(fromEs);
    assertNotNull(cacheService.get(getAlertId(alert)));
    assertTrue(
        fromEs
            .getAlert()
            .getAlertUpdateDetails()
            .getUpdateTypesList()
            .contains(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.EVENT_ADDED));

    // AlertUpdateType will be updated by OS
    PSCRProto.CbAlertUpdateDetails cbAlertUpdateDetails =
        PSCRProto.CbAlertUpdateDetails.newBuilder()
            .addUpdateTypes(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.EVENT_ADDED)
            .build();
    copyUaeFieldsToUds(updatedBuilder);
    updatedBuilder.getAlertBuilder().setAlertUpdateDetails(cbAlertUpdateDetails);

    assertEquals(originalBuilder.getEvent(), fromEs.getEvent());
  }

  @Test
  void testScoreUpdateAlerts() throws IOException, NonRetryableDecompressionException {
    Timestamp now = TimeHelper.timestampNow();

    // Preparation
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg batch =
        UdsInputSamples.mappedThreatProcessEvent()
            .withTriggeredAlertId("test-alert-" + System.currentTimeMillis())
            .withAlertUpdateTimestamp(now)
            .withAlertCreationTimestamp(now)
            .build();

    PSCAlert.CbUaeAlert.Builder originalEventWithPrimaryTrue =
        UaeAlertCreator.convertAlerts(batch).get(0).toBuilder();
    originalEventWithPrimaryTrue
        .getEventBuilder()
        .getHeaderBuilder()
        .setSensorAction(PSCRProto.CbEventHeader.CbSensorAction.ACTION_ALLOW);
    PSCAlert.CbUaeAlert original = originalEventWithPrimaryTrue.build();
    final List<PSCAlert.CbUaeAlert.Builder> cbUaeAlerts = new ArrayList<>();
    final Timestamp timestamp = CbUaeAlertUtil.getCurrentProtoBufTimeStamp();
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder(original);
    builder
        .getAlertBuilder()
        .setAlertCreationTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    cbUaeAlerts.add(builder);
    ReflectionTestUtils.setField(writer, "isAggregationCacheEnabled", true);
    cacheService.deleteAll();

    // Publish to OS
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.NEW);
    assertNotNull(cacheService.get(getAlertId(original)));

    // Update Alert with new Score change
    builder = PSCAlert.CbUaeAlert.newBuilder(original);
    builder
        .getAlertBuilder()
        .setAlertCreationTimestamp(timestamp)
        .setAlertUpdateTimestamp(timestamp);
    builder.setUpdateCount(5L);
    builder.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(8);
    builder.getAlertBuilder().setAlertUpdateTimestamp(TimeHelper.timestampNow());
    cbUaeAlerts.add(builder);
    cacheService.deleteAll();
    // Publishing for updated alert
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.SCORE);

    // Verification
    final PSCAlert.CbUaeAlert response = lookupAlertFromOpenSearch(original);
    assertNotNull(response);
    assertNotNull(cacheService.get(getAlertId(original)));
    assertTrue(
        response
            .getAlert()
            .getAlertUpdateDetails()
            .getUpdateTypesList()
            .contains(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.THREAT_SCORE_INCREASED));

    assertEquals(8, response.getAlert().getAlertThreatScore().getThreatScore());
    assertEquals(2, response.getAlert().getAlertUpdateDetails().getOldThreatScore());
  }

  @Test
  void testTimestampUpdateAlertsAndRemovedFromCache() throws IOException {
    final PSCAlert.CbUaeAlert alert = UaeAlertSamples.hbfwAlert();
    final List<PSCAlert.CbUaeAlert.Builder> cbUaeAlerts = new ArrayList<>();
    Timestamp timestamp = CbUaeAlertUtil.getCurrentProtoBufTimeStamp();
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder(alert);
    String id = "test-" + System.currentTimeMillis();
    builder.getAlertBuilder().setAlertIdGuid(id);
    builder.getAlertBuilder().getAlertAggregationBuilder().setAggregateGuid(id);
    builder.getAlertBuilder().setAlertCreationTimestamp(timestamp);
    builder.setUpdateCount(5L);
    cbUaeAlerts.add(builder);
    ReflectionTestUtils.setField(writer, "isAggregationCacheEnabled", true);
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.NEW);

    // Build Alert update the timestamp and count
    builder = PSCAlert.CbUaeAlert.newBuilder(builder.build());
    builder
        .getAlertBuilder()
        .setAlertCreationTimestamp(timestamp)
        .setAlertUpdateTimestamp(timestamp);
    builder.setLastEventTimestamp(timestamp);
    builder.setUpdateCount(5L);
    cbUaeAlerts.add(builder);
    cacheService.deleteAll();
    // Publish
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.EVICT);

    // Verification
    final PSCAlert.CbUaeAlert response =
        await()
            .atMost(Duration.ofSeconds(60))
            .pollInterval(Duration.ofMillis(250))
            .until(() -> alertEsClient.findAlertById(id).orElse(null), Objects::nonNull);
    assertNotNull(response);
    assertNull(cacheService.get(getAlertId(alert)));
    assertEquals(1, response.getUpdateCount());
    assertEquals(timestamp, response.getLastEventTimestamp());
  }

  @Test
  void testUpdateCount() throws IOException {
    final PSCAlert.CbUaeAlert alert = UaeAlertSamples.hbfwAlert();
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder(alert);
    final Timestamp timestamp = TimeHelper.timestampNow();
    builder
        .getAlertBuilder()
        .setAlertUpdateTimestamp(timestamp)
        .setAlertIdGuid("dong-noop-test-" + UUID.randomUUID());

    final List<PSCAlert.CbUaeAlert.Builder> cbUaeAlerts = new ArrayList<>();
    cbUaeAlerts.add(builder);
    PSCAlert.CbUaeAlert beforeCachedData = cacheService.get(getAlertId(builder));
    assertNull(beforeCachedData);

    // Publishing it
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.NEW);

    // write alerts with the latest event timestamp
    builder.getAlertBuilder().setAlertUpdateTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    builder.setLastEventTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    cbUaeAlerts.clear();
    cbUaeAlerts.add(builder);
    cacheService.delete(getAlertId(builder));
    // Publishing the updated
    writer.writeAndClearAlerts(cbUaeAlerts, AggregationUpdateType.NEW);

    // verification
    PSCAlert.CbUaeAlert fromEs = lookupAlertFromOpenSearch(builder);
    assertNotNull(fromEs);
    assertEquals(1, fromEs.getUpdateCount());
  }

  /**
   * Copy Uae Fields to UDS. copy happening in alert-upsert(ES) painless script so modifying the
   * expected results to reflect the same.
   *
   * @param uaeAlertBuilder PSCAlert.CbUaeAlert.
   */
  private static void copyUaeFieldsToUds(final PSCAlert.CbUaeAlert.Builder uaeAlertBuilder) {
    final PSCRProto.CbAlert.Builder udsAlertBuilder = uaeAlertBuilder.getAlertBuilder();
    if (uaeAlertBuilder.hasCreateTimestamp()) {
      udsAlertBuilder.setBackendTimestamp(uaeAlertBuilder.getCreateTimestamp());
      uaeAlertBuilder.setSystemUpdateTimestamp(uaeAlertBuilder.getCreateTimestamp());
    }
    if (uaeAlertBuilder.hasSystemUpdateTimestamp()) {
      udsAlertBuilder.setBackendUpdateTimestamp(uaeAlertBuilder.getSystemUpdateTimestamp());
    }
    if (uaeAlertBuilder.hasUserUpdateTimestamp()) {
      udsAlertBuilder.setUserUpdateTimestamp(uaeAlertBuilder.getUserUpdateTimestamp());
    }
    udsAlertBuilder.setAlertNotesPresent(uaeAlertBuilder.getAlertNotesPresent());
    udsAlertBuilder.setMdrAlertNotesPresent(uaeAlertBuilder.getMdrAlertNotesPresent());
  }

  @NotNull
  private PSCAlert.CbUaeAlert lookupAlertFromOpenSearch(PSCAlert.CbUaeAlertOrBuilder original)
      throws IOException {
    final String index = partitionStrategy.getIndexForDocument(original);
    GetResponse doc =
        client.get(
            new GetRequest(index, original.getAlert().getAlertIdGuid()), RequestOptions.DEFAULT);
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder();
    JsonFormat.parser().merge(doc.getSourceAsString(), builder);
    return builder.build();
  }
}
