package com.vmware.psc.alerting.persister.kinesis.producer;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.*;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRAugmentedEventBatch;
import com.vmware.psc.alerting.data.uae.UaeAlertCreator;
import com.vmware.psc.alerting.persister.exception.PublishException;
import com.vmware.psc.alerting.persister.kinesis.producer.properties.KinesisOutputStreamProperties;
import com.vmware.psc.alerting.persister.metrics.PerAlertTypeCounter;
import com.vmware.psc.alerting.persister.properties.KinesisConfig;
import com.vmware.psc.alerting.persister.util.CbUaeAlertUtil;
import com.vmware.psc.alerting.sampledata.UaeAlertSamples;
import io.micrometer.core.instrument.Counter;
import io.micrometer.core.instrument.MeterRegistry;
import java.util.Optional;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.test.util.ReflectionTestUtils;

class ProducerSupportTest {

  @Captor private ArgumentCaptor<byte[]> captorBatch;

  @Mock private IKinesisProducer iKinesisProducer;

  @Mock private KinesisConfig kinesisConfiguration;

  @Mock private KinesisOutputStreamProperties kinesisOutputStreamProperties;

  @Mock private MeterRegistry meterRegistry;

  @Mock private Counter counter;

  private ProducerSupport producerSupport;

  @Mock private PerAlertTypeCounter alertsPublishedToUds;
  @Mock private PerAlertTypeCounter alertsFailedToPublishToUds;
  @Mock private PerAlertTypeCounter alertsSkippedToPublishToUds;

  @BeforeEach
  public void setup() {
    MockitoAnnotations.openMocks(this);
    when(meterRegistry.counter(anyString())).thenReturn(counter);
    when(meterRegistry.counter(anyString(), any(Iterable.class))).thenReturn(counter);

    producerSupport =
        new ProducerSupport(Optional.of(iKinesisProducer), kinesisConfiguration, meterRegistry);

    ReflectionTestUtils.setField(producerSupport, "alertsPublishedToUds", alertsPublishedToUds);
    ReflectionTestUtils.setField(
        producerSupport, "alertsFailedToPublishToUds", alertsFailedToPublishToUds);
    ReflectionTestUtils.setField(
        producerSupport, "alertsSkippedToPublishToUds", alertsSkippedToPublishToUds);
  }

  @Test
  void publishEventBatchWithProducerType() throws Exception {
    when(kinesisConfiguration.getOutputStream()).thenReturn(kinesisOutputStreamProperties);
    when(kinesisOutputStreamProperties.getStreamName()).thenReturn("test");

    PSCAlert.CbUaeAlert alert = UaeAlertSamples.watchlistIngressAlert();

    producerSupport.publishUaeAlert(alert, CbUaeAlertUtil.getAlertId(alert));

    verify(iKinesisProducer).addRawRecord(any(), captorBatch.capture(), anyString());
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg responseBatch =
        PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg.parseFrom(captorBatch.getValue());
    assertNotNull(responseBatch);
    assertEquals(
        PSCRAugmentedEventBatch.CbBatchProducerMetadata.ProducerType.ALERT_PERSISTER,
        responseBatch.getHeader().getProducerMetadata().getProducerType());

    // assert cloud dispatch tag present
    PSCAlert.CbUaeAlert uaeAlert =
        UaeAlertCreator.convertAlerts(responseBatch).stream().findFirst().orElseThrow();
    assertTrue(
        uaeAlert
            .getEvent()
            .getHeader()
            .getCloudDispatchTagsList()
            .contains(
                PSCRAugmentedEventBatch.CbBatchProducerMetadata.ProducerType.ALERT_PERSISTER
                    .name()));

    verify(alertsPublishedToUds).increment(any(PSCAlert.CbUaeAlert.class));

    verify(alertsFailedToPublishToUds, never()).increment(any(PSCAlert.CbUaeAlert.class));
    verify(alertsSkippedToPublishToUds, never()).increment(any(PSCAlert.CbUaeAlert.class));
  }

  @Test
  void testDispatchTagNotPresent() throws Exception {
    when(kinesisConfiguration.getOutputStream()).thenReturn(kinesisOutputStreamProperties);
    when(kinesisOutputStreamProperties.getStreamName()).thenReturn("test");

    PSCAlert.CbUaeAlert alert = UaeAlertSamples.watchlistIngressAlert();
    PSCAlert.CbUaeAlert.Builder builder = PSCAlert.CbUaeAlert.newBuilder(alert);
    builder
        .getEventBuilder()
        .getHeaderBuilder()
        .addCloudDispatchTags(
            PSCRAugmentedEventBatch.CbBatchProducerMetadata.ProducerType.ALERT_PERSISTER.name());

    producerSupport.publishUaeAlert(builder.build(), CbUaeAlertUtil.getAlertId(alert));

    verify(iKinesisProducer).addRawRecord(any(), captorBatch.capture(), anyString());
    PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg responseBatch =
        PSCRAugmentedEventBatch.CbAugmentedEventBatchMsg.parseFrom(captorBatch.getValue());
    assertNotNull(responseBatch);
    assertEquals(
        PSCRAugmentedEventBatch.CbBatchProducerMetadata.ProducerType.ALERT_PERSISTER,
        responseBatch.getHeader().getProducerMetadata().getProducerType());

    // assert cloud dispatch tag present
    PSCAlert.CbUaeAlert uaeAlert =
        UaeAlertCreator.convertAlerts(responseBatch).stream().findFirst().orElseThrow();
    assertTrue(
        uaeAlert
            .getEvent()
            .getHeader()
            .getCloudDispatchTagsList()
            .contains(
                PSCRAugmentedEventBatch.CbBatchProducerMetadata.ProducerType.ALERT_PERSISTER
                    .name()));
  }

  @Test
  void testPublishToUdsFailed() throws Exception {
    when(kinesisConfiguration.getOutputStream()).thenReturn(kinesisOutputStreamProperties);
    when(kinesisOutputStreamProperties.getStreamName()).thenReturn("test");
    doThrow(new RuntimeException()).when(iKinesisProducer).addRawRecord(any(), any(), any());

    PSCAlert.CbUaeAlert alert = UaeAlertSamples.watchlistIngressAlert();
    for (int i = 0; i < 5; i++) {
      assertThrows(
          PublishException.class,
          () -> {
            producerSupport.publishUaeAlert(alert, CbUaeAlertUtil.getAlertId(alert));
          });
    }

    verify(alertsFailedToPublishToUds, times(5)).increment(any(PSCAlert.CbUaeAlert.class));

    verify(alertsPublishedToUds, never()).increment(any(PSCAlert.CbUaeAlert.class));
    verify(alertsSkippedToPublishToUds, never()).increment(any(PSCAlert.CbUaeAlert.class));
  }

  @Test
  void testPublishToUdsSkipped() throws Exception {
    when(kinesisConfiguration.getOutputStream()).thenReturn(kinesisOutputStreamProperties);
    when(kinesisOutputStreamProperties.getStreamName()).thenReturn("test");
    ReflectionTestUtils.setField(producerSupport, "kinesisProducer", null);

    PSCAlert.CbUaeAlert alert = UaeAlertSamples.watchlistIngressAlert();
    for (int i = 0; i < 5; i++) {
      producerSupport.publishUaeAlert(alert, CbUaeAlertUtil.getAlertId(alert));
    }

    verify(iKinesisProducer, never()).addRawRecord(any(), any(), any());
    verify(alertsSkippedToPublishToUds, times(5)).increment(any(PSCAlert.CbUaeAlert.class));

    verify(alertsPublishedToUds, never()).increment(any(PSCAlert.CbUaeAlert.class));
    verify(alertsFailedToPublishToUds, never()).increment(any(PSCAlert.CbUaeAlert.class));
  }
}
