package com.vmware.psc.alerting.persister.service;

import static org.junit.jupiter.api.Assertions.*;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.google.protobuf.Timestamp;
import com.vmware.psc.alerting.data.uae.UaeAlertCreator;
import com.vmware.psc.alerting.exception.NonRetryableDecompressionException;
import com.vmware.psc.alerting.persister.enums.AggregationUpdateType;
import com.vmware.psc.alerting.persister.util.CbUaeAlertUtil;
import com.vmware.psc.alerting.sampledata.UaeAlertSamples;
import com.vmware.psc.alerting.sampledata.UdsInputSampleBuilder;
import com.vmware.psc.alerting.sampledata.UdsInputSamples;
import com.vmware.time.helper.TimeHelper;
import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;

class AggregationServiceTest {

  @Test
  void aggregationUpdateTypeForNewAlert() throws IOException {
    final PSCAlert.CbUaeAlert newAlert = UaeAlertSamples.hbfwAlert();
    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(null, newAlert, Duration.ZERO);
    assertNotNull(updateType);
    assertEquals(AggregationUpdateType.NEW, updateType);
  }

  @Test
  void aggregationUpdateTypeForScoreChangeAlert() throws IOException {
    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    incomingAlert.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(10);
    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    cachedAlertBuilder.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(5);
    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlertBuilder.build(), incomingAlert.build(), Duration.ZERO);
    assertNotNull(updateType);
    assertEquals(AggregationUpdateType.SCORE, updateType);
  }

  @Test
  void aggregationUpdateTypeFlushIntervalAlert() throws IOException {
    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    incomingAlert.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(5);
    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    cachedAlertBuilder.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(10);
    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlertBuilder.build(), incomingAlert.build(), Duration.ZERO);
    assertNotNull(updateType);
    assertEquals(AggregationUpdateType.FLUSH_INTERVAL, updateType);
  }

  @Test
  void aggregationUpdateTypeNoneAlert() throws IOException {
    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    incomingAlert.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(5);
    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    cachedAlertBuilder.getAlertBuilder().getAlertThreatScoreBuilder().setThreatScore(5);
    cachedAlertBuilder.setCreateTimestamp(TimeHelper.timestampNow());
    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlertBuilder.build(), incomingAlert.build(), Duration.ofMinutes(5));
    assertNotNull(updateType);
    assertEquals(AggregationUpdateType.NO_CHANGE, updateType);
  }

  @Test
  void aggregationUpdateTypeJavaAnalyticsAlertWithLatestLastEventTime()
      throws IOException, NonRetryableDecompressionException {
    UdsInputSampleBuilder builder = UdsInputSamples.mappedThreatProcessEvent();
    var uaeAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertCreator.convertAlerts(builder.build()).get(0));
    Instant lastEventTime = TimeHelper.toInstant(uaeAlertBuilder.getLastEventTimestamp());

    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(uaeAlertBuilder.build());
    incomingAlert.setLastEventTimestamp(
        TimeHelper.toTimestamp(lastEventTime.plus(Duration.ofMinutes(5))));
    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(uaeAlertBuilder.build());

    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlertBuilder.build(), incomingAlert.build(), Duration.ZERO);
    assertNotNull(updateType);
    assertEquals(AggregationUpdateType.JAVA_ANALYTICS_PRIMARY, updateType);
  }

  @Test
  void aggregationUpdateTypeJavaAnalyticsAlertWithEarlierFirstEventTime()
      throws IOException, NonRetryableDecompressionException {
    UdsInputSampleBuilder builder = UdsInputSamples.mappedThreatProcessEvent();
    var uaeAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertCreator.convertAlerts(builder.build()).get(0));
    Instant firstEventTime = TimeHelper.toInstant(uaeAlertBuilder.getFirstEventTimestamp());

    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(uaeAlertBuilder.build());
    incomingAlert.setFirstEventTimestamp(
        TimeHelper.toTimestamp(firstEventTime.minus(Duration.ofMinutes(5))));
    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(uaeAlertBuilder.build());

    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlertBuilder.build(), incomingAlert.build(), Duration.ZERO);
    assertNotNull(updateType);
    assertEquals(AggregationUpdateType.JAVA_ANALYTICS_PRIMARY, updateType);
  }

  @Test
  void aggregationUpdateTypeJAAlertWithOlderFirstEventTimeAndEarlierLastEventTime()
      throws IOException, NonRetryableDecompressionException {
    UdsInputSampleBuilder builder = UdsInputSamples.mappedThreatProcessEvent();
    var uaeAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertCreator.convertAlerts(builder.build()).get(0));
    Instant lastEventTime = TimeHelper.toInstant(uaeAlertBuilder.getLastEventTimestamp());
    Instant firstEventTime = TimeHelper.toInstant(uaeAlertBuilder.getFirstEventTimestamp());

    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(uaeAlertBuilder.build());
    incomingAlert.setFirstEventTimestamp(
        TimeHelper.toTimestamp(firstEventTime.plus(Duration.ofMinutes(5))));
    incomingAlert.setLastEventTimestamp(
        TimeHelper.toTimestamp(lastEventTime.minus(Duration.ofMinutes(5))));
    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(uaeAlertBuilder.build());

    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlertBuilder.build(), incomingAlert.build(), Duration.ZERO);
    assertNotNull(updateType);
    assertNotEquals(AggregationUpdateType.JAVA_ANALYTICS_PRIMARY, updateType);
  }

  @Test
  @DisplayName(
      "Test AggregationUpdateType is JAVA_ANALYTICS_PRIMARY when the cached alert is non-primary and the cached alert is older than the incoming alert")
  void testUpdateJavaAnalyticsPrimaryForOlderNonPrimaryCache()
      throws IOException, NonRetryableDecompressionException {
    UdsInputSampleBuilder builder = UdsInputSamples.mappedThreatProcessEvent();
    var cachedAlertBuilder = UaeAlertCreator.convertAlertsAsMutable(builder.build()).get(0);
    Instant lastEventTime = TimeHelper.toInstant(cachedAlertBuilder.getLastEventTimestamp());

    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(cachedAlertBuilder.build());

    // Incoming alert is newer than cached alert and is primary
    incomingAlert.setLastEventTimestamp(
        TimeHelper.toTimestamp(lastEventTime.plus(Duration.ofMinutes(5))));
    incomingAlert.getAlertBuilder().setDefenseAnalyticsPrimaryEvent(true);

    // Cached alert is not the primary event
    cachedAlertBuilder.getAlertBuilder().setDefenseAnalyticsPrimaryEvent(false);

    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlertBuilder.build(), incomingAlert.build(), Duration.ofMinutes(10));

    assertEquals(AggregationUpdateType.JAVA_ANALYTICS_PRIMARY, updateType);
  }

  @Test
  @DisplayName(
      "Test AggregationUpdateType is JAVA_ANALYTICS_PRIMARY when the cached alert is non-primary and the cached alert is newer than the incoming alert")
  void testUpdateJavaAnalyticsPrimaryForNewerNonPrimaryCache()
      throws IOException, NonRetryableDecompressionException {
    UdsInputSampleBuilder builder = UdsInputSamples.mappedThreatProcessEvent();
    var cachedAlertBuilder = UaeAlertCreator.convertAlertsAsMutable(builder.build()).get(0);
    Instant lastEventTime = TimeHelper.toInstant(cachedAlertBuilder.getLastEventTimestamp());

    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(cachedAlertBuilder.build());

    // Incoming alert is older than cached alert and is primary
    incomingAlert.setLastEventTimestamp(
        TimeHelper.toTimestamp(lastEventTime.minus(Duration.ofMinutes(5))));
    incomingAlert.getAlertBuilder().setDefenseAnalyticsPrimaryEvent(true);

    // Cached alert is not the primary event
    cachedAlertBuilder.getAlertBuilder().setDefenseAnalyticsPrimaryEvent(false);

    AggregationUpdateType updateType =
        AggregationService.findAlertUpdateType(
            cachedAlertBuilder.build(), incomingAlert.build(), Duration.ofMinutes(10));

    assertEquals(AggregationUpdateType.JAVA_ANALYTICS_PRIMARY, updateType);
  }

  @Test
  void prepareDataForNewAlert() throws IOException {
    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    PSCAlert.CbUaeAlert.Builder builder =
        AggregationService.updateIncomingAlertFromCache(
            incomingAlert, null, AggregationUpdateType.NEW);
    assertNotNull(builder);
  }

  @Test
  void prepareDataForScoreChangeAlert() throws IOException {
    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    cachedAlertBuilder.setUpdateCount(25L);

    PSCAlert.CbUaeAlert.Builder builder =
        AggregationService.updateIncomingAlertFromCache(
            incomingAlert, cachedAlertBuilder.build(), AggregationUpdateType.SCORE);
    assertNotNull(builder);
    assertEquals(incomingAlert.getFirstEventTimestamp(), builder.getFirstEventTimestamp());
    assertEquals(incomingAlert.getLastEventTimestamp(), builder.getLastEventTimestamp());
  }

  @Test
  void testScoreChangeAndEventTimestampIsLatestInCache() throws IOException {
    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    cachedAlertBuilder.setUpdateCount(25L);
    cachedAlertBuilder.setLastEventTimestamp(TimeHelper.timestampNow());
    cachedAlertBuilder.setFirstEventTimestamp(
        TimeHelper.toTimestamp(
            TimeHelper.toInstant(cachedAlertBuilder.getFirstEventTimestamp())
                .minus(Duration.ofMinutes(2))));

    PSCAlert.CbUaeAlert.Builder builder =
        AggregationService.updateIncomingAlertFromCache(
            incomingAlert, cachedAlertBuilder.build(), AggregationUpdateType.SCORE);
    assertNotNull(builder);
    assertEquals(cachedAlertBuilder.getFirstEventTimestamp(), builder.getFirstEventTimestamp());
    assertEquals(cachedAlertBuilder.getLastEventTimestamp(), builder.getLastEventTimestamp());
  }

  @Test
  void prepareDataForNoneAlert() throws IOException {
    final PSCAlert.CbUaeAlert.Builder incomingAlert =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    final PSCAlert.CbUaeAlert.Builder cachedAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    cachedAlertBuilder.setUpdateCount(25L);

    PSCAlert.CbUaeAlert.Builder builder =
        AggregationService.updateIncomingAlertFromCache(
            incomingAlert, cachedAlertBuilder.build(), AggregationUpdateType.SCORE);
    assertNotNull(builder);
    assertEquals(cachedAlertBuilder.getFirstEventTimestamp(), builder.getFirstEventTimestamp());
  }

  @Test
  void testUpdateAlertTimestamp() throws IOException {
    final PSCAlert.CbUaeAlert.Builder alertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    alertBuilder.setUpdateCount(10L);
    final Timestamp timestamp = CbUaeAlertUtil.getCurrentProtoBufTimeStamp();
    alertBuilder.getAlertBuilder().setAlertUpdateTimestamp(timestamp);
    final PSCAlert.CbUaeAlert.Builder responseBuilder =
        AggregationService.updateTimestamp(alertBuilder.build(), alertBuilder);
    assertNotNull(responseBuilder);
    // App handles updateCount to track the aggregation but that is not get stored in OS
    //  OS upsert will handle this logic
    assertEquals(11L, responseBuilder.getUpdateCount());
    assertEquals(timestamp, responseBuilder.getAlert().getAlertUpdateTimestamp());
  }

  @Test
  void testPreserveFieldsFirstSeenTimeWithCachedTimeOldest() throws IOException {

    final PSCAlert.CbUaeAlert.Builder incomingAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    PSCAlert.CbUaeAlert cachedAlert = incomingAlertBuilder.build();
    incomingAlertBuilder.setFirstEventTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    PSCAlert.CbUaeAlert.Builder response =
        AggregationService.preserveFields(cachedAlert, incomingAlertBuilder);
    assertNotNull(response);
    assertNotNull(response.getFirstEventTimestamp());
    assertEquals(cachedAlert.getFirstEventTimestamp(), response.getFirstEventTimestamp());
  }

  @Test
  void testPreserveFieldsFirstSeenTimeWithCachedTimeLatest() throws IOException {

    final PSCAlert.CbUaeAlert.Builder alertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    alertBuilder.setFirstEventTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    PSCAlert.CbUaeAlert cachedAlert = alertBuilder.build();

    final PSCAlert.CbUaeAlert.Builder incomingAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    PSCAlert.CbUaeAlert.Builder response =
        AggregationService.preserveFields(cachedAlert, incomingAlertBuilder);
    assertNotNull(response);
    assertNotNull(response.getFirstEventTimestamp());
    assertEquals(incomingAlertBuilder.getFirstEventTimestamp(), response.getFirstEventTimestamp());
  }

  @Test
  void testPreserveFieldsLatestSeenTimeWithCachedTimeEarliest() throws IOException {

    final PSCAlert.CbUaeAlert.Builder incomingAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    PSCAlert.CbUaeAlert cachedAlert = incomingAlertBuilder.build();
    incomingAlertBuilder.setLastEventTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    PSCAlert.CbUaeAlert.Builder response =
        AggregationService.preserveFields(cachedAlert, incomingAlertBuilder);
    assertNotNull(response);
    assertNotNull(response.getLastEventTimestamp());
    assertEquals(incomingAlertBuilder.getLastEventTimestamp(), response.getLastEventTimestamp());
  }

  @Test
  void testPreserveFieldsLastSeenTimeWithCachedTimeOldest() throws IOException {

    final PSCAlert.CbUaeAlert.Builder alertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    alertBuilder.setLastEventTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    PSCAlert.CbUaeAlert cachedAlert = alertBuilder.build();

    final PSCAlert.CbUaeAlert.Builder incomingAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    PSCAlert.CbUaeAlert.Builder response =
        AggregationService.preserveFields(cachedAlert, incomingAlertBuilder);
    assertNotNull(response);
    assertNotNull(response.getLastEventTimestamp());
    assertEquals(cachedAlert.getLastEventTimestamp(), response.getLastEventTimestamp());
  }

  @Test
  void testPreserveFieldsWithHighestAlertThreatStateAndCacheHasPrevented() throws IOException {

    final PSCAlert.CbUaeAlert.Builder alertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    alertBuilder
        .getAlertBuilder()
        .setHighestAlertExecutionState(
            PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_PREVENTED);
    alertBuilder.setLastEventTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    PSCAlert.CbUaeAlert cachedAlert = alertBuilder.build();

    final PSCAlert.CbUaeAlert.Builder incomingAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    incomingAlertBuilder
        .getAlertBuilder()
        .setHighestAlertExecutionState(
            PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_UNKNOWN);

    PSCAlert.CbUaeAlert.Builder response =
        AggregationService.preserveFields(cachedAlert, incomingAlertBuilder);
    assertNotNull(response);
    assertNotNull(response.getAlert().getHighestAlertExecutionState());
    assertEquals(
        PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_PREVENTED,
        response.getAlert().getHighestAlertExecutionState());
  }

  @Test
  void testPreserveFieldsWithHighestAlertThreatStateAndCacheNotHasPrevented() throws IOException {

    final PSCAlert.CbUaeAlert.Builder alertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    alertBuilder
        .getAlertBuilder()
        .setHighestAlertExecutionState(
            PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_UNKNOWN);
    alertBuilder.setLastEventTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    PSCAlert.CbUaeAlert cachedAlert = alertBuilder.build();

    final PSCAlert.CbUaeAlert.Builder incomingAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    incomingAlertBuilder
        .getAlertBuilder()
        .setHighestAlertExecutionState(
            PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_PREVENTED);

    PSCAlert.CbUaeAlert.Builder response =
        AggregationService.preserveFields(cachedAlert, incomingAlertBuilder);
    assertNotNull(response);
    assertNotNull(response.getAlert().getHighestAlertExecutionState());
    assertEquals(
        PSCRProto.CbAlert.ThreatExecutionState.THREAT_EXECUTION_PREVENTED,
        response.getAlert().getHighestAlertExecutionState());
  }

  @Test
  void testPreserveFieldsBackendTimestampWithCachedTime() throws IOException {

    final PSCAlert.CbUaeAlert.Builder alertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    alertBuilder
        .getAlertBuilder()
        .setBackendTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    PSCAlert.CbUaeAlert cachedAlert = alertBuilder.build();
    final PSCAlert.CbUaeAlert.Builder incomingAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());

    PSCAlert.CbUaeAlert.Builder response =
        AggregationService.preserveDetailsFromCache(cachedAlert, incomingAlertBuilder);
    assertNotNull(response);
    assertNotNull(response.getAlert().getBackendTimestamp());
    assertEquals(
        cachedAlert.getAlert().getBackendTimestamp(), response.getAlert().getBackendTimestamp());
  }

  @Test
  void testPreserveFieldsWithCachedTime() throws IOException {
    final PSCAlert.CbUaeAlert.Builder alertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    alertBuilder
        .getAlertBuilder()
        .setBackendUpdateTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    alertBuilder
        .getAlertBuilder()
        .setAlertFirstSeenTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    alertBuilder
        .getAlertBuilder()
        .setAlertLastSeenTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    alertBuilder
        .getAlertBuilder()
        .setUserUpdateTimestamp(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
    alertBuilder.getAlertBuilder().setAlertNotesPresent(true);
    alertBuilder.getAlertBuilder().setMdrAlertNotesPresent(true);
    PSCAlert.CbUaeAlert cachedAlert = alertBuilder.build();

    final PSCAlert.CbUaeAlert.Builder incomingAlertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    PSCAlert.CbUaeAlert.Builder response =
        AggregationService.preserveDetailsFromCache(cachedAlert, incomingAlertBuilder);
    assertNotNull(response);
    assertEquals(
        cachedAlert.getAlert().getBackendUpdateTimestamp(),
        response.getAlert().getBackendUpdateTimestamp());
    assertEquals(
        cachedAlert.getAlert().getAlertFirstSeenTimestamp(),
        response.getAlert().getAlertFirstSeenTimestamp());
    assertEquals(
        cachedAlert.getAlert().getAlertLastSeenTimestamp(),
        response.getAlert().getAlertLastSeenTimestamp());
    assertEquals(
        cachedAlert.getAlert().getAlertNotesPresent(), response.getAlert().getAlertNotesPresent());
    assertEquals(
        cachedAlert.getAlert().getMdrAlertNotesPresent(),
        response.getAlert().getMdrAlertNotesPresent());
    assertEquals(
        cachedAlert.getAlert().getUserUpdateTimestamp(),
        response.getAlert().getUserUpdateTimestamp());
  }
}
