package com.vmware.psc.alerting.persister.util;

import static com.vmware.psc.alerting.persister.util.CbUaeAlertUtil.*;
import static org.junit.jupiter.api.Assertions.*;

import com.carbonblack.protocols.psc.response.PSCAlert;
import com.carbonblack.protocols.psc.response.PSCRProto;
import com.google.protobuf.Timestamp;
import com.vmware.psc.alerting.sampledata.UaeAlertSamples;
import java.io.IOException;
import org.junit.jupiter.api.Test;

class CbUaeAlertUtilTest {

  @Test
  void testThreatScoreWithValidData() throws IOException {
    int score = getThreatScore(UaeAlertSamples.hbfwAlert());
    assertEquals(4, score);
  }

  @Test
  void testThreatScoreFromBuilderWithValidData() throws IOException {
    int score = getThreatScore(PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert()));
    assertEquals(4, score);
  }

  @Test
  void testThreatScoreFromBuilderWithNullData() {
    int score = getThreatScore(null);
    assertEquals(0, score);
  }

  @Test
  void getAlertIdWithGuId() throws IOException {
    final String alertId = getAlertId(UaeAlertSamples.hbfwAlert());
    assertNotNull(alertId);
    assertEquals("4fd8c8bf-2545-11ed-9bd0-000c298a7d8a", alertId);
  }

  @Test
  void getAlertIdFromBuilderWithGuId() throws IOException {
    final String alertId = getAlertId(PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert()));
    assertNotNull(alertId);
    assertEquals("4fd8c8bf-2545-11ed-9bd0-000c298a7d8a", alertId);
  }

  @Test
  void getAlertIdWithAggregationId() throws IOException {
    PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    builder.getAlertBuilder().getAlertAggregationBuilder().setAggregateGuid("test");
    String alertId = getAlertId(builder.build());
    assertNotNull(alertId);
    assertEquals("test", alertId);
  }

  @Test
  void testAlertUpdateTimestamp() throws IOException {
    PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    Timestamp timestamp = CbUaeAlertUtil.getAlertUpdateTimestamp(builder);
    assertNotNull(timestamp);
    assertEquals(builder.getAlert().getAlertUpdateTimestamp(), timestamp);
  }

  @Test
  void testAlertUpdateTimestampWithNull() throws IOException {
    PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    Timestamp timestamp = CbUaeAlertUtil.getAlertUpdateTimestamp(builder);
    assertNotNull(timestamp);
  }

  @Test
  void testUpdateCount() throws IOException {
    PSCAlert.CbUaeAlert.Builder alertBuilder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    alertBuilder.setUpdateCount(10L);
    Long updateCount = CbUaeAlertUtil.getUpdateCount(alertBuilder.build());
    assertNotNull(updateCount);
    assertEquals(10L, updateCount);
  }

  @Test
  void testUpdateCountWithNull() {
    Long updateCount = CbUaeAlertUtil.getUpdateCount(null);
    assertNotNull(updateCount);
    assertEquals(0L, updateCount);
  }

  @Test
  void getAlertIdFromBuilderWithAggregationId() throws IOException {
    PSCAlert.CbUaeAlert.Builder builder =
        PSCAlert.CbUaeAlert.newBuilder(UaeAlertSamples.hbfwAlert());
    builder.getAlertBuilder().getAlertAggregationBuilder().setAggregateGuid("test");
    String alertId = getAlertId(builder);
    assertNotNull(alertId);
    assertEquals("test", alertId);
  }

  @Test
  void getAlertUpdateTimestamp() throws IOException {
    final Timestamp timestamp = CbUaeAlertUtil.getAlertUpdateTimestamp(UaeAlertSamples.hbfwAlert());
    assertNotNull(timestamp);
    assertEquals(timestamp, UaeAlertSamples.hbfwAlert().getAlert().getAlertUpdateTimestamp());
  }

  @Test
  void getCurrentProtoBufTimeStamp() {
    assertNotNull(CbUaeAlertUtil.getCurrentProtoBufTimeStamp());
  }

  @Test
  void testFirstSeenTimeWithValidValue() throws IOException {
    PSCAlert.CbUaeAlert alert = UaeAlertSamples.hbfwAlert();
    final Timestamp timestamp = CbUaeAlertUtil.getFirstEventTimestamp(alert);
    assertNotNull(timestamp);
    assertEquals(alert.getFirstEventTimestamp(), timestamp);
  }

  @Test
  void testLastSeenTimeWithValidValue() throws IOException {
    PSCAlert.CbUaeAlert alert = UaeAlertSamples.hbfwAlert();
    final Timestamp timestamp = CbUaeAlertUtil.getLastEventTimestamp(alert);
    assertNotNull(timestamp);
    assertEquals(alert.getLastEventTimestamp(), timestamp);
  }

  @Test
  void testHighestThreatStateWithValidValue() throws IOException {
    PSCAlert.CbUaeAlert alert = UaeAlertSamples.hbfwAlert();
    final PSCRProto.CbAlert.ThreatExecutionState threatExecutionState =
        CbUaeAlertUtil.getHighestAlertExecutionState(alert);
    assertEquals(alert.getAlert().getHighestAlertExecutionState(), threatExecutionState);
  }

  @Test
  void testUpdateTypeAlertCreated() throws IOException {
    final PSCAlert.CbUaeAlert alert = UaeAlertSamples.hbfwAlert();
    final PSCAlert.CbUaeAlert.Builder alertBuilder = PSCAlert.CbUaeAlert.newBuilder(alert);
    PSCRProto.CbAlertUpdateDetails cbAlertUpdateDetails =
        PSCRProto.CbAlertUpdateDetails.newBuilder()
            .addUpdateTypes(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.ALERT_CREATED)
            .build();
    alertBuilder.getAlertBuilder().setAlertUpdateDetails(cbAlertUpdateDetails);
    boolean hasEventAdded = updateTypeIsEventAdded(alertBuilder);
    assertFalse(hasEventAdded);
  }

  @Test
  void testUpdateTypeEventAdded() throws IOException {
    final PSCAlert.CbUaeAlert alert = UaeAlertSamples.hbfwAlert();
    final PSCAlert.CbUaeAlert.Builder alertBuilder = PSCAlert.CbUaeAlert.newBuilder(alert);
    PSCRProto.CbAlertUpdateDetails cbAlertUpdateDetails =
        PSCRProto.CbAlertUpdateDetails.newBuilder()
            .addUpdateTypes(PSCRProto.CbAlertUpdateDetails.AlertUpdateType.EVENT_ADDED)
            .build();
    alertBuilder.getAlertBuilder().setAlertUpdateDetails(cbAlertUpdateDetails);
    boolean hasEventAdded = updateTypeIsEventAdded(alertBuilder);
    assertTrue(hasEventAdded);
  }

  @Test
  void testUpdateTypeWithNull() throws IOException {
    final PSCAlert.CbUaeAlert alert = UaeAlertSamples.hbfwAlert();
    final PSCAlert.CbUaeAlert.Builder alertBuilder = PSCAlert.CbUaeAlert.newBuilder(alert);
    boolean hasEventAdded = updateTypeIsEventAdded(alertBuilder);
    assertFalse(hasEventAdded);
  }
}
